/**
 * @file c_application.cc
 *
 * @brief Service Placement application for ns-2.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: c_application.cc 3719 2010-08-23 13:00:36Z gwitten $
 */

#include <float.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <queue>
#include <vector>

#include "app.h"
#include "packet.h"
#include "rng.h"
#include "scheduler.h"
#include "timer-handler.h"

#include "c_application.h"



/**
 * @brief Group the output of vprintf() by time.
 */
//#define VPRINTF_GROUP_OUTPUT_BY_TIME

/**
 * @brief Debug level for this component.
 */
#define C_APPLICATION_DEBUG_LEVEL	C_MSG_WARN



#define STATE_INIT		0x0	///< Application is initializing.
#define STATE_STOPPED		0x1	///< Application is stopped.
#define STATE_RUNNING		0x2	///< Application is running.
#define STATE_TRANSMIT_STOPPED	0x3	///< Application is transmitting a packet while stopped.
#define STATE_TRANSMIT_RUNNING	0x4	///< Application is transmitting a packet while running.
#define STATE_FINALIZED		0x5	///< Application is finalized.

/**
 * @brief Convert a numeric state represenstation into the corresponding
 * string.
 *
 * @param state Numeric representation of the state.
 * @return String representing the state.
 */
const char* STATE_TO_STRING(int state) {
	switch(state) {
	case STATE_INIT:		return "INIT";
	case STATE_STOPPED:		return "STOPPED";
	case STATE_RUNNING:		return "RUNNING";
	case STATE_TRANSMIT_STOPPED:	return "TRANSMIT/STOPPED";
	case STATE_TRANSMIT_RUNNING:	return "TRANSMIT/RUNNING";
	case STATE_FINALIZED:		return "FINALIZED";
	default: printf("\t\t\t\tCApplication: Illegal state (%d), aborting...\n", state); abort();
	}
}



using namespace C;

// Glue code to the Tcl interpreter.
static class CApplicationClass : public TclClass {
public:
	CApplicationClass() : TclClass("Application/C") {}
	TclObject* create(int, const char*const*) {
		return (new CApplication());
	}
} class_c_application;



CApplication* CApplication::instance_ = NULL;
double CApplication::timeOfLastCallToVprintf_ = 0;

/**
 * @brief CApplication constructor.
 */
CApplication::CApplication() : \
	Application()
{
	// One-time global initialization.
	if(instance_ == NULL) {
		// ...
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Creating new instance...\n", this);
	instance_ = this;
	timer_ = new CApplicationTimer(this);
	rng_ = new RNG();
	rng_->set_seed(RNG::HEURISTIC_SEED_SOURCE, 0);
// 	rng_->set_seed(RNG::PREDEF_SEED_SOURCE, 0);
	state_ = STATE_INIT;
	timeOfNextCallback_ = 0.0;

	// Export C variables.
	bind("DEBUG_ERRORs_", &C_DEBUG_ERRORs);
	bind("DEBUG_WARNs_", &C_DEBUG_WARNs);
	bind("averageNodeDegree_", &C_average_node_degree);
	bind("averageNodeDistance_", &C_average_node_distance);
	for(int i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++) {
		char str[64];	// Hack to bind array.
		snprintf(str, 64, "unicastPacketsPerHopCount_%d_", i);
		bind(str, &C_unicast_packets_per_hop_count[i]);
	}
	bind("transmittedUnicastPackets_", &C_transmitted_unicast_packets);
	bind("unicastTransmissionFailures_", &C_unicast_transmission_failures);
	bind("receivedUnicastPackets_", &C_received_unicast_packets);
	bind("transmittedBroadcastPackets_", &C_transmitted_broadcast_packets);
	bind("receivedBroadcastPackets_", &C_received_broadcast_packets);
	bind("transmittedBytes_", &C_transmitted_bytes);
	bind("receivedBytes_", &C_received_bytes);
	bind("transmittedDatarate_", &C_transmitted_datarate);
	bind("receivedDatarate_", &C_received_datarate);
	bind("networkMappingOverhead_", &C_network_mapping_overhead);
	bind("DYMO_transmittedRREQs_", &C_DYMO_transmitted_RREQs);
	bind("DYMO_transmittedRREPs_", &C_DYMO_transmitted_RREPs);
	bind("DYMO_transmittedRERRs_", &C_DYMO_transmitted_RERRs);
	bind("DYMO_transmittedUnicastPackets_", &C_DYMO_transmitted_unicast_packets);
	bind("DYMO_receivedUnicastPackets_", &C_DYMO_received_unicast_packets);
	bind("DYMO_transmittedBroadcastPackets_", &C_DYMO_transmitted_broadcast_packets);
	bind("DYMO_receivedBroadcastPackets_", &C_DYMO_received_broadcast_packets);
	bind("RUDP_totalTransmissions_", &C_RUDP_total_transmissions);
	bind("RUDP_successfulTransmissions_", &C_RUDP_successful_transmissions);
	bind("ICMP_transmittedNops_", &C_ICMP_transmitted_nops);
	bind("ICMP_receivedNops_", &C_ICMP_received_nops);
	bind("SLP_totalQueries_", &C_SLP_total_queries);
	bind("SLP_successfulQueries_", &C_SLP_successful_queries);
	bind("SERVICE_totalDataRequests_", &C_SERVICE_total_data_requests);
	bind("SERVICE_successfulDataRequests_", &C_SERVICE_successful_data_requests);
	bind("SERVICE_totalSuccessfulDataRequestDelay_", &C_SERVICE_total_successful_data_request_delay);
	bind("SERVICE_totalDataUpdates_", &C_SERVICE_total_data_updates);
	bind("SERVICE_successfulDataUpdates_", &C_SERVICE_successful_data_updates);
	bind("SERVICE_totalStreamRequests_", &C_SERVICE_total_stream_requests);
	bind("SERVICE_successfulStreamRequests_", &C_SERVICE_successful_stream_requests);
	bind("SERVICE_expectedStreamPackets_", &C_SERVICE_expected_stream_packets);
	bind("SERVICE_receivedStreamPackets_", &C_SERVICE_received_stream_packets);
	bind("SERVICE_totalRequests_", &C_SERVICE_total_requests);
	bind("SERVICE_successfulRequests_", &C_SERVICE_successful_requests);
	bind("SERVICE_failedRequests_", &C_SERVICE_failed_requests);
	bind("SERVICE_transmittedBytes_", &C_SERVICE_transmitted_bytes);
	bind("SERVICE_receivedBytes_", &C_SERVICE_received_bytes);
	bind("SERVICE_transmittedDatarate_", &C_SERVICE_transmitted_datarate);
	bind("SERVICE_receivedDatarate_", &C_SERVICE_received_datarate);
	for(int i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++) {
		char str[64];	// Hack to bind array.
		snprintf(str, 64, "SERVICE_requestPacketsPerHopCount_%d_", i);
		bind(str, &C_SERVICE_request_packets_per_hop_count[i]);
	}
	bind("SERVICE_currentlyHostedServices_", &C_SERVICE_currently_hosted_services);
	bind("SERVICE_averageHostedServices_", &C_SERVICE_average_hosted_services);
	bind("SERVICE_syncOverheadBytes_", &C_SERVICE_sync_overhead_bytes);
	bind("SERVICE_totalReplications_", &C_SERVICE_total_replications);
	bind("SERVICE_successfulReplications_", &C_SERVICE_successful_replications);
	bind("SERVICE_totalMigrations_", &C_SERVICE_total_migrations);
	bind("SERVICE_successfulMigrations_", &C_SERVICE_successful_migrations);
	bind("PLACEMENT_currentServiceCost_", &C_PLACEMENT_current_service_cost);
	bind("PLACEMENT_currentlyOptimalServiceCost_", &C_PLACEMENT_currently_optimal_service_cost);
	bind("PLACEMENT_currentlyOptimalNumberOfHosts_", &C_PLACEMENT_currently_optimal_number_of_hosts);
	bind("PLACEMENT_currentlyEstimatedAdaptationCost_", &C_PLACEMENT_currently_estimated_adaptation_cost);
	bind("PLACEMENT_totalAdaptations_", &C_PLACEMENT_total_adaptations);
	bind("PLACEMENT_successfulAdaptations_", &C_PLACEMENT_successful_adaptations);
	bind("PLACEMENT_failedAdaptations_", &C_PLACEMENT_failed_adaptations);

	state_ = STATE_STOPPED;

	// C code initializer is called when an agent is attached because we
	// don't know the local address at this point.
}

/**
 * @brief CApplication destructor.
 */
CApplication::~CApplication() {
	if(state_ != STATE_FINALIZED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to delete application, but it is not finalized (%s)...\n", this, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Deleting application...\n", this);

	// Cancel any remaining callbacks.
	if(timeOfNextCallback_ != 0.0)
		timer_->cancel();

	delete timer_;
	delete rng_;
}

/*
 * ns-2 entry points.
 */

int CApplication::command(int argc, const char*const* argv) {
	if(state_ == STATE_FINALIZED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Application has already been finalized, ignoring command \"%s\"...\n", this, argv[1]);
		return TCL_OK;
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Processing command \"%s\"...\n", this, argv[1]);
	instance_ = this;

//	Tcl& tcl = Tcl::instance();
	if(argc == 2) {
		if(strcmp(argv[1], "finalize") == 0) {
			C_finalize();
			state_ = STATE_FINALIZED;
			fflush(NULL);	// Flush all I/O streams.
			return TCL_OK;
		}
	} else if(argc == 3) {
		// Intercept "attach-agent" command, initialize the local
		// address and call C code initializer.
		if(strcmp(argv[1], "attach-agent") == 0) {
			int ret = Application::command(argc, argv);
			if(ret == TCL_OK) {
				C_init(agent_->addr(), (C_addr_t) IP_BROADCAST);
			}
			return ret;
		}
	} else if(argc == 4) {
		if(strcmp(argv[1], "seed") == 0) {
			int i = atoi(argv[3]);
			if(strcmp(argv[2], "raw") == 0) {
				rng_->set_seed(RNG::RAW_SEED_SOURCE, i);
			} else if(strcmp(argv[2], "predef") == 0) {
				rng_->set_seed(RNG::PREDEF_SEED_SOURCE, i);
			} else if(strcmp(argv[2], "heuristic") == 0) {
				rng_->set_seed(RNG::HEURISTIC_SEED_SOURCE, 0);
			}
			return TCL_OK;
		}
	}

	if(C_command(argc, argv))
		return TCL_OK;

	return Application::command(argc, argv);
}

void CApplication::start() {
	if(state_ != STATE_STOPPED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to start application, but it is not stopped (%s)...\n", this, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Starting...\n", this);
	instance_ = this;
	state_ = STATE_RUNNING;
	C_start();
}

void CApplication::stop() {
	if(state_ == STATE_FINALIZED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Ignoring request to stop application, which has already been finalized...\n", this);
		return;
	}

	if(state_ != STATE_RUNNING) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to stop application, but it is not running (%s)...\n", this, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Stopping...\n", this);
	instance_ = this;
	C_stop();
	state_ = STATE_STOPPED;
}

// Unused ns-2 entry point.
void CApplication::send(int nbytes) {
	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
		printf("\t\t\t\tCApplication(%p): Sending %d byte(s) on unused interface.\n", this, nbytes);
	abort();
}

// Unused ns-2 entry point.
void CApplication::recv(int nbytes) {
	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
		printf("\t\t\t\tCApplication(%p): Received %d byte(s) on unused interface.\n", this, nbytes);
	abort();
}

// ns-2 entry point for received packets.
void CApplication::recv(Packet* pkt, Handler*) {
	if(pkt == NULL) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Received a NULL packet.\n", this);
		abort();
	} else if(pkt->userdata()->size() < ((int) sizeof(net_header_t))) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Received %d byte(s) but expected at least %d bytes.\n", this, pkt->userdata()->size(), sizeof(net_header_t));
		abort();
	}

	unsigned char* p = ((PacketData*) pkt->userdata())->data();
	net_header_t* header = (net_header_t*) p;
#ifndef C_USE_INTERNAL_ROUTING
	header->hop_count += hdr_cmn::access(pkt)->num_forwards();
	header->last_hop_address = hdr_cmn::access(pkt)->prev_hop();
#endif	// C_USE_INTERNAL_ROUTING

	void* payload = NULL;
	if(header->payload_length > 0)
		payload = p + sizeof(net_header_t);

	if(state_ == STATE_FINALIZED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Trying to process %d byte(s) of application data from node #%d, but application is FINALIZED. Ignoring...\n", this, pkt->userdata()->size(), header->source_address);
		return;
	}

	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to process %d byte(s) of application data from node #%d, but application is neither running nor stopped (%s)...\n", this, pkt->userdata()->size(), header->source_address, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_INFO)
		printf("\t\t\t\tCApplication(%p): Processing %d byte(s) of application data from node #%d (seqno=%d, hops=%d, last_hop=%d)...\n", this, pkt->userdata()->size(), header->source_address, header->seqno, header->hop_count, header->last_hop_address);
	Tcl::instance().evalf("[Simulator instance] puts-nam-traceall \"r -t %f -s %d -d %d -e %d -c %d -i %d\"", Scheduler::instance().clock(), header->last_hop_address, header->next_hop_address, (sizeof(net_header_t) + header->payload_length), header->source_address, header->seqno);

	instance_ = this;
	C_recv(header, payload);
}

// Unused ns-2 entry point.
void CApplication::process_data(int size, AppData* data) {
	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
		printf("\t\t\t\tCApplication(%p): Received %d byte(s) of application data at %p on unused interface. Is processPackets_ in Agent/MessagePassing set to 1?\n", this, size, data);
	abort();
}

// ns-2 entry point upon completed packet transmission.
void CApplication::resume() {
	if(state_ == STATE_FINALIZED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Application has already been finalized, ignoring attempt to resume...\n", this);
		return;
	}
	if((state_ != STATE_TRANSMIT_STOPPED) && (state_ != STATE_TRANSMIT_RUNNING)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to resume application, but it is not transmitting (%s)...\n", this, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Resuming...\n", this);
	if(state_ == STATE_TRANSMIT_STOPPED)
		state_ = STATE_STOPPED;
	else	// state == STATE_TRANSMIT_RUNNING
		state_ = STATE_RUNNING;
}

// ns-2 entry point upon failed packet transmission.
void CApplication::rt_ll_failed(Packet *pkt) {
	if(pkt == NULL) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Failed to transmit a NULL packet.\n", this);
		abort();
	} else if(pkt->userdata()->size() < ((int) sizeof(net_header_t))) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Failed to transmit %d byte(s) of %d bytes.\n", this, pkt->userdata()->size(), sizeof(net_header_t));
		abort();
	}

	if(state_ == STATE_FINALIZED) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Application has already been finalized, ignoring report of broken link...\n", this);
		return;
	}

	unsigned char* p = ((PacketData*) pkt->userdata())->data();
	net_header_t* header = (net_header_t*) p;
	void* payload = NULL;
	if(header->payload_length > 0)
		payload = p + sizeof(net_header_t);

	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED) && (state_ != STATE_TRANSMIT_RUNNING) && (state_ != STATE_TRANSMIT_STOPPED)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Failed to transmit %d byte(s) of application data to node #%d, but application is neither running nor stopped nor transmitting while running nor transmitting while stopped (%s)...\n", this, pkt->userdata()->size(), header->destination_address, STATE_TO_STRING(state_));
		abort();
	} else if(header->destination_address == C_BROADCAST_ADDRESS) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Failed to transmit %d byte(s) of application data to the broadcast address, for which there should be no feedback...\n", this, pkt->userdata()->size());
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_INFO)
		printf("\t\t\t\tCApplication(%p): Failed to transmit %d byte(s) of application data to node #%d (seqno=%d, hops=%d, last_hop=%d)...\n", this, pkt->userdata()->size(), header->destination_address, header->seqno, header->hop_count, header->last_hop_address);
	instance_ = this;
	C_xmit_failure(header, payload);
}

/*
 * C code entry points.
 */

/**
 * @brief ns-2 specific implementation of __C_vprintf().
 */
void CApplication::__C_vprintf(const char *format, va_list ap) {
#ifdef VPRINTF_GROUP_OUTPUT_BY_TIME
	if(timeOfLastCallToVprintf_ != Scheduler::instance().clock())
		printf("\n");
#endif	// VPRINTF_GROUP_OUTPUT_BY_TIME
	vprintf(format, ap);
	fflush(NULL);
	timeOfLastCallToVprintf_ = Scheduler::instance().clock();
}

/**
 * @brief ns-2 specific implementation of __C_random().
 */
double CApplication::__C_random(double max) {
	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED) && (state_ != STATE_TRANSMIT_RUNNING) && (state_ != STATE_TRANSMIT_STOPPED)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to generate random number between 0 and %f, but application is neither running nor stopped nor transmitting while running  nor transmitting while stopped (%s)...\n", this, max, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_VERBOSE)
		printf("\t\t\t\tCApplication(%p): Generating random number between 0 and %f...\n", this, max);
	return rng_->uniform(max);
}

/**
 * @brief ns-2 specific implementation of __C_time().
 */
C_time_t CApplication::__C_time() {
	return Scheduler::instance().clock();
}

/**
 * @brief ns-2 specific implementation of __C_send().
 */
void CApplication::__C_send(const net_header_t* header, const void* payload) {
	unsigned int packet_size = sizeof(net_header_t) + header->payload_length;

	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying to send packet of %d byte(s) with sequence number %d to node #%d, but application is neither running nor stopped (%s)...\n", this, packet_size, header->seqno, header->destination_address, STATE_TO_STRING(state_));
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_INFO)
		printf("\t\t\t\tCApplication(%p): Sending packet #%d from node #%d to node #%d (%d bytes, 0x%X:%X)...\n", this, header->seqno, header->source_address, header->destination_address, packet_size, header->app, header->type);
//	Tcl::instance().evalf("[Simulator instance] puts-nam-traceall \"l -t %f -s %d -d %d -S UP -c black\"", Scheduler::instance().clock(), header->last_hop_address, header->next_hop_address);
	Tcl::instance().evalf("[Simulator instance] puts-nam-traceall \"h -t %f -s %d -d %d -e %d -c %d -i %d\"", Scheduler::instance().clock(), header->last_hop_address, header->next_hop_address, (sizeof(net_header_t) + header->payload_length), header->source_address, header->seqno);

	if(state_ == STATE_STOPPED)
		state_ = STATE_TRANSMIT_STOPPED;
	else	// state == STATE_RUNNING
		state_ = STATE_TRANSMIT_RUNNING;

#ifdef C_USE_INTERNAL_ROUTING
	ns_addr_t dst = {header->next_hop_address, agent_->port()};
#else
	ns_addr_t dst = {header->destination_address, agent_->port()};
#endif	// C_USE_INTERNAL_ROUTING
	PacketData* data = new PacketData(packet_size);
	memcpy(data->data(), header, sizeof(net_header_t));
	if(header->payload_length > 0)
		memcpy(data->data() + sizeof(net_header_t), payload, header->payload_length);
	agent_->sendto(packet_size, data, NULL, dst);
}

/**
 * @brief ns-2 specific implementation of __C_activate_callback().
 */
void CApplication::__C_activate_callback(C_time_t time) {
	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED) && (state_ != STATE_TRANSMIT_RUNNING) && (state_ != STATE_TRANSMIT_STOPPED)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trying activate callback to be called at %.2fs, but application is neither running nor stopped nor transmitting while running  nor transmitting while stopped (%s)...\n", this, time, STATE_TO_STRING(state_));
		abort();
	}

	if(time != 0.0) {

		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Activating callback to be called at %.2fs...\n", this, time);
	
		// Schedule next callback.
		double delay = time - Scheduler::instance().clock();
		if(delay < (0 - FLT_EPSILON)) {
			if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
				printf("\t\t\t\tCApplication(%p): Trying to activate callback in the past (%f < %f)...\n", this, time, Scheduler::instance().clock());
			abort();
		}
		timeOfNextCallback_ = time;
		timer_->resched(delay);

	} else {

		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
			printf("\t\t\t\tCApplication(%p): Cancelling callback...\n", this);
	
		timeOfNextCallback_ = 0.0;
		timer_->cancel();

	}
}

/**
 * @brief Size of the trace output buffer.
 */
#define TRACE_OUTPUT_SIZE	256

/**
 * @brief ns-2 specific implementation of __C_trace().
 */
void CApplication::__C_trace(const char *format, va_list ap) {
	char trace_output[TRACE_OUTPUT_SIZE];
	if(vsnprintf(trace_output, TRACE_OUTPUT_SIZE, format, ap) >= TRACE_OUTPUT_SIZE) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Trace output is too big for allocated buffer.\n", this);
		abort();
	}

	Tcl::instance().evalf("[Simulator instance] puts-nam-traceall \"%s\"", trace_output);
	Tcl::instance().evalf("[Simulator instance] flush-trace");
}



/*
 * Timer entry points.
 */

void CApplication::expire() {
	double currentTime = Scheduler::instance().clock();

	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_WARN)
			printf("\t\t\t\tCApplication(%p): Trying to execute callback, but application is neither running nor stopped (%s)...\n", this, STATE_TO_STRING(state_));
		return;
	} else if(timeOfNextCallback_ == 0.0) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Timer expired but none was expected.\n", this);
		abort();
	} else if(fabsf(currentTime - timeOfNextCallback_) > FLT_EPSILON) {
		if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_FATAL)
			printf("\t\t\t\tCApplication(%p): Timer expired at incorrect time (%f != %f).\n", this, currentTime, timeOfNextCallback_);
		abort();
	}

	if(C_APPLICATION_DEBUG_LEVEL <= C_MSG_DEBUG)
		printf("\t\t\t\tCApplication(%p): Timer expired, executing callback...\n", this);

	// Execute callback.
	instance_ = this;
	timeOfNextCallback_ = 0.0;
	C_execute_callback();
}



/**
 * CApplicationTimer
 **/

CApplicationTimer::CApplicationTimer(CApplication* app) : \
	TimerHandler(), app_(app)
{}

void CApplicationTimer::expire(Event*) {
	app_->expire();
}



/**
 * Include C code
 **/

// Calls into C code.
#define C_init CApplication::C_init
#define C_start CApplication::C_start
#define C_stop CApplication::C_stop
#define C_finalize CApplication::C_finalize
#define C_recv CApplication::C_recv
#define C_xmit_failure CApplication::C_xmit_failure
#define C_execute_callback CApplication::C_execute_callback
#define C_command CApplication::C_command

// Calls from C code.
#define __C_vprintf CApplication::instance_->__C_vprintf
#define __C_random CApplication::instance_->__C_random
#define __C_time CApplication::instance_->__C_time
#define __C_send CApplication::instance_->__C_send
#define __C_activate_callback CApplication::instance_->__C_activate_callback
#define __C_trace CApplication::instance_->__C_trace

// Global C variables.
#define C_DEBUG_ERRORs CApplication::instance_->C_DEBUG_ERRORs
#define C_DEBUG_WARNs CApplication::instance_->C_DEBUG_WARNs
#define C_state CApplication::instance_->C_state
#define C_init_time CApplication::instance_->C_init_time
#define C_stop_time CApplication::instance_->C_stop_time
#define C_check_consistency_on_shutdown CApplication::instance_->C_check_consistency_on_shutdown
#define C_callbacks CApplication::instance_->C_callbacks
#define C_next_callback_id CApplication::instance_->C_next_callback_id
#define C_registered_callbacks CApplication::instance_->C_registered_callbacks
#define C_executed_callbacks CApplication::instance_->C_executed_callbacks
#define C_local_address CApplication::instance_->C_local_address
#define C_broadcast_address CApplication::instance_->C_broadcast_address
#define C_addr_to_str_buffer CApplication::instance_->C_addr_to_str_buffer
#define C_addr_to_str_buffer_index CApplication::instance_->C_addr_to_str_buffer_index
#define C_sequence_number CApplication::instance_->C_sequence_number
#define C_send_callback_id CApplication::instance_->C_send_callback_id
#define C_received_packet_cache CApplication::instance_->C_received_packet_cache
#define C_known_nodes CApplication::instance_->C_known_nodes
#define C_network_map CApplication::instance_->C_network_map
#define C_network_map_may_need_purging CApplication::instance_->C_network_map_may_need_purging
#define C_beacon_callback_id CApplication::instance_->C_beacon_callback_id
#define C_edge_cost_matrix CApplication::instance_->C_edge_cost_matrix
#define C_distance_matrix CApplication::instance_->C_distance_matrix
#define C_periodic_statistics_measurement_callback_id CApplication::instance_->C_periodic_statistics_measurement_callback_id
#define C_periodic_statistics_measurement_samples CApplication::instance_->C_periodic_statistics_measurement_samples
#define C_node_degree_running_sum CApplication::instance_->C_node_degree_running_sum
#define C_average_node_degree CApplication::instance_->C_average_node_degree
#define C_node_distance_running_sum CApplication::instance_->C_node_distance_running_sum
#define C_average_node_distance CApplication::instance_->C_average_node_distance
#define C_unicast_packets_per_hop_count CApplication::instance_->C_unicast_packets_per_hop_count
#define C_transmitted_unicast_packets CApplication::instance_->C_transmitted_unicast_packets
#define C_unicast_transmission_failures CApplication::instance_->C_unicast_transmission_failures
#define C_received_unicast_packets CApplication::instance_->C_received_unicast_packets
#define C_transmitted_broadcast_packets CApplication::instance_->C_transmitted_broadcast_packets
#define C_received_broadcast_packets CApplication::instance_->C_received_broadcast_packets
#define C_transmitted_bytes CApplication::instance_->C_transmitted_bytes
#define C_received_bytes CApplication::instance_->C_received_bytes
#define C_transmitted_datarate_stat CApplication::instance_->C_transmitted_datarate_stat
#define C_transmitted_datarate CApplication::instance_->C_transmitted_datarate
#define C_received_datarate_stat CApplication::instance_->C_received_datarate_stat
#define C_received_datarate CApplication::instance_->C_received_datarate
#define C_network_mapping_overhead CApplication::instance_->C_network_mapping_overhead
// DYMO
#define C_DYMO_own_seq_num CApplication::instance_->C_DYMO_own_seq_num
#define C_DYMO_routing_table CApplication::instance_->C_DYMO_routing_table
#define C_DYMO_packet_queue CApplication::instance_->C_DYMO_packet_queue
#define C_DYMO_transmitted_RREQs CApplication::instance_->C_DYMO_transmitted_RREQs
#define C_DYMO_transmitted_RREPs CApplication::instance_->C_DYMO_transmitted_RREPs
#define C_DYMO_transmitted_RERRs CApplication::instance_->C_DYMO_transmitted_RERRs
#define C_DYMO_transmitted_unicast_packets CApplication::instance_->C_DYMO_transmitted_unicast_packets
#define C_DYMO_received_unicast_packets CApplication::instance_->C_DYMO_received_unicast_packets
#define C_DYMO_transmitted_broadcast_packets CApplication::instance_->C_DYMO_transmitted_broadcast_packets
#define C_DYMO_received_broadcast_packets CApplication::instance_->C_DYMO_received_broadcast_packets
// ICMP-light
#define C_ICMP_active_echo_requests CApplication::instance_->C_ICMP_active_echo_requests
#define C_ICMP_next_echo_id CApplication::instance_->C_ICMP_next_echo_id
#define C_ICMP_transmitted_nops CApplication::instance_->C_ICMP_transmitted_nops
#define C_ICMP_received_nops CApplication::instance_->C_ICMP_received_nops
// RUDP-light
#define C_RUDP_transmission_queue CApplication::instance_->C_RUDP_transmission_queue
#define C_RUDP_received_packet_cache CApplication::instance_->C_RUDP_received_packet_cache
#define C_RUDP_total_transmissions CApplication::instance_->C_RUDP_total_transmissions
#define C_RUDP_successful_transmissions CApplication::instance_->C_RUDP_successful_transmissions
#define C_RUDP_failed_transmissions CApplication::instance_->C_RUDP_failed_transmissions
// SLP-light
#define C_SLP_active_queries CApplication::instance_->C_SLP_active_queries
#define C_SLP_total_queries CApplication::instance_->C_SLP_total_queries
#define C_SLP_successful_queries CApplication::instance_->C_SLP_successful_queries
#define C_SLP_failed_queries CApplication::instance_->C_SLP_failed_queries
#define C_SLP_service_cache CApplication::instance_->C_SLP_service_cache
#define C_SLP_unavailable_services CApplication::instance_->C_SLP_unavailable_services
// Service
#define C_SERVICE_total_data_requests CApplication::instance_->C_SERVICE_total_data_requests
#define C_SERVICE_successful_data_requests CApplication::instance_->C_SERVICE_successful_data_requests
#define C_SERVICE_failed_data_requests CApplication::instance_->C_SERVICE_failed_data_requests
#define C_SERVICE_total_successful_data_request_delay CApplication::instance_->C_SERVICE_total_successful_data_request_delay
#define C_SERVICE_total_data_updates CApplication::instance_->C_SERVICE_total_data_updates
#define C_SERVICE_successful_data_updates CApplication::instance_->C_SERVICE_successful_data_updates
#define C_SERVICE_failed_data_updates CApplication::instance_->C_SERVICE_failed_data_updates
#define C_SERVICE_total_stream_requests CApplication::instance_->C_SERVICE_total_stream_requests
#define C_SERVICE_successful_stream_requests CApplication::instance_->C_SERVICE_successful_stream_requests
#define C_SERVICE_failed_stream_requests CApplication::instance_->C_SERVICE_failed_stream_requests
#define C_SERVICE_expected_stream_packets CApplication::instance_->C_SERVICE_expected_stream_packets
#define C_SERVICE_received_stream_packets CApplication::instance_->C_SERVICE_received_stream_packets
#define C_SERVICE_total_requests CApplication::instance_->C_SERVICE_total_requests
#define C_SERVICE_successful_requests CApplication::instance_->C_SERVICE_successful_requests
#define C_SERVICE_failed_requests CApplication::instance_->C_SERVICE_failed_requests
#define C_SERVICE_transmitted_bytes CApplication::instance_->C_SERVICE_transmitted_bytes
#define C_SERVICE_received_bytes CApplication::instance_->C_SERVICE_received_bytes
#define C_SERVICE_transmitted_datarate_stat CApplication::instance_->C_SERVICE_transmitted_datarate_stat
#define C_SERVICE_transmitted_datarate CApplication::instance_->C_SERVICE_transmitted_datarate
#define C_SERVICE_received_datarate_stat CApplication::instance_->C_SERVICE_received_datarate_stat
#define C_SERVICE_received_datarate CApplication::instance_->C_SERVICE_received_datarate 
#define C_SERVICE_request_packets_per_hop_count CApplication::instance_->C_SERVICE_request_packets_per_hop_count
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
#define C_SERVICE_indicate_request_callback_id CApplication::instance_->C_SERVICE_indicate_request_callback_id
#define C_SERVICE_indicate_activity_callback_id CApplication::instance_->C_SERVICE_indicate_activity_callback_id
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
#define C_SERVICE_hosted_services CApplication::instance_->C_SERVICE_hosted_services
#define C_SERVICE_served_streams CApplication::instance_->C_SERVICE_served_streams
#define C_SERVICE_outgoing_replications CApplication::instance_->C_SERVICE_outgoing_replications
#define C_SERVICE_incoming_replications CApplication::instance_->C_SERVICE_incoming_replications
#define C_SERVICE_currently_hosted_services CApplication::instance_->C_SERVICE_currently_hosted_services
#define C_SERVICE_hosted_services_running_sum CApplication::instance_->C_SERVICE_hosted_services_running_sum
#define C_SERVICE_average_hosted_services CApplication::instance_->C_SERVICE_average_hosted_services
#define C_SERVICE_available_until_time CApplication::instance_->C_SERVICE_available_until_time
#define C_SERVICE_sync_overhead_bytes CApplication::instance_->C_SERVICE_sync_overhead_bytes
#define C_SERVICE_total_replications CApplication::instance_->C_SERVICE_total_replications
#define C_SERVICE_successful_replications CApplication::instance_->C_SERVICE_successful_replications
#define C_SERVICE_failed_replications CApplication::instance_->C_SERVICE_failed_replications
#define C_SERVICE_total_migrations CApplication::instance_->C_SERVICE_total_migrations
#define C_SERVICE_successful_migrations CApplication::instance_->C_SERVICE_successful_migrations
#define C_SERVICE_failed_migrations CApplication::instance_->C_SERVICE_failed_migrations
// Placement
#define C_PLACEMENT_currently_outgoing_adaptation CApplication::instance_->C_PLACEMENT_currently_outgoing_adaptation
#define C_PLACEMENT_services_to_shutdown_after_delay CApplication::instance_->C_PLACEMENT_services_to_shutdown_after_delay
#define C_PLACEMENT_current_service_cost CApplication::instance_->C_PLACEMENT_current_service_cost
#define C_PLACEMENT_currently_optimal_service_cost CApplication::instance_->C_PLACEMENT_currently_optimal_service_cost
#define C_PLACEMENT_currently_optimal_number_of_hosts CApplication::instance_->C_PLACEMENT_currently_optimal_number_of_hosts
#define C_PLACEMENT_currently_estimated_adaptation_cost CApplication::instance_->C_PLACEMENT_currently_estimated_adaptation_cost
#define C_PLACEMENT_total_adaptations CApplication::instance_->C_PLACEMENT_total_adaptations
#define C_PLACEMENT_successful_adaptations CApplication::instance_->C_PLACEMENT_successful_adaptations
#define C_PLACEMENT_failed_adaptations CApplication::instance_->C_PLACEMENT_failed_adaptations
// Connectivity/interference graphs.
#define C_INTERFERENCE_active_measurements CApplication::instance_->C_INTERFERENCE_active_measurements
#define C_INTERFERENCE_local_data CApplication::instance_->C_INTERFERENCE_local_data

// C code files.
#include "../../interface.c"
#include "../../util/debug.c"
#include "../../util/list.c"
#include "../../util/callback.c"
#include "../../util/net.c"
#include "../../util/mapping.c"
#include "../../util/stats.c"
#include "../../util/trigger.c"
#include "../../dymo/dymo.c"
#include "../../icmp-light/icmp-light.c"
#include "../../rudp-light/rudp-light.c"
#include "../../slp-light/slp-light.c"
#include "../../slp-light/slp_cache.c"
#include "../../service/service.c"
#include "../../service/service_client.c"
#include "../../service/service_server.c"
#include "../../service/service_management.c"
#include "../../service/service_replication.c"
#include "../../service/service_stats.c"
#include "../../service/service_util.c"
#include "../../placement/placement.c"
#include "../../placement/simple_heuristic/simple_heuristic.c"
#include "../../placement/graph_cost_single_instance/graph_cost_single_instance.c"
#include "../../placement/graph_cost_multiple_instances/graph_cost_multiple_instances.c"
#include "../../placement/graph_cost_multiple_instances/gcmi_cost_calculation.c"
#include "../../placement/graph_cost_multiple_instances/gcmi_clustering.c"
#include "../../placement/linkpull_migration/linkpull_migration.c"
#include "../../placement/peerpull_migration/peerpull_migration.c"
#include "../../placement/topocenter_migration/topocenter_migration.c"
#include "../../placement/tree-topology_migration/tree-topology_migration.c"
#include "../../placement/asg_simple_algorithm/asg_simple_algorithm.c"
#include "../../placement/asg_event_flow_tree/asg_event_flow_tree.c"
#include "../../interference/interference.c"
#include "../../interference/interference_local.c"
