/**
 * @file interference.c
 *
 * @brief Interference graphs.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interference.c 3289 2010-02-13 19:49:19Z gwitten $
 */

#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../icmp-light/icmp-light.h"
#include "../rudp-light/rudp-light.h"
#include "interference.h"
#include "interference_types.h"
#include "interference_vars.h"
#include "interference_local.h"



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the INTERFERENCE module.
 */
void INTERFERENCE_init() {
	C_INTERFERENCE_active_measurements = NULL;
	list_alloc(&C_INTERFERENCE_active_measurements);
	C_INTERFERENCE_local_data = NULL;
}

/**
 * @brief Start the INTERFERENCE module.
 */
void INTERFERENCE_start() {
}

/**
 * @brief Stop the INTERFERENCE module.
 */
void INTERFERENCE_stop() {
}

/**
 * @brief Finalize the INTERFERENCE module.
 */
void INTERFERENCE_finalize() {
	// TODO: Properly free the memory pointed to by the list items.
	list_free_including_data(&C_INTERFERENCE_active_measurements, free);
}



/*
 * Utility functions.
 */

/**
 * @brief Free a global data node structure.
 *
 * @param p Pointer to global data node structure to free.
 */
static void free_global_data_node_list(void* p) {
	INTERFERENCE_global_data_node_t* global_data_node = (INTERFERENCE_global_data_node_t*) p;
	if(global_data_node->local_data_list != NULL)
		list_free_including_data(&global_data_node->local_data_list, free_local_data_list);
	FREE(global_data_node);
}

/*
 * @brief Free a global data node structure.
 *
 * @param global_data_node Pointer to pointer to global data node structure to
 * free.
 */
/*  Unused, but retained for interface completeness.
static void free_global_data_node(INTERFERENCE_global_data_node_t** global_data_node) {
	free_global_data_node_list(*global_data_node);
	*global_data_node = NULL;
}
*/


/*
 * Packet handling functions.
 */

/**
 * @brief Check whether two global data nodes have the same address.
 *
 * @param item Data node to check.
 * @param cmp_item Reference data node.
 * @return <code>true</code> if the data nodes have the same address,
 * <code>false</code> otherwise.
 */
static bool cmp_global_data_node_address(const void* item, const void* cmp_item) {
	INTERFERENCE_global_data_node_t* data_node = (INTERFERENCE_global_data_node_t*) item;
	INTERFERENCE_global_data_node_t* cmp_data_node = (INTERFERENCE_global_data_node_t*) cmp_item;
	return (data_node->node.address == cmp_data_node->node.address);
}

static void measure_iterate(INTERFERENCE_measurement_t* graph_measurement);

/**
 * @brief RUDP callback for sending a INTERFERENCE REQUEST packet.
 *
 * @param parameter RUDP transmission result.
 */
static void send_INTERFERENCE_REQUEST_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	C_addr_t* to = (C_addr_t*) result->parameter;

	if(result->error_code == C_ERROR_OK) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: INTERFERENCE REQUEST to %s was successful.\n", addr_to_str(*to));
	} else {
		INTERFERENCE_measurement_t* graph_measurement = NULL;
		INTERFERENCE_global_data_node_t* data_node = NULL;
		INTERFERENCE_global_data_node_t cmp_data_node;

		// This timeout might trigger after the graph measurement is
		// already complete, hence we have to check. Note that there is
		// a slight chance that we actually getting a reference to the
		// NEXT graph measurement process, but in reality this
		// shouldn't happen as graph measurement is only done once.
		if(list_empty(C_INTERFERENCE_active_measurements)) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: INTERFERENCE REQUEST to %s timed out, but graph measurement has already been completed...\n", addr_to_str(*to));
			return;
		}

		// NOTE: Since only one graph measurement process can be
		// active, we can just get the head of the list of active
		// measurements.
		graph_measurement = (INTERFERENCE_measurement_t*) list_head(C_INTERFERENCE_active_measurements);
		memset(&cmp_data_node, 0, sizeof(INTERFERENCE_global_data_node_t));
		cmp_data_node.node.address = *to;
		data_node = (INTERFERENCE_global_data_node_t*) list_find(graph_measurement->result->data_nodes, &cmp_data_node, cmp_global_data_node_address);
		C_ABORT_ON(data_node == NULL, "INTERFERENCE: INTERFERENCE REQUEST to %s timed out, but it is not in network list...\n", addr_to_str(*to));
		if(data_node->local_data_list != NULL) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: INTERFERENCE REQUEST to %s timed out, but node has already received data...\n", addr_to_str(*to));
			return;
		} else if(*to != graph_measurement->active_node) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: INTERFERENCE REQUEST to %s timed out, but active node is %s...\n", addr_to_str(*to), addr_to_str(graph_measurement->active_node));
			return;
		}

		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: INTERFERENCE REQUEST failed (%d), skipping %s...\n", result->error_code, addr_to_str(*to));
		callback_cancel(graph_measurement->callback_id);
		list_alloc(&data_node->local_data_list);

		// Continue iterating over all nodes.
		measure_iterate(graph_measurement);
	}

	FREE(to);
	FREE(result);
}



/**
 * @brief Send a INTERFERENCE REQUEST packet.
 *
 * @param to Address of the destination node.
 * @param graph_measurement Pointer to graph measurement data structure.
 */
static void send_INTERFERENCE_REQUEST(C_addr_t to, INTERFERENCE_measurement_t* graph_measurement) {
	node_t* node_array = NULL;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Sending INTERFERENCE REQUEST to node #%d...\n", to);
	MALLOC_ARRAY(node_array, node_t, list_size(graph_measurement->nodes));
	list_to_array(graph_measurement->nodes, node_array, sizeof(node_t));
	MALLOC(result, RUDP_transmission_result_t);
	MALLOC(result->parameter, C_addr_t);
	*((C_addr_t*) result->parameter) = to;
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_INTERFERENCE, INTERFERENCE_PACKET_TYPE_REQUEST, node_array, (list_size(graph_measurement->nodes) * sizeof(node_t)), send_INTERFERENCE_REQUEST_callback, result);
	FREE(node_array);
}

/**
 * @brief Handle an incoming INTERFERENCE REPLY packet.
 *
 * Add the data collected by the remote node to the global data and start the
 * next iteration of the measurement process. If a reply is received from a
 * node other than the node that is currantly active, a rollback is scheduled.
 *
 * @param header Pointer to header of the incoming packet.
 * @param payload Pointer to the payload that contains a list of nodes with
 * data on network conditions and connectivty.
 */
static void handle_INTERFERENCE_REPLY(const net_header_t* header, const void* payload) {
	INTERFERENCE_measurement_t* graph_measurement = NULL;
	INTERFERENCE_global_data_node_t* data_node = NULL;
	INTERFERENCE_global_data_node_t cmp_data_node;

	// NOTE: Since only one graph measurement process can be active, we
	// can just get the head of the list of active measurements.
	if(list_empty(C_INTERFERENCE_active_measurements)) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Received INTERFERENCE REPLY from %s, but no graph measurement is active.\n", addr_to_str(header->source_address));
		return;
	}

	graph_measurement = (INTERFERENCE_measurement_t*) list_head(C_INTERFERENCE_active_measurements);
	if(header->source_address != graph_measurement->active_node) {
		if(graph_measurement->rollback_necessary) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Received INTERFERENCE REPLY from %s, but active node is %s and a rollback has already been scheduled previously to %s.\n", addr_to_str(header->source_address), addr_to_str(graph_measurement->active_node), addr_to_str(graph_measurement->rollback_node_address));

			// TODO: Properly handle multiple rollbacks.
			// When dealing with multiple rollbacks we have to
			// rollback to the node that is first in the list of
			// all nodes in the network. The implementation of this
			// a bit nasty. The error has a very low probability of
			// ever occuring, so we just report it for now.

			return;
		}

		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Received INTERFERENCE REPLY from %s, but active node is %s. Scheduling rollback to %s (inclusive)...\n", addr_to_str(header->source_address), addr_to_str(graph_measurement->active_node), addr_to_str(header->source_address));
		graph_measurement->rollback_necessary = true;
		graph_measurement->rollback_node_address = header->source_address;
		return;
	}

	memset(&cmp_data_node, 0, sizeof(INTERFERENCE_global_data_node_t));
	cmp_data_node.node.address = header->source_address;
	data_node = (INTERFERENCE_global_data_node_t*) list_find(graph_measurement->result->data_nodes, &cmp_data_node, cmp_global_data_node_address);
	if(data_node == NULL) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Received INTERFERENCE REPLY from %s, which is not in network list...\n", addr_to_str(header->source_address));
		return;
	}

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Received INTERFERENCE REPLY from %s...\n", addr_to_str(header->source_address));
	callback_cancel(graph_measurement->callback_id);
	list_alloc(&data_node->local_data_list);
	list_from_array(data_node->local_data_list, payload, sizeof(INTERFERENCE_local_data_node_t), header->payload_length);

	// Continue iterating over all nodes.
	measure_iterate(graph_measurement);
}

/**
 * @brief Handle an incoming INTERFERENCE RESET packet.
 *
 * Add the newly discovered nodes to the global list of all nodes in the
 * network, discard the data collected so far and restart the collection
 * process.
 *
 * @param header Pointer to header of the incoming packet.
 * @param payload Pointer to the payload that contains a list of newly
 * discovered nodes.
 */
static void handle_INTERFERENCE_RESET(const net_header_t* header, const void* payload) {
	INTERFERENCE_measurement_t* graph_measurement = NULL;
	list_t* newly_discovered_nodes = NULL;
	INTERFERENCE_global_data_node_t* old_data_node = NULL;
	list_iterator_t data_node_iterator;

	// NOTE: Since only one graph measurement process can be active, we
	// can just get the head of the list of active measurements.
	if(list_empty(C_INTERFERENCE_active_measurements)) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Received INTERFERENCE RESET from %s, but no graph measurement is active.\n", addr_to_str(header->source_address));
		return;
	}

	graph_measurement = (INTERFERENCE_measurement_t*) list_head(C_INTERFERENCE_active_measurements);
	if(header->source_address != graph_measurement->active_node) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Received INTERFERENCE RESET from %s, but active node is %s.\n", addr_to_str(header->source_address), addr_to_str(graph_measurement->active_node));

		// TODO: Handle resets at inactive nodes.
		// The correct way of handling this case is similar to dealing
		// with multiple REPLY packets: We have to wait for the other
		// node to report back and then clean up. The implementation
		// of this a bit nasty, since theoretically this can happen
		// between several nodes, not only two, and we have to add the
		// newly discovered only once to the global list of nodes. The
		// error has a very low probability of ever occuring, so we
		// just report it for now.

		return;
	}

	graph_measurement->rollback_necessary = false;
	graph_measurement->rollback_node_address = C_BROADCAST_ADDRESS;
	callback_cancel(graph_measurement->callback_id);

	list_alloc(&newly_discovered_nodes);
	list_from_array(newly_discovered_nodes, payload, sizeof(node_t), header->payload_length);
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_NOTE, "INTERFERENCE: Received INTERFERENCE RESET from %s, adding %d node(s) to node list: ", addr_to_str(header->source_address), list_size(newly_discovered_nodes));
	list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), newly_discovered_nodes, print_node_address);
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), "\n");

	// Discard previous results.
	list_clear_including_data(graph_measurement->result->data_nodes, free_global_data_node_list);

	// Add newly discovered nodes to the global list of nodes.
	while(!list_empty(newly_discovered_nodes)) {
		node_t* node = (node_t*) list_head(newly_discovered_nodes);
		list_add(graph_measurement->nodes, node);
		list_remove(newly_discovered_nodes, node);
	}
	list_free(&newly_discovered_nodes);
	list_sort(graph_measurement->nodes, node_address_less_than);

	// Re-add all nodes to the list of data nodes.
	list_iterator_init(graph_measurement->nodes, &data_node_iterator);
	while((old_data_node = (INTERFERENCE_global_data_node_t*) list_iterator_next(&data_node_iterator)) != NULL) {
		INTERFERENCE_global_data_node_t* new_data_node = NULL;
		MALLOC(new_data_node, INTERFERENCE_global_data_node_t);
		memcpy(&new_data_node->node, old_data_node, sizeof(node_t));
		new_data_node->local_data_list = NULL;
		list_add(graph_measurement->result->data_nodes, new_data_node);
	}

	// Print list of nodes for debugging.
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: %d node(s) in network: ", list_size(graph_measurement->nodes));
	list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), graph_measurement->nodes, print_node_address);
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");

	// Restart iterating over all nodes.
	measure_iterate(graph_measurement);
}

/**
 * @brief INTERFERENCE packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void INTERFERENCE_recv(const net_header_t* header, const void* payload) {
	C_ABORT_ON(header->app != APP_TYPE_INTERFERENCE, "INTERFERENCE: Received non-INTERFERENCE packet.\n");

	switch(header->type) {
	case INTERFERENCE_PACKET_TYPE_REQUEST:
		handle_INTERFERENCE_REQUEST(header, payload);
		break;
	case INTERFERENCE_PACKET_TYPE_REPLY:
		handle_INTERFERENCE_REPLY(header, payload);
		break;
	case INTERFERENCE_PACKET_TYPE_RESET:
		handle_INTERFERENCE_RESET(header, payload);
		break;
	case INTERFERENCE_PACKET_TYPE_TEST_REQUEST:
		C_ABORT_ON(header->payload_length != sizeof(INTERFERENCE_TEST_REQUEST_packet_t), "INTERFERENCE: Received TEST REQUEST packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(INTERFERENCE_TEST_REQUEST_packet_t));
		handle_INTERFERENCE_TEST_REQUEST(header, (INTERFERENCE_TEST_REQUEST_packet_t*) payload);
		break;
	case INTERFERENCE_PACKET_TYPE_TEST_DATA:
		C_ABORT_ON(header->payload_length != INTERFERENCE_INTERFERENCE_PAYLOAD_SIZE, "INTERFERENCE: Received TEST DATA packet with wrong payload length (%d != %d).\n", header->payload_length, INTERFERENCE_INTERFERENCE_PAYLOAD_SIZE);
		// NOTE: We're ignoring that the payload also contains dummy
		// data for testing.
		handle_INTERFERENCE_TEST_DATA(header, (INTERFERENCE_TEST_DATA_packet_t*) payload);
		break;
	default:
		C_ABORT_ON(true, "INTERFERENCE: Received packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Timeout for request for local information.
 *
 * This timeout is triggered if a node hasn't replied after a certain delay.
 *
 * @param parameter Graph measurement data structure.
 */
static void measure_iterate_timeout(callback_parameter_t parameter) {
	INTERFERENCE_measurement_t* graph_measurement = (INTERFERENCE_measurement_t*) parameter;
	INTERFERENCE_global_data_node_t* data_node = NULL;
	INTERFERENCE_global_data_node_t cmp_data_node;
	cmp_data_node.node.address = graph_measurement->active_node;
	data_node = (INTERFERENCE_global_data_node_t*) list_find(graph_measurement->result->data_nodes, &cmp_data_node, cmp_global_data_node_address);
	C_ABORT_ON(data_node == NULL, "INTERFERENCE: Requesting data from %s timed out, but it is not in network list...\n", addr_to_str(graph_measurement->active_node));
	C_ABORT_ON(data_node->local_data_list != NULL, "INTERFERENCE: Requesting data from %s timed out, but node has already received data...\n", addr_to_str(graph_measurement->active_node));

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Requesting data from %s has timed out, proceeding with next node...\n", addr_to_str(graph_measurement->active_node));
	list_alloc(&data_node->local_data_list);

	// Continue iterating over all nodes.
	measure_iterate(graph_measurement);
}

/**
 * @brief Delayed sending of the request for local information.
 *
 * The intention of this delay is to allow the network to settle down between
 * requests and to make sure that queues on all layers are empty when starting
 * the next request.
 *
 * @param parameter Graph measurement data structure.
 */
static void measure_iterate_delay(callback_parameter_t parameter) {
	INTERFERENCE_measurement_t* graph_measurement = (INTERFERENCE_measurement_t*) parameter;
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_NOTE, "INTERFERENCE: Requesting data from %s...\n", addr_to_str(graph_measurement->active_node));

	// The data request timeout depends on how many nodes need to be
	// queried locally by each node.
	graph_measurement->callback_id = callback_add(measure_iterate_timeout, graph_measurement, INTERFERENCE_PER_NODE_TIMEOUT);
	send_INTERFERENCE_REQUEST(graph_measurement->active_node, graph_measurement);
}

/**
 * @brief Check whether a global data node has no local data list.
 *
 * @param item Node to check.
 * @param cmp_item Reference node (not used).
 * @return <code>true</code> if the node has no local data list,
 * <code>false</code> otherwise.
 */
static bool node_without_local_data_list(const void* item, const void* cmp_item) {
	INTERFERENCE_global_data_node_t* data_node = (INTERFERENCE_global_data_node_t*) item;
	return (data_node->local_data_list == NULL);
}

/**
 * @brief Iterate over all nodes in the network and request each node to
 * measure its connectivity and interference lists.
 *
 * In case a rollback is scheduled, it is performed in this function prior to
 * the next iteration.
 *
 * @param graph_measurement Graph measurement data structure.
 */
static void measure_iterate(INTERFERENCE_measurement_t* graph_measurement) {
	INTERFERENCE_global_data_node_t* data_node = NULL;

	// Perform any scheduled rollback.
	if(graph_measurement->rollback_necessary) {
		unsigned int i = 0;
		bool clear_results = false;

		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_NOTE, "INTERFERENCE: Rolling back to %s (inclusive)...\n", addr_to_str(graph_measurement->rollback_node_address));
		while(i < list_size(graph_measurement->result->data_nodes)) {
			INTERFERENCE_global_data_node_t* data_node = (INTERFERENCE_global_data_node_t*) list_get(graph_measurement->result->data_nodes, i);
			if(clear_results)
				list_free_including_data(&data_node->local_data_list, free_local_data_list);
			if(data_node->node.address == graph_measurement->rollback_node_address)
				clear_results = true;
			if(data_node->node.address == graph_measurement->active_node)
				break;
			i++;
		}
		graph_measurement->rollback_necessary = false;
		graph_measurement->rollback_node_address = C_BROADCAST_ADDRESS;
	}

	// Request data from the next node.
	data_node = (INTERFERENCE_global_data_node_t*) list_find(graph_measurement->result->data_nodes, NULL, node_without_local_data_list);
	if(data_node != NULL) {
		graph_measurement->active_node = data_node->node.address;
		callback_add(measure_iterate_delay, graph_measurement, C_NETWORK_FLUSH_TIME);
	} else {
		// All nodes have measureed their graphs, we're done.
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_NOTE, "INTERFERENCE: Received data from all nodes.\n");
		graph_measurement->result->error_code = C_ERROR_OK;
		graph_measurement->function(graph_measurement->result);
		list_free_including_data(&graph_measurement->nodes, free_node_list);
		list_remove(C_INTERFERENCE_active_measurements, graph_measurement);
		FREE(graph_measurement);
	}
}

/**
 * @brief Callback for echo request of the measure function.
 *
 * @param parameter The results of the echo request.
 */
static void measure_echo_request_callback(callback_parameter_t parameter) {
	ICMP_echo_request_result_t* echo_request_result = (ICMP_echo_request_result_t*) parameter;
	INTERFERENCE_measurement_t* graph_measurement = (INTERFERENCE_measurement_t*) echo_request_result->parameter;
	node_t* node = NULL;
	INTERFERENCE_global_data_node_t* old_data_node = NULL;
	list_iterator_t node_iterator;

	switch(echo_request_result->error_code) {
	case C_ERROR_OK:
		// Add local node to the list of nodes.
		alloc_node(&node, C_LOCAL_ADDRESS);
		list_add(graph_measurement->nodes, node);
		list_sort(graph_measurement->nodes, node_address_less_than);

		// Print list of nodes for debugging.
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: %d node(s) in network: ", list_size(graph_measurement->nodes));
		list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), graph_measurement->nodes, print_node_address);
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");

		// Add all nodes to the list of data nodes.
		list_iterator_init(graph_measurement->nodes, &node_iterator);
		while((old_data_node = (INTERFERENCE_global_data_node_t*) list_iterator_next(&node_iterator)) != NULL) {
			INTERFERENCE_global_data_node_t* new_data_node = NULL;
			MALLOC(new_data_node, INTERFERENCE_global_data_node_t);
			memcpy(&new_data_node->node, old_data_node, sizeof(node_t));
			new_data_node->local_data_list = NULL;
			list_add(graph_measurement->result->data_nodes, new_data_node);
		}

		// Start collecting data from all nodes.
		measure_iterate(graph_measurement);

		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: No other nodes have been found.\n");
		graph_measurement->result->error_code = C_ERROR_TIMEOUT;
		graph_measurement->function(graph_measurement->result);
		list_free_including_data(&graph_measurement->nodes, free_node_list);
		list_remove(C_INTERFERENCE_active_measurements, graph_measurement);
		FREE(graph_measurement);
		break;
	default:
		C_ABORT_ON(true, "INTERFERENCE: Callback with unknown error code 0x%X.\n", echo_request_result->error_code);
	}
	FREE(echo_request_result);
}

/**
 * @brief Measure connectivity/interference graphs.
 *
 * @param function Pointer to function to call upon completion.
 * @param measurement_result Data structure for the resulting connectivity/interference graphs.
 */
void INTERFERENCE_measure(callback_function_t function, INTERFERENCE_measurement_result_t* measurement_result) {
	INTERFERENCE_measurement_t* graph_measurement = NULL;
	ICMP_echo_request_result_t* echo_request_result = NULL;

	if((function == NULL) || (measurement_result == NULL)) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Callback function or parameter not given (%p, %p).\n", function, measurement_result);
		return;
	} else if(!list_empty(C_INTERFERENCE_active_measurements)) {
		// Allow only one graph measurement to be active.
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Trying to measure connectivity/interference graphs, but a measurement is in progress...\n");
		measurement_result->error_code = C_ERROR_BUSY;
		function(measurement_result);
		return;
	}

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_NOTE, "INTERFERENCE: Starting to measure connectivity/interference graphs...\n");
	MALLOC(graph_measurement, INTERFERENCE_measurement_t);
	list_alloc(&graph_measurement->nodes);
	graph_measurement->active_node = C_BROADCAST_ADDRESS;
	graph_measurement->rollback_necessary = false;
	graph_measurement->rollback_node_address = C_BROADCAST_ADDRESS;
	graph_measurement->callback_id = 0;
	graph_measurement->function = function;
	graph_measurement->result = measurement_result;
	list_add(C_INTERFERENCE_active_measurements, graph_measurement);

	MALLOC(echo_request_result, ICMP_echo_request_result_t);
	echo_request_result->nodes = graph_measurement->nodes;
	echo_request_result->parameter = graph_measurement;
	ICMP_send_echo_request(C_BROADCAST_ADDRESS, C_DEFAULT_TTL, measure_echo_request_callback, echo_request_result);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Print the address of a global data node structure.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to data node whose address is to be printed.
 */
static void print_data_node_address(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "%s", addr_to_str(((INTERFERENCE_global_data_node_t*) item)->node.address));
}

/**
 * @brief Print the connectivity field of a local data node structure.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to data node whose connectivity is to be printed.
 */
static void print_data_node_connectivity(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "%.0f%%", (((float) ((INTERFERENCE_local_data_node_t*) item)->connectivity_count) / INTERFERENCE_CONNECTIVITY_SAMPLES * 100));
}

/**
 * @brief Print connectivity data in local data lists of a global node data structure.
 *
 * @param item Pointer to data node whose graph list is to be processed.
 */
static void print_connectivity_in_local_data_list(const void* item) {
	INTERFERENCE_global_data_node_t* data_node = (INTERFERENCE_global_data_node_t*) item;
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: %s:\t", addr_to_str(data_node->node.address));
	list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), data_node->local_data_list, print_data_node_connectivity);
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
}

/**
 * @brief Print the interference field of a local data node structure.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to data node whose interference is to be printed.
 */
static void print_data_node_interference(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "%.0f%%", (((INTERFERENCE_local_data_node_t*) item)->interference * 100));
}

/**
 * @brief Print interference data in local data lists of a global node data structure.
 *
 * @param item Pointer to data node whose graph list is to be processed.
 */
static void print_interference_in_local_data_list(const void* item) {
	INTERFERENCE_global_data_node_t* data_node = (INTERFERENCE_global_data_node_t*) item;
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: %s:\t", addr_to_str(data_node->node.address));
	list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), data_node->local_data_list, print_data_node_interference);
	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
}

/**
 * @brief Callback for ECHO REQUESTs of the ICMP command function.
 *
 * @param parameter The results of the echo request.
 */
static void INTERFERENCE_command_measure_callback(callback_parameter_t parameter) {
	INTERFERENCE_measurement_result_t* measurement_result = (INTERFERENCE_measurement_result_t*) parameter;

	switch(measurement_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_NOTE, "INTERFERENCE: Network graph measurement complete, %d node(s) in network.\n", list_size(measurement_result->data_nodes));
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Connectivity graph:\n");
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: \t");
		list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), measurement_result->data_nodes, print_data_node_address);
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		list_map(measurement_result->data_nodes, print_connectivity_in_local_data_list);
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Interference graph:\n");
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: \t");
		list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), measurement_result->data_nodes, print_data_node_address);
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		list_map(measurement_result->data_nodes, print_interference_in_local_data_list);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Network graph measurement in progress.\n");
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Network graph measurement aborted because no other nodes are reachable.\n");
		break;
	default:
		C_ABORT_ON(true, "INTERFERENCE: Callback with unknown error code 0x%X.\n", measurement_result->error_code);
	}

	list_free_including_data(&measurement_result->data_nodes, free_global_data_node_list);
	FREE(measurement_result);
}

/**
 * @brief INTERFERENCE command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool INTERFERENCE_command(int argc, const char*const* argv) {
	if(argc == 2) {
		if(strcmp(argv[1], "INTERFERENCE_measure") == 0) {
			INTERFERENCE_measurement_result_t* measurement_result = NULL;
			MALLOC(measurement_result, INTERFERENCE_measurement_result_t);
			list_alloc(&measurement_result->data_nodes);
			INTERFERENCE_measure(INTERFERENCE_command_measure_callback, measurement_result);
			return true;
		}
	}
	return false;
}
