/**
 * @file interference.h
 *
 * @brief Interference graphs (header).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interference.h 3177 2010-01-05 16:43:45Z wittenbu $
 */

#ifndef INTERFERENCE_H
#define INTERFERENCE_H

#include "../types.h"
#include "interference_types.h"



#define INTERFERENCE_PACKET_TYPE_REQUEST	0x0	///< INTERFERENCE REQUEST packet type.
#define INTERFERENCE_PACKET_TYPE_REPLY		0x1	///< INTERFERENCE REPLY packet type.
#define INTERFERENCE_PACKET_TYPE_RESET		0x2	///< INTERFERENCE RESET packet type.
#define INTERFERENCE_PACKET_TYPE_TEST_REQUEST	0x3	///< INTERFERENCE TEST REQUEST packet type.
#define INTERFERENCE_PACKET_TYPE_TEST_DATA	0x4	///< INTERFERENCE TEST DATA packet type.

/**
 * @brief Number of samples to collect to establish the connectivity between
 * two nodes.
 */
#define INTERFERENCE_CONNECTIVITY_SAMPLES		100

/**
 * @brief Number of samples to collect to establish the interference between
 * two nodes.
 */
#define INTERFERENCE_INTERFERENCE_SAMPLES		100

/**
 * @brief Time delay before the interfering node starts sending.
 */
#define INTERFERENCE_INTERFERENCE_TEST_DELAY		C_NETWORK_FLUSH_TIME

/**
 * @brief Time delay before the reference node starts sending.
 */
#define INTERFERENCE_INTERFERENCE_REFERENCE_DELAY	(3.0 * INTERFERENCE_INTERFERENCE_TEST_DELAY)

/**
 * @brief Time interval between interfering test packets.
 */
#define INTERFERENCE_INTERFERENCE_TEST_INTERVAL	0.01

/**
 * @brief Time interval between reference packets.
 */
#define INTERFERENCE_INTERFERENCE_REFERENCE_INTERVAL	0.1

/**
 * @brief Duration of a single interference test.
 */
#define INTERFERENCE_INTERFERENCE_TEST_DURATION	(3.0 * (INTERFERENCE_INTERFERENCE_REFERENCE_DELAY - INTERFERENCE_INTERFERENCE_TEST_DELAY) \
						+ INTERFERENCE_INTERFERENCE_SAMPLES * INTERFERENCE_INTERFERENCE_REFERENCE_INTERVAL)

/**
 * @brief Timeout time for an interference test.
 */
#define INTERFERENCE_INTERFERENCE_TEST_TIMEOUT	(INTERFERENCE_INTERFERENCE_TEST_DURATION + C_TIMEOUT)

/**
 * @brief Size of test packet payload in bytes.
 */
#define INTERFERENCE_INTERFERENCE_PAYLOAD_SIZE	C_max_net_payload_size()

/**
 * @brief Maximal number of nodes in the network.
 *
 * We need this value to estimate the timeout. Alternatively,
 * it could be established during a discovery phase.
 */
#define INTERFERENCE_MAX_NUMBER_OF_NODES		100

/**
 * @brief Time (in seconds) to wait for individual nodes to collect their data
 * and have it sent back to the central node.
 */
#define INTERFERENCE_PER_NODE_TIMEOUT		(INTERFERENCE_CONNECTIVITY_SAMPLES * C_SINGLE_HOP_RTT \
						+ INTERFERENCE_MAX_NUMBER_OF_NODES * INTERFERENCE_INTERFERENCE_TEST_TIMEOUT \
						+ 2 * C_TIMEOUT)

// Interface to main module.
void INTERFERENCE_init();
void INTERFERENCE_start();
void INTERFERENCE_stop();
void INTERFERENCE_finalize();
void INTERFERENCE_recv(const net_header_t* header, const void* payload);
bool INTERFERENCE_command(int argc, const char*const* argv);

// Interface provided by this module.
void INTERFERENCE_measure(callback_function_t function, INTERFERENCE_measurement_result_t* measurement_result);

#endif	// INTERFERENCE_H
