/**
 * @file main.c
 *
 * @brief Service Placement main file.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: main.c 3722 2010-08-23 13:40:10Z gwitten $
 */

/**
 * @mainpage
 *
 * The problem of service placement in ad hoc networks can be stated as follows:
 * Given a physical network topology and service demands of client nodes, adapt
 * the number and location of services in the network over time such that the
 * service demands are met at a minimal cost. The cost function may include
 * metrics such as network traffic, energy expenditure or service-dependant
 * Quality of Service (QoS) metrics (availability, latency, etc.).
 *
 * The purpose of this framework is to evaluate different approaches to service
 * placement in ad hoc networks. It is designed to run on different platforms
 * ranging from network simulators to major operating systems and embedded
 * devices. The framework is described in detail in:
 *
 *	Georg Wittenburg. Service Placement in Ad Hoc Networks. PhD thesis,
 *	Freie Universität Berlin, Berlin, Germany, 2010. (handed in for review)
 *
 * For more information, please have a look at the project webpage at
 * http://cst.mi.fu-berlin.de/projects/SPi/.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 */



#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef ARCH_LINUX
#include <errno.h>
#include <error.h>
#include <netdb.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#elif defined(ARCH_WIN32)
#include <winsock2.h>
#include <windows.h>
#endif	// ARCH_*

#include "config.h"
#include "macros.h"
#include "types.h"
#include "vars.h"
#include "main.h"
#include "util/debug.h"
#include "util/list.h"
#include "util/callback.h"
#include "util/net.h"
#include "icmp-light/icmp-light.h"
#include "service/service.h"
#include "placement/placement.h"



/**
 * @brief Architecture-specific abort macros.
 */
#ifdef ARCH_LINUX
	#define ABORT(errno)	error(1, errno, NULL)
#else
	#define ABORT(errno)	abort()
#endif	// ARCH_*



#ifdef ARCH_WIN32

/**
 * @brief Redefine errno to function of the Winsock API.
 */
#undef errno
#define errno WSAGetLastError()

/**
 * @brief Define error INTR error code of the Winsock API.
 */
#define EINTR WSAEINTR

#endif	// ARCH_WIN32



/**
 * @brief Status of the application.
 */
bool running;

/**
 * @brief File that contains commands for this application.
 */
FILE* command_file;

/**
 * @brief Size of buffer for current command in bytes.
 */
#define CURRENT_COMMAND_SIZE	256

/**
 * @brief The command read from the command file that is to be executed next.
 */
char current_command[CURRENT_COMMAND_SIZE];

/**
 * @brief Size of buffer for received packets in bytes.
 */
#define RECV_BUFFER_SIZE	C_MTU

/**
 * @brief Buffer for received packets in bytes.
 */
unsigned char recv_buffer[RECV_BUFFER_SIZE];

/**
 * @brief Time interval at which to insert a mark into logfile in seconds.
 */
#define LOG_MARK_INTERVAL	10

/**
 * @brief Time at which the last mark was inserted into logfile.
 */
C_time_t last_mark_in_log_time;

/**
 * @brief List of neighboring nodes according to the topology file (if set).
 */
list_t* topology_neighbors;



/**
 * @brief Initialization of the application.
 *
 * Seed the random number generator, setup and configure a socket, establish
 * the local address and initial
 */
static C_addr_t init() {
	unsigned int seed;
	struct protoent *udp = NULL;
	struct sockaddr_in addr;
	C_addr_t local_address = 0;
#ifdef ARCH_LINUX
	int opt = 1;
	int i = 1;
#elif defined(ARCH_WIN32)
	BOOL opt = TRUE;
	WSADATA wsa;
	int wsastartup_error;
#endif	// ARCH_*

	running = false;
	command_file = NULL;
	memset(current_command, 0, (sizeof(char) * CURRENT_COMMAND_SIZE));
	memset(recv_buffer, 0, (sizeof(char) * RECV_BUFFER_SIZE));
	last_mark_in_log_time = C_time();
	topology_neighbors = NULL;

#ifdef ARCH_WIN32
	// Start Windows Sockets API.
	wsastartup_error = WSAStartup(MAKEWORD(2,0), &wsa);
	if(wsastartup_error != 0) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while starting Windows Sockets API (%d)...\n", wsastartup_error);
		ABORT();
	}
#endif	// ARCH_WIN32

	// Seed random number generator.
	seed = (unsigned int) (C_time() * 1000000);
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Setting random seed to %d...\n", seed);
	srand(seed);

	// Setup socket.
	udp = getprotobyname("UDP");
	if(udp == NULL) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while retrieving protocol information for UDP (%d)...\n", errsv);
		ABORT(errsv);
	}
	if((C_udp_socket = socket(PF_INET, SOCK_DGRAM, udp->p_proto)) < 0) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while creating UDP socket (%d)...\n", errsv);
		ABORT(errsv);
	}
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(C_DEFAULT_UDP_PORT);
	if(bind(C_udp_socket, (struct sockaddr*) &addr, sizeof(struct sockaddr_in)) < 0) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while binding UDP socket on port %d (%d)...\n", C_DEFAULT_UDP_PORT, errsv);
		ABORT(errsv);
	}

	// Enable support for broadcasting.
#ifdef ARCH_LINUX
	if(setsockopt(C_udp_socket, SOL_SOCKET, SO_BROADCAST, &opt, sizeof(int)) < 0) {
#elif defined(ARCH_WIN32)
	if(setsockopt(C_udp_socket, SOL_SOCKET, SO_BROADCAST, (LPSTR) &opt, sizeof(BOOL)) < 0) {
#endif	// ARCH_*
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while setting socket options (%d)...\n", errsv);
		ABORT(errsv);
	}

	// Establish local IP address.
#ifdef ARCH_LINUX
	// Adapted from http://www.hungry.com/~alves/local-ip-in-C.html (no license).
	while(true) {
		struct ifreq ifr;
		struct sockaddr_in *sin = (struct sockaddr_in*) &ifr.ifr_addr;
		ifr.ifr_ifindex = i++;
		if(ioctl(C_udp_socket, SIOCGIFNAME, &ifr) < 0)
			break;
		if(
			(ioctl(C_udp_socket, SIOCGIFADDR, &ifr) < 0)
			|| (strncmp(inet_ntoa(sin->sin_addr), "127.0.0.1", 10) == 0)
			|| (i > 5)	// NOTE: This is an ugly hack to work around detecting IPs of additional interfaces.
		)
			continue;
		local_address = sin->sin_addr.s_addr;
		// printf("%d: %s\n", i, addr_to_str(local_address));
	}
	if(local_address == 0) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Cannot establish local IP address (%d).\n", errsv);
		ABORT(errsv);
	}
#elif defined(ARCH_WIN32)
	// Adapted from http://tangentsoft.net/wskfaq/examples/ipaddr.html (public domain).
	{
		char hostname[256];
		struct hostent *host = NULL;
		int i;

		if(gethostname(hostname, sizeof(hostname)) == SOCKET_ERROR) {
			int errsv = errno;
			C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Cannot establish local hostname (%d).\n", errsv);
			ABORT(errsv);
		}
		host = gethostbyname(hostname);
		if(host == NULL) {
			int errsv = errno;
			C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Cannot establish local host (%d).\n", errsv);
			ABORT(errsv);
		}
		for(i = 0; host->h_addr_list[i] != NULL; ++i) {
			struct in_addr addr;
			memcpy(&addr, host->h_addr_list[i], sizeof(struct in_addr));
			local_address = addr.s_addr;
		}
	}
#endif	// ARCH_*
	return local_address;
}

/**
 * @brief Handle incoming packets.
 *
 * Retrieve meta-information and hand packet up to components.
 *
 */
static void handle_packet() {
	struct sockaddr_in addr;
#ifdef ARCH_LINUX
	socklen_t addrlen = sizeof(struct sockaddr_in);
	ssize_t size = recvfrom(C_udp_socket, recv_buffer, RECV_BUFFER_SIZE, 0, (struct sockaddr*) &addr, &addrlen);
#elif defined(ARCH_WIN32)
	int addrlen = sizeof(struct sockaddr_in);
	int size = recvfrom(C_udp_socket, recv_buffer, RECV_BUFFER_SIZE, 0, (struct sockaddr*) &addr, &addrlen);
#endif	// ARCH_*
	net_header_t* header = NULL;
	void* payload = NULL;

	if(size < 0) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while receiving packet (%d).\n", errsv);
		ABORT(errsv);
	}
	C_ABORT_ON(size == 0, "Non-existant connection has been shutdown by peer (recvfrom() returned 0).\n");
	C_ABORT_ON(size < ((int) sizeof(net_header_t)), "Received %d byte(s) but expected at least %d header bytes.\n", size, sizeof(net_header_t));

	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Received a packet of size %d from %s:%d.\n", size, addr_to_str(addr.sin_addr.s_addr), ntohs(addr.sin_port));

	header = (net_header_t*) recv_buffer;

	C_ABORT_ON(addr.sin_addr.s_addr != header->last_hop_address, "OS packet header and internal packet header disagree about the last hop of this packet (%s != %s).\n", addr_to_str(addr.sin_addr.s_addr), addr_to_str(header->last_hop_address));

#ifdef C_TRANSMIT_PACKETS_VIA_BROADCAST
	if(
		(header->next_hop_address != C_BROADCAST_ADDRESS)
		&& (header->next_hop_address != C_LOCAL_ADDRESS)
	)
		return;
#endif	// C_TRANSMIT_PACKETS_VIA_BROADCAST

	// Only process packets that originate from a neighbor, if running on
	// an artificial topology.
	if(topology_neighbors != NULL) {
		C_addr_t* neighbor_address = NULL;
		list_iterator_t topology_neighbors_iterator;
		bool is_topology_neighbor = false;

		list_iterator_init(topology_neighbors, &topology_neighbors_iterator);
		while((neighbor_address = (C_addr_t*) list_iterator_next(&topology_neighbors_iterator)) != NULL) {
			if(*neighbor_address == header->last_hop_address) {
				is_topology_neighbor = true;
				break;
			}
		}
		if(!is_topology_neighbor)
			return;
	}

	C_ABORT_ON(size != ((int) (sizeof(net_header_t) + header->payload_length)), "Received %d byte(s) but expected at least %d (= %d + %d) bytes.\n", (sizeof(net_header_t) + header->payload_length), size, sizeof(net_header_t), header->payload_length);

#ifndef C_USE_INTERNAL_ROUTING
	// TODO: Gather information from the routing protocol (possibly via OS interfaces).
	header->hop_count = 1;
	header->last_hop_address = addr.sin_addr.s_addr;
#endif	// C_USE_INTERNAL_ROUTING

	if(header->payload_length > 0)
		payload = recv_buffer + sizeof(net_header_t);
	C_recv(header, payload);
}

/**
 * @brief Main loop iteration.
 *
 * Wait for either the next callback or for a packet to arrive.
 */
static void loop_iteration() {
	C_time_t time_until_next_mark = -1;
	struct timeval timeout;
	fd_set incoming_sockets;
	int select_return = 0;

	// Set timeout to the next time of marking the log file. As a side
	// effect, timeout is always set to an appropriate value even if there
	// is no callback.
	time_until_next_mark = last_mark_in_log_time + LOG_MARK_INTERVAL - C_time();
	if(time_until_next_mark < 0.0)
		time_until_next_mark = 0.0;
	memset(&timeout, 0, sizeof(struct timeval));
#ifdef ARCH_LINUX
	timeout.tv_sec = time_until_next_mark;
	timeout.tv_usec = (time_until_next_mark - ((C_time_t) timeout.tv_sec)) * 1000000;
#elif defined(ARCH_WIN32)
	timeout.tv_sec = (long) time_until_next_mark;
	timeout.tv_usec = (long) ((time_until_next_mark - ((C_time_t) timeout.tv_sec)) * 1000000);
#endif	// ARCH_*

	// Adjust timeout to next callback, if necessary.
	if(!list_empty(C_callbacks)) {
		C_time_t time_until_next_callback = ((callback_t*) list_head(C_callbacks))->time - C_time();
		if(time_until_next_callback <= time_until_next_mark) {
			if(time_until_next_callback < 0.0)
				time_until_next_callback = 0.0;
#ifdef ARCH_LINUX
			timeout.tv_sec = time_until_next_callback;
			timeout.tv_usec = (time_until_next_callback - ((C_time_t) timeout.tv_sec)) * 1000000;
#elif defined(ARCH_WIN32)
			timeout.tv_sec = (long) time_until_next_callback;
			timeout.tv_usec = (long) ((time_until_next_callback - ((C_time_t) timeout.tv_sec)) * 1000000);
#endif	// ARCH_*
		}
	}

	// Wait for either a packet to arrive or for the timeout to expire.
	FD_ZERO(&incoming_sockets);
	FD_SET(C_udp_socket, &incoming_sockets);
	select_return = select(C_udp_socket + 1, &incoming_sockets, NULL, NULL, &timeout);
	if(select_return < 0) {
		int errsv = errno;
		if(errsv != EINTR) {
			C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while waiting for packet (%d).\n", errsv);
			ABORT(errsv);
		}
	}
	C_ABORT_ON(select_return > 1, "Error while waiting for packet, more than one file descriptor was set (%d).\n", select_return);

	// Return if aborted.
	if(!running)
		return;

	// Handle packet.
	if(select_return == 1) {
		C_ABORT_ON(!FD_ISSET(C_udp_socket, &incoming_sockets), "Error while waiting for packet, file descriptor was not set.\n");
		handle_packet();
	}

	// Handle callback, if any.
	if(
		!list_empty(C_callbacks)
		&& (((callback_t*) list_head(C_callbacks))->time < C_time())
	) {
		callback_expire(NULL);
	}

	// Periodically insert mark into log file.
	if(C_time() >= (last_mark_in_log_time + LOG_MARK_INTERVAL)) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "-- MARK --\n");
		last_mark_in_log_time = C_time();
	}
}

/**
 * @brief Service ID for periodic service data requests.
 */
SERVICE_id_t* requests_service_id = NULL;

/**
 * @brief Service request offset for periodic service data requests.
 */
uint16_t requests_service_offset = 0;

/**
 * @brief Service request size for periodic service data requests.
 */
size_t requests_service_size = 0;

/**
 * @brief Service request interval for periodic service data requests in
 * seconds.
 */
C_time_t requests_service_interval = 0.0;

/**
 * @brief Callback ID for periodic service data requests.
 */
callback_id_t request_data_callback_id = 0;

/**
 * @brief Periodically request data from a service.
 *
 * @param parameter Not used.
 */
static void request_data_callback(callback_parameter_t parameter) {
	SERVICE_request_data(requests_service_id, requests_service_offset, requests_service_size);
	request_data_callback_id = callback_add(request_data_callback, NULL, requests_service_interval);
}

/**
 * @brief Process topology information for this node.
 *
 * @param topology_file_name File name of the topology file to process.
 */
static void process_topology(const char* topology_file_name) {
	FILE* topology_file = fopen(topology_file_name, "r");
	char topology_line[256];
	char* c = NULL;
	char* neighbor_address_list_str = NULL;
	char* neighbor_address_str = NULL;

	if(topology_file == NULL) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Error while opening topology file at \"%s\" (%d)...\n", topology_file_name, errsv);
		ABORT(errsv);
	}
	while(fgets(topology_line, 256, topology_file) != NULL) {
		for(c = topology_line; *c != '\0'; c++) {
			if(*c == ':') {
				*c = '\0';
				break;
			}
		}
		if(inet_addr(topology_line) == C_LOCAL_ADDRESS) {
			c++;
			break;
		}
		*c = '\0';	// Set c as indication for abort condition below.
	}
	C_ABORT_ON(*c == '\0', "Topology file \"%s\" does not contain information for this host.\n", topology_file_name);

	neighbor_address_list_str = c;
	for(c = neighbor_address_list_str; *c != '\0'; c++) {
		if(*c == '\n') {
			*c = '\0';
			break;
		}
	}

	list_alloc(&topology_neighbors);
	while((neighbor_address_str = strsep(&neighbor_address_list_str, ",")) != NULL) {
		C_addr_t* neighbor_address = NULL;
		C_ABORT_ON(strlen(neighbor_address_str) == 0, "Found empty string while parsing topology file \"%s\"...\n", topology_file_name);
		MALLOC(neighbor_address, C_addr_t);
		*neighbor_address = inet_addr(neighbor_address_str);
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_INFO, "Adding %s to topology neighbors...\n", addr_to_str(*neighbor_address));
		list_add(topology_neighbors, neighbor_address);
	}
	C_ABORT_ON(list_empty(topology_neighbors), "This node has no neighbors according to topology file \"%s\".\n", topology_file_name);
}

static void schedule_command(char* command_str);

/**
 * @brief Process a command.
 *
 * @param parameter Pointer to command string to process.
 */
static void process_command(callback_parameter_t parameter) {
	char* command_str = (char*) parameter;
	char* str = command_str;
	char* token = NULL;
	int argc = 1;
	const char* argv[16];
	argv[0] = NULL;	// Used by ns-2 for something else.

	while((token = strsep(&str, " ")) != NULL) {
		if(argc >= 16) {
			C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Failed to parse command \"%s\", too many parameters.\n", command_str);
			abort();
		}
		argv[argc++] = token;
	}
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "Processing command \"%s\" with %d parameter(s)...\n", argv[1], (argc - 2));
	if(argc == 2) {
		if(strcmp(argv[1], "start") == 0) {
			C_start();
			goto next_command;
		} else if(strcmp(argv[1], "stop") == 0) {
			C_stop();
			goto next_command;
		} else if(strcmp(argv[1], "finalize") == 0) {
			C_finalize();
			running = false;
			return;
		}
	} else if(argc == 3) {
		if(strcmp(argv[1], "set_topology") == 0) {
			process_topology(argv[2]);
			goto next_command;
		}
	}
	if(!C_command(argc, argv)) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_ERROR, "Invalid command \"%s\", aborting...\n", argv[1]);
		abort();
	}

next_command:
	if(fgets(current_command, CURRENT_COMMAND_SIZE, command_file) != NULL)
		schedule_command(current_command);
}

/**
 * @brief Schedule a command for processing.
 *
 * @param command_str_with_delay Pointer to command string to schedule (incl.
 * delay).
 */
static void schedule_command(char* command_str_with_delay) {
	char* command_str = command_str_with_delay;
	C_time_t command_delay = atof(command_str_with_delay) + C_init_time - C_time();
	if(command_delay < 0.0)
		command_delay = 0.0;

	// Remove trailing line break.
	while(*(++command_str) != '\n') {}
	*command_str = '\0';

	// Remove the delay from the string.
	command_str = command_str_with_delay;
	while(*(command_str++) != ' ') {}

	if(command_delay >= 10.0)
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_INFO, "Scheduling next command \"%s\" in %.1fs...\n", command_str, command_delay);
	else
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "Scheduling next command \"%s\" in %.1fs...\n", command_str, command_delay);
	callback_add(process_command, command_str, command_delay);
}

/**
 * @brief Process command line arguments.
 *
 * @param argc Number of command line arguments.
 * @param argv Array of command line arguments.
 * @param run Flag whether to take action based on the command line arguments
 * (as opposed to merely checking them for validity).
 */
static bool process_args(int argc, char *argv[], bool run) {
	int i;

	if(argc < 2)
		return false;

	for(i = 1; i < argc; i++) {
		if(strcmp(argv[i], "--srv") == 0) {
			if((i + 3) < argc) {
				if(run) {
					SERVICE_create_service(argv[i + 1], atoi(argv[i + 2]), atof(argv[i + 3]), C_PLACEMENT_ALGORITHM_NONE);
				}
				i += 3;
			} else {
				fprintf(stderr, "Not enough parameters for option \"--srv\" (need ID, SIZE, SYNC).\n");
				return false;
			}
		} else if(strcmp(argv[i], "--req") == 0) {
			if((i + 3) < argc) {
				if(run) {
					SERVICE_request_data(argv[i + 1], atoi(argv[i + 2]), atoi(argv[i + 3]));
				}
				i += 3;
			} else {
				fprintf(stderr, "Not enough parameters for option \"--req\" (need ID, OFFSET, SIZE).\n");
				return false;
			}
		} else if(strcmp(argv[i], "--reqs") == 0) {
			if((i + 4) < argc) {
				if(run) {
					SERVICE_alloc_service_id(&requests_service_id, argv[i + 1]);
					requests_service_offset = atoi(argv[i + 2]);
					requests_service_size = atoi(argv[i + 3]);
					requests_service_interval = (1.0 / atof(argv[i + 4]));
					request_data_callback_id = callback_add(request_data_callback, NULL, requests_service_interval);
				}
				i += 4;
			} else {
				fprintf(stderr, "Not enough parameters for option \"--reqs\" (need ID, OFFSET, SIZE, RATE).\n");
				return false;
			}
		} else if(strcmp(argv[i], "--passive") == 0) {
			// Nothing to do.
		} else if(strcmp(argv[i], "--nop") == 0) {
			if((i + 1) < argc) {
				if(run) {
					ICMP_send_nop(inet_addr(argv[i + 1]), C_DEFAULT_TTL, 0);
				}
				i += 1;
			} else {
				if(run) {
					ICMP_send_nop(C_BROADCAST_ADDRESS, C_DEFAULT_TTL, 0);
				}
			}
		} else if(strcmp(argv[i], "--file") == 0) {
			if((i + 1) < argc) {
				if(run) {
					command_file = fopen(argv[i + 1], "r");
					if(command_file == NULL) {
						int errsv = errno;
						C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_ERROR, "Error while opening command file at \"%s\" (%d)...\n", argv[i + 1], errsv);
						abort();
					}
					if(fgets(current_command, CURRENT_COMMAND_SIZE, command_file) != NULL) {
						schedule_command(current_command);
					} else {
						C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Command file \"%s\" does not contain any commands.\n", argv[i + 1]);
					}
				}
				i += 1;
			} else {
				fprintf(stderr, "Not enough parameters for option \"--file\" (need FILE).\n");
				return false;
			}
		} else if(strcmp(argv[i], "--topology") == 0) {
			if((i + 1) < argc) {
				if(run) {
					process_topology(argv[i + 1]);
				}
				i += 1;
			} else {
				fprintf(stderr, "Not enough parameters for option \"--topology\" (need FILE).\n");
				return false;
			}
		} else if(strcmp(argv[i], "--show-address") == 0) {
			if(run) {
				printf("Local address is %s, broadcast address is %s.\n", addr_to_str(C_LOCAL_ADDRESS), addr_to_str(C_BROADCAST_ADDRESS));
				exit(0);
			}
		} else if(strcmp(argv[i], "--help") == 0) {
			return false;
		} else {
			fprintf(stderr, "Unknown argument \"%s\".\n", argv[i]);
			return false;
		}
	}
	return true;
}

/**
 * @brief Print usage information.
 */
static void print_usage() {
	printf("\n");
	printf("%s (%s", SRVPLACE_NAME, SRVPLACE_VERSION);
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	printf("-SP");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
	printf("-SPI");
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION
	printf("-ASG");
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION
#ifdef C_ARTIFICIAL_LINK_RELIABILITY
	printf("-%dP_LINKS", (int) (C_ARTIFICIAL_LINK_RELIABILITY * 100.0));
#endif	// C_ARTIFICIAL_LINK_RELIABILITY
#ifdef C_TESTBED_EVALUATION
	printf("-TESTBED");
#endif	// C_TESTBED_EVALUATION
	printf(", %s %s)\n\n", __DATE__, __TIME__);
	printf("Usage: %s [OPTION], where OPTION may be any of the following:\n", SRVPLACE_NAME);
	printf("--srv ID SIZE SYNC\t\tstart service ID with SIZE bytes and\n");
	printf("\t\t\t\tsynchroniztion traffic ratio SYNC\n");
	printf("--req ID OFFSET SIZE\t\tsend service request for SIZE bytes at OFFSET\n");
	printf("\t\t\t\tof service ID\n");
	printf("--reqs ID OFFSET SIZE RATE\tsend service request for SIZE bytes at OFFSET\n");
	printf("\t\t\t\tof service ID at rate RATE\n");
	printf("--passive\t\t\trun passive node that takes no action\n");
	printf("--nop\t\t\t\tbroadcast an ICMP NOP (useful for debugging)\n");
	printf("--file FILE\t\t\tread commands from FILE\n");
	printf("--topology FILE\t\t\tread topology from FILE\n");
	printf("--show-address\t\t\tprint address information and exit\n");
	printf("--help\t\t\t\tdisplay this text and exit\n");
	printf("\n");
}

/**
 * @brief Stop the main loop.
 *
 * This function is called in response to a SIGINT signal.
 */
#ifdef ARCH_LINUX
static void handle_SIGINT() {
#elif defined(ARCH_WIN32)
static void handle_SIGINT(int UNUSED) {
#endif	// ARCH_*
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_INFO, "Interrupted by user.\n");

	if(request_data_callback_id != 0)
		callback_cancel(request_data_callback_id);

	C_check_consistency_on_shutdown = false;
	running = false;
}

/**
 * @brief The main function.
 *
 * Initialize and start all components, run the main loop and on shutdown stop
 * and finalize all components.
 *
 * @param argc Number of command line arguments.
 * @param argv Array of command line arguments
 */
int main(int argc, char *argv[]) {
	C_addr_t local_address = init();
	C_addr_t broadcast_address = INADDR_BROADCAST;
	char broadcast_address_str[16];
	struct in_addr tmp_addr;
	tmp_addr.s_addr = local_address;
        unsigned int n1, n2, n3, n4;

	// Set the broadcast address.
	if(sscanf(inet_ntoa(tmp_addr), "%u.%u.%u.%u", &n1, &n2, &n3, &n4) != 4) {
		int errsv = errno;
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Unable to parse local address (%d).\n", errsv);
		ABORT(errsv);
	}
	// TODO: Find a way to establish the correct broadcast address.
#ifdef C_TESTBED_EVALUATION
// 	snprintf(broadcast_address_str, 16, "%u.%u.%u.%u", n1, n2, n3, 127);	// Used for TESTBED
	snprintf(broadcast_address_str, 16, "%u.%u.%u.%u", n1, n2, 255, 255);	// Used for TESTBED
#else
	snprintf(broadcast_address_str, 16, "%u.%u.%u.%u", n1, n2, n3, 255);	// Used for POOL.
#endif	// C_TESTBED_EVALUATION
	inet_aton(broadcast_address_str, &tmp_addr);
	broadcast_address = tmp_addr.s_addr;

	signal(SIGINT, handle_SIGINT);

	// Check command line arguments.
	if(!process_args(argc, argv, 0)) {
		print_usage();
		exit(1);
	}

	// Initialize and start components.
	C_init(local_address, broadcast_address);

	// Process command line arguments.
	if(!process_args(argc, argv, 1)) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, "Failed to process supposedly valid command line arguments.\n");
		abort();
	}

	// Main loop.
	running = true;
	while(running)
		loop_iteration();

	if(requests_service_id != NULL)
		FREE(requests_service_id);
	if(command_file != NULL)
		fclose(command_file);
	if(topology_neighbors != NULL)
		list_free_including_data(&topology_neighbors, free);

	// Close socket.
#ifdef ARCH_LINUX
	close(C_udp_socket);
#elif defined(ARCH_WIN32)
	closesocket(C_udp_socket);
#endif	// ARCH_*

	return 0;
}
