/**
 * @file asg_simple_algorithm.c
 *
 * @brief ASG simple algorithm.
 *
 * This is an implementation of the "ASG simple" service placement algorithm
 * proposed in:
 * Klaus Herrmann. Self-Organizing Infrastructures for Ambient Services. PhD
 * Thesis. Berlin University of Technology, July 2006.
 *
 * It is applicable to a single service with multiple service instances. These
 * service instances are migrated and replicated (and dissolved) iteratively to
 * neighboring nodes according to a set of rules.
 *
 * The ASG architecture also describes a mechanism that optimizes the
 * distribution of information about active service instances. Instead of
 * re-implementing the entire mechanism, we decided to simply enable our own
 * optimization in the SLP component when evaluating this algorithm.
 *
 * Similarly, we enable our own optimizations for service migration and routing
 * as closely related concepts are mentioned by Herrmann.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: asg_simple_algorithm.c 3642 2010-07-07 22:10:27Z gwitten $
 */

#include <math.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/callback.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/net.h"
#include "../../slp-light/slp-light.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../placement_types.h"
#include "asg_simple_algorithm.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION

#ifndef C_ENABLE_ASG_OPTIMIZATIONS
	#warning Compiling the ASG simple algorithm without ASG optimizations...
#endif	// C_ENABLE_ASG_OPTIMIZATIONS



/**
 * @brief Debug level for this component.
 */
#define ASG_SIMPLE_DEBUG_LEVEL		C_MSG_NOTE



/**
 * @brief Interface to ASG simple algorithm.
 */
PLACEMENT_algorithm_t asg_simple_algorithm = {
	C_PLACEMENT_ALGORITHM_ASG_SIMPLE,
	asg_simple_algorithm_start,
	asg_simple_algorithm_stop,
	NULL,
	NULL,
	NULL,
	NULL,
	asg_simple_algorithm_report_known_service_instance,
	asg_simple_algorithm_report_invalid_service_instance,
	asg_simple_algorithm_get_synchronization_hosts
};

/**
 * @brief ASG simple algorithm data.
 */
typedef struct {
	SERVICE_id_t*	service_id;			///< Service ID.
	list_t*		known_service_instances;	///< List of addresses of known service instances.
	callback_id_t	callback_id;			///< ID of the periodic callback.
} asg_simple_algorithm_data_t;

/**
 * @brief Time interval between runs of the adaptation algorithm (in seconds).
 *
 * Herrmann gives the value of 5000 "ticks" for this parameter (p. 124).
 * However, earlier on he states that "there is no satisfactory mapping of
 * simulation time [ticks] to real time or vice versa." (p. 92).
 *
 * For our evaluation, we assume a value of 60s, which should provide a good
 * granularity for adaptation decisions in the range of request rates that we
 * are evaluating.
 */
#define ASG_SIMPLE_ADAPTATION_INTERVAL	60

/**
 * @brief Load threshold to justify the existence of a service instance.
 *
 * Herrmann does not give a value for this parameter, but only mentions that it
 * should be positive "to avoid that a single client that sends sporadic
 * requests can keep a replica alive" (p. 102, Replica Garbage Collection).
 *
 * Since the minimal request data rate used in our experiments is 102.4 B/s
 * (25.6 on the testbed), a value of 50 (B/s) (12 on the testbed) should match
 * the above criteria.
 */
#ifndef C_TESTBED_EVALUATION
	#define ASG_SIMPLE_ALPHA	50
#else
	#define ASG_SIMPLE_ALPHA	12
#endif	// C_TESTBED_EVALUATION



/**
 * @brief Implementation of the idle rule.
 *
 * @param service Pointer to the service for which to check whether a migration
 * is necessary.
 * @return <code>true</code> if this service instance is idle,
 * <code>false</code> otherwise.
 */
bool asg_simple_idle(SERVICE_service_t* service) {
	// NOTE: Since we're using data rate stats, there's no need to worry
	// about stability.
	return (calculate_datarate(service->host_stat->datarate_stat) < ASG_SIMPLE_ALPHA);
}

/**
 * @brief Implementation of the replication rule.
 *
 * @param service Pointer to the service for which to check whether a migration
 * is necessary.
 * @param replication_target_address_1 Pointer to the address of the first
 * replication target node (return value).
 * @param replication_target_address_2 Pointer to the address of the second
 * replication target node (return value).
 */
void asg_simple_get_replication_targets(SERVICE_service_t* service, C_addr_t* replication_target_address_1, C_addr_t* replication_target_address_2) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	unsigned int number_of_neighbors = 0;
	float highest_average_distance = 0.0;	// Uses hop count as distance metric.
	uint32_t highest_message_flow = 0;	// Uses data rate as metric for message flows.
	float rho = 0.0;			// Replication radius.
	uint8_t maximal_node_distance_in_hops = 0;
	list_t* neighbor_nodes_with_highest_average_distance = NULL;
	list_t* neighbor_nodes_with_highest_message_flow = NULL;

	*replication_target_address_1 = C_BROADCAST_ADDRESS;
	*replication_target_address_2 = C_BROADCAST_ADDRESS;

	// Check whether we have more than 2 neighbors.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		if(node_stat->node->minimal_hop_count <= 1)
			number_of_neighbors++;
	}
	if(number_of_neighbors < 2)
		return;

	// Calculate highest average distance, highest message flow, and
	// replication radius (rho).
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		float average_hop_count = 0.0;
		uint32_t datarate = calculate_datarate(node_stat->last_hop_datarate_stat);

		if(node_stat->last_hop_request_count > 0)
			average_hop_count = ((float) node_stat->last_hop_hop_count_running_sum) / ((float) node_stat->last_hop_request_count);
		if(average_hop_count > highest_average_distance)
			highest_average_distance = average_hop_count;

		if(datarate > highest_message_flow)
			highest_message_flow = datarate;

		if(
			(node_stat->node->minimal_hop_count <= C_MAXIMAL_NETWORK_DIAMETER)
			&& (node_stat->node->minimal_hop_count > maximal_node_distance_in_hops)
		) {
			maximal_node_distance_in_hops = node_stat->node->minimal_hop_count;
		}
	}
	// Herrmann states that (rho ^ 0.5) should be proportional to the
	// network diameter (p. 126). The network diameter is unknown, but
	// (1.5 * maximal_node_distance_in_hops) should be a good approximation.
	rho = sqrt(1.5 * ((float) maximal_node_distance_in_hops));
	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Replication radius for service \"%s\" is %.2f (maximal node distance is %d hop(s)).\n", service->id, rho, maximal_node_distance_in_hops);

	// Calculate set of neighbors with highest average distance and set of
	// neighbors with highest message flow.
	list_alloc(&neighbor_nodes_with_highest_average_distance);
	list_alloc(&neighbor_nodes_with_highest_message_flow);
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		float average_hop_count = 0.0;
		uint32_t datarate = calculate_datarate(node_stat->last_hop_datarate_stat);

		if(node_stat->last_hop_request_count > 0)
			average_hop_count = ((float) node_stat->last_hop_hop_count_running_sum) / ((float) node_stat->last_hop_request_count);
		if(average_hop_count >= highest_average_distance) {
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_VERBOSE, "ASG/SIMPLE: Adding %s to set of neighbors with highest average distance (%.2f hops)...\n", addr_to_str(node_stat->node->address), average_hop_count);
			list_add(neighbor_nodes_with_highest_average_distance, node_stat);
		}

		if(datarate >= highest_message_flow) {
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_VERBOSE, "ASG/SIMPLE: Adding %s to set of neighbors with highest message flow (%d B/s)...\n", addr_to_str(node_stat->node->address), datarate);
			list_add(neighbor_nodes_with_highest_message_flow, node_stat);
		}
	}

	// Check replication condition.
	list_iterator_init(neighbor_nodes_with_highest_average_distance, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		float average_hop_count = 0.0;
		SERVICE_node_stat_t cmp_node_stat;
		node_t cmp_node;
		uint32_t second_highest_message_flow = 0;	// Uses data rate as metric for message flows.
		SERVICE_node_stat_t* node_with_second_highest_message_flow = NULL;
		SERVICE_node_stat_t* node_stat2 = NULL;
		list_iterator_t node_stats_iterator2;
		uint32_t total_datarate_via_other_last_hops = 0;

		// Calculate average hop count.
		if(node_stat->last_hop_request_count > 0)
			average_hop_count = ((float) node_stat->last_hop_hop_count_running_sum) / ((float) node_stat->last_hop_request_count);

		// First check: Does the traffic forwarded by this neighbor
		// come from nodes that are far enough away?
		if(average_hop_count <= rho) {
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Average hop count of traffic via replication candidate %s is smaller than the replication radius (%.2f <= %.2f), skipping candidate...\n", addr_to_str(node_stat->node->address), average_hop_count, rho);
			continue;
		}

		memset(&cmp_node_stat, 0, sizeof(SERVICE_node_stat_t));
		memset(&cmp_node, 0, sizeof(node_t));
		cmp_node_stat.node = &cmp_node;
		cmp_node_stat.node->address = node_stat->node->address;

		// Second check: Is the neighbor contained in the set
		// of neighbors with the highest message flow?
		if(list_find(neighbor_nodes_with_highest_message_flow, &cmp_node_stat, SERVICE_matches_node_stat_address) == NULL) {
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Replication candidate %s is not contained in set of neighbors with highest message flow (%d element(s)), skipping candidate...\n", addr_to_str(node_stat->node->address), list_size(neighbor_nodes_with_highest_message_flow));
			continue;
		}

		// Set the first replication target.
		*replication_target_address_1 = node_stat->node->address;

		// Select node with the highest message flow other than the
		// first replication target.
		list_iterator_init(service->node_stats, &node_stats_iterator2);
		while((node_stat2 = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator2)) != NULL) {
			uint32_t datarate = calculate_datarate(node_stat2->last_hop_datarate_stat);
			if(
				(node_stat2->node->address != *replication_target_address_1)
				&& (datarate > second_highest_message_flow)
			) {
				second_highest_message_flow = datarate;
				node_with_second_highest_message_flow = node_stat2;
			}
		}

		// Calculate the data flow forwarded via all other hops (except
		// the first replication target and the other node with the
		// highest replication target).
		list_iterator_init(service->node_stats, &node_stats_iterator2);
		while((node_stat2 = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator2)) != NULL) {
			if(
				(node_stat2->node->address != *replication_target_address_1)
				&& (node_stat2 != node_with_second_highest_message_flow)
			) {
				total_datarate_via_other_last_hops += calculate_datarate(node_stat2->last_hop_datarate_stat);
			}
		}

		// Check whether the other node does NOT forward enough
		// messages.
		// NOTE: This is done implicitly by the stability check (Fig.
		// 8.9, line 11), which we otherwise ignore since our data rate
		// statistics are stable enough.
		if(!(second_highest_message_flow > total_datarate_via_other_last_hops)) {
			*replication_target_address_2 = C_LOCAL_ADDRESS;
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Replicating one instance of service \"%s\" to %s (%d service demand, with %.2f average request distance, after %d requests, replication radius is %.2f)...\n", service->id, addr_to_str(*replication_target_address_1), calculate_datarate(node_stat->last_hop_datarate_stat), average_hop_count, service->host_stat->request_count, rho);
		} else {
			*replication_target_address_2 = node_with_second_highest_message_flow->node->address;
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Replicating two instances of service \"%s\" to %s and %s (%d service demand, with %.2f average request distance, and %d service demand after %d requests, replication radius is %.2f)...\n", service->id, addr_to_str(*replication_target_address_1), addr_to_str(*replication_target_address_2), calculate_datarate(node_stat->last_hop_datarate_stat), average_hop_count, calculate_datarate(node_with_second_highest_message_flow->last_hop_datarate_stat), service->host_stat->request_count, rho);
		}

		break;	// Returns (replication_target_address_1, replication_target_address_2).
	}

	list_free(&neighbor_nodes_with_highest_average_distance);
	list_free(&neighbor_nodes_with_highest_message_flow);
}

/**
 * @brief Implementation of the migration rule.
 *
 * NOTE: This function only implements the "simple migration condition" (8.27)
 * explicitly. However, since its decision is based on the data rate
 * calculation, it is stable enough and implementing the "advanced migration
 * condition" (8.31) is not necessary.
 *
 * @param service Pointer to the service for which to check whether a migration
 * is necessary.
 * @param neighbor_address_to_ignore Address of a neighboring node which is to
 * be ignored while establishing a migration target, or C_BROADCAST_ADDRESS if
 * all nodes are to be considered. (Used by ASG/Event Flow Tree).
 * @return Address of the migration target node, or C_BROADCAST_ADDRESS if no
 * migration is necessary.
 */
C_addr_t asg_simple_get_migration_target(SERVICE_service_t* service, C_addr_t neighbor_address_to_ignore) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	uint32_t max_last_hop_datarate = 0;
	node_t* node_with_max_last_hop_datarate = NULL;
	uint32_t total_datarate_via_other_last_hops = 0;

	// Look for node that was the last hop to forward most of the service
	// demand to this node.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		uint32_t datarate = 0;

		if(neighbor_address_to_ignore == node_stat->node->address)
			continue;

		datarate = calculate_datarate(node_stat->last_hop_datarate_stat);
		if(datarate > max_last_hop_datarate) {
			max_last_hop_datarate = datarate;
			node_with_max_last_hop_datarate = node_stat->node;
		}
	}
	if(node_with_max_last_hop_datarate == NULL) {
		C_ABORT_ON(neighbor_address_to_ignore == C_BROADCAST_ADDRESS, "ASG/SIMPLE: %d service request(s) for service \"%s\" but no service demand.\n", service->host_stat->request_count, service->id);
		return C_BROADCAST_ADDRESS;
	}

	// Calculate the sum of the service demand forwarded to this node
	// via all other neighbors.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		if(neighbor_address_to_ignore == node_stat->node->address)
			continue;

		if(node_stat->node != node_with_max_last_hop_datarate) {
			total_datarate_via_other_last_hops += calculate_datarate(node_stat->last_hop_datarate_stat);
		}
	}

	// Migrate to the last hop that forwarded most of the service demand if
	// this service demand is greater than the total service demand via all
	// other neighbors.
	if(node_with_max_last_hop_datarate->address == C_LOCAL_ADDRESS) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Instance of service \"%s\" is placed optimally on the current node (%d service demand after %d requests, %d service demand from all other nodes).\n", service->id, max_last_hop_datarate, service->host_stat->request_count, total_datarate_via_other_last_hops);
		return C_BROADCAST_ADDRESS;
	} else if(max_last_hop_datarate > total_datarate_via_other_last_hops) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Migrating instance of service \"%s\" to %s (%d service demand after %d requests, %d service demand from all other nodes)...\n", service->id, addr_to_str(node_with_max_last_hop_datarate->address), max_last_hop_datarate, service->host_stat->request_count, total_datarate_via_other_last_hops);
		return node_with_max_last_hop_datarate->address;
	} else {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Migration of instance of service \"%s\" is not necessary (%d service demand after %d requests, %d service demand from all other nodes).\n", service->id, max_last_hop_datarate, service->host_stat->request_count, total_datarate_via_other_last_hops);
		return C_BROADCAST_ADDRESS;
	}
}

static void asg_simple_algorithm_run(callback_parameter_t parameter);

/**
 * @brief Replication callback data.
 */
typedef struct {
	SERVICE_id_t	service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_addr_t	replication_target_address_2;	///< Address of the second replication target.
} asg_simple_replication_data_t;

/**
 * @brief Second service replication callback.
 *
 * @param parameter Replication result.
 */
static void asg_simple_replication_2_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* replication_result = (SERVICE_replication_result_t*) parameter;
	asg_simple_replication_data_t* replication_data = (asg_simple_replication_data_t*) replication_result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(replication_data->service_id);
	asg_simple_algorithm_data_t* placement_data = NULL;

	C_ABORT_ON(service == NULL, "ASG/SIMPLE: Trying to call second replication callback of ASG simple algorithm for service \"%s\", but service is NULL.\n", replication_data->service_id);
	if(service->placement_data == NULL) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Second replication of service \"%s\" failed (%d), but ASG simple algorithm is not active anymore.\n", replication_data->service_id, replication_result->error_code);
		goto free_replication_2_parameters;
	}
	placement_data = (asg_simple_algorithm_data_t*) service->placement_data;

	switch(replication_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Second replication of service \"%s\" was successful. Shutting down service instance on this node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		SERVICE_shutdown_service(replication_data->service_id);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Second replication of service \"%s\" failed because the service was found running on the destination node. Shutting down service instance on this node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		SERVICE_shutdown_service(replication_data->service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Second replication of service \"%s\" timed out, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Second replication of service \"%s\" has been aborted, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		// Don't re-add the callback if aborted by the user.
		break;
	case C_ERROR_BUSY:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Second replication of service \"%s\" failed because the local system is busy, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Second replication of service \"%s\" is already in progress, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Second replication of service \"%s\" failed due to a communication error, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Second replication of service \"%s\" failed because the remote system refused to host the service, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	default:
		C_ABORT_ON(true, "ASG/SIMPLE: Second replication callback with unknown error code 0x%X.\n", replication_result->error_code);
	}

free_replication_2_parameters:
	FREE(replication_data);
	FREE(replication_result);
}

/**
 * @brief First service replication callback.
 *
 * @param parameter Replication result.
 */
static void asg_simple_replication_1_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* replication_result = (SERVICE_replication_result_t*) parameter;
	asg_simple_replication_data_t* replication_data = (asg_simple_replication_data_t*) replication_result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(replication_data->service_id);
	asg_simple_algorithm_data_t* placement_data = NULL;

	C_ABORT_ON(service == NULL, "ASG/SIMPLE: Trying to call first replication callback of ASG simple algorithm for service \"%s\", but service is NULL.\n", replication_data->service_id);
	if(service->placement_data == NULL) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: First replication of service \"%s\" failed (%d), but ASG simple algorithm is not active anymore.\n", replication_data->service_id, replication_result->error_code);
		goto free_replication_1_parameters;
	}
	placement_data = (asg_simple_algorithm_data_t*) service->placement_data;

	switch(replication_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: First replication of service \"%s\" was successful.\n", replication_data->service_id);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: First replication of service \"%s\" failed because the service was found running on the destination node.\n", replication_data->service_id);
		SERVICE_reset_service_stats(replication_data->service_id);	// NOTE: This indicates that the stats that lead to this replication decision were wrong. Reset stats to take this into account.
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: First replication of service \"%s\" timed out.\n", replication_data->service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: First replication of service \"%s\" has been aborted.\n", replication_data->service_id);
		// Don't consider second replication if aborted by the user.
		goto free_replication_1_parameters;
	case C_ERROR_BUSY:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: First replication of service \"%s\" failed because the local system is busy.\n", replication_data->service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: First replication of service \"%s\" is already in progress.\n", replication_data->service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: First replication of service \"%s\" failed due to a communication error.\n", replication_data->service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: First replication of service \"%s\" failed because the remote system refused to host the service.\n", replication_data->service_id);
		break;
	default:
		C_ABORT_ON(true, "ASG/SIMPLE: First replication callback with unknown error code 0x%X.\n", replication_result->error_code);
	}

	if(replication_data->replication_target_address_2 != C_LOCAL_ADDRESS) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Starting second replication of service \"%s\" to %s...\n", replication_data->service_id, addr_to_str(replication_data->replication_target_address_2));
		SERVICE_replicate_service(replication_data->service_id, replication_data->replication_target_address_2, asg_simple_replication_2_callback, replication_result);
		return;		// Don't free parameters.
	} else {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Second replication of service \"%s\" is not necessary, service instance remains on the current node.\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		// The service demand has changed since this node now has a
		// neighboring service instance. Reset stats to take this into
		// account.
		SERVICE_reset_service_stats(replication_data->service_id);
		// Make sure the new service instance knows about this one for
		// synchronization.
		SLP_announce_service(replication_data->service_id, C_BROADCAST_ADDRESS);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
	}

free_replication_1_parameters:
	FREE(replication_data);
	FREE(replication_result);
}

/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void asg_simple_migration_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* migration_result = (SERVICE_migration_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) migration_result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(service_id);
	asg_simple_algorithm_data_t* placement_data = NULL;

	if(migration_result->error_code != C_ERROR_OK) {
		C_ABORT_ON(service == NULL, "ASG/SIMPLE: Trying to call migration callback of ASG simple algorithm for service \"%s\", but service is NULL.\n", service_id);
		if(service->placement_data == NULL) {
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Migration of service \"%s\" failed (%d), but ASG simple algorithm is not active anymore.\n", service_id, migration_result->error_code);
			goto free_migration_parameters;
		}
		placement_data = (asg_simple_algorithm_data_t*) service->placement_data;
	}

	switch(migration_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Migration of service \"%s\" was successful.\n", service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", service_id);
		SERVICE_unlock_service(service_id);
		SERVICE_shutdown_service(service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Migration of service \"%s\" timed out.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Migration of service \"%s\" has been aborted.\n", service_id);
		SERVICE_unlock_service(service_id);
		// Don't re-add the callback if aborted by the user.
		break;
	case C_ERROR_BUSY:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Migration of service \"%s\" failed because the local system is busy.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Migration of service \"%s\" is already in progress.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Migration of service \"%s\" failed due to a communication error.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Migration of service \"%s\" failed because the remote system refused to host the service.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		break;
	default:
		C_ABORT_ON(true, "ASG/SIMPLE: Migration callback with unknown error code 0x%X.\n", migration_result->error_code);
	}

free_migration_parameters:
	FREE(service_id);
	FREE(migration_result);
}

/**
 * @brief ASG simple algorithm.
 *
 * @param parameter Pointer to service for which to run this algorithm.
 */
static void asg_simple_algorithm_run(callback_parameter_t parameter) {
	SERVICE_id_t* service_id = (SERVICE_id_t*) parameter;
	SERVICE_service_t* service = SERVICE_get_running_service(service_id);
	asg_simple_algorithm_data_t* placement_data = NULL;

	if(service == NULL) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/SIMPLE: Trying to run ASG simple algorithm for service \"%s\", but service is not running.\n", service->id);
		return;
	}

	C_ABORT_ON(service->placement_data == NULL, "ASG/SIMPLE: Trying to run ASG simple algorithm for service \"%s\", but placement data is NULL.\n", service->id);
	placement_data = (asg_simple_algorithm_data_t*) service->placement_data;
	placement_data->callback_id = 0;

	// Make sure all service instances know about this instance for
	// synchronization.
	SLP_announce_service(service->id, C_BROADCAST_ADDRESS);

	if(service->host_stat->request_count == 0) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_WARN, "ASG/SIMPLE: Trying to run ASG simple algorithm for service \"%s\", but no service requests have been received.\n", service->id);
		placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		return;
	}

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Checking whether an adaptation is required for service \"%s\"...\n", service->id);

	/// This following corresponds to the code snippet show in Figure 8.5
	/// (page 101) of Herrmann.

	if(asg_simple_idle(service)) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Service \"%s\" is idle, shutting down service...\n", service->id);
		SERVICE_shutdown_service(service->id);
	} else {
		C_addr_t replication_target_address_1 = C_BROADCAST_ADDRESS;
		C_addr_t replication_target_address_2 = C_BROADCAST_ADDRESS;
		C_addr_t migration_target_address = C_BROADCAST_ADDRESS;

		asg_simple_get_replication_targets(service, &replication_target_address_1, &replication_target_address_2);
		if(
			(replication_target_address_1 != C_BROADCAST_ADDRESS)
			&& (replication_target_address_2 != C_BROADCAST_ADDRESS)
		) {
			SERVICE_replication_result_t* replication_result = NULL;
			asg_simple_replication_data_t* replication_data = NULL;

			if(replication_target_address_2 == C_LOCAL_ADDRESS)
				C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Replicating service \"%s\" to %s...\n", service->id, addr_to_str(replication_target_address_1));
			else
				C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Replicating service \"%s\" to %s and %s...\n", service->id, addr_to_str(replication_target_address_1), addr_to_str(replication_target_address_2));
			SERVICE_lock_service(service->id);
			MALLOC(replication_result, SERVICE_replication_result_t);
			MALLOC(replication_data, asg_simple_replication_data_t);
			SERVICE_copy_service_id(replication_data->service_id, service->id);
			replication_data->replication_target_address_2 = replication_target_address_2;
			replication_result->parameter = replication_data;
			SERVICE_replicate_service(service->id, replication_target_address_1, asg_simple_replication_1_callback, replication_result);
		} else if((migration_target_address = asg_simple_get_migration_target(service, C_BROADCAST_ADDRESS)) != C_BROADCAST_ADDRESS) {
			SERVICE_migration_result_t* migration_result = NULL;

			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Migrating service \"%s\" to %s...\n", service->id, addr_to_str(migration_target_address));
			SERVICE_lock_service(service->id);
			MALLOC(migration_result, SERVICE_migration_result_t);
			SERVICE_alloc_service_id((SERVICE_id_t**) &migration_result->parameter, service->id);
			SERVICE_migrate_service(service->id, migration_target_address, asg_simple_migration_callback, migration_result);
		} else {
			C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/SIMPLE: Nothing to do for service \"%s\"...\n", service->id);
			placement_data->callback_id = callback_add(asg_simple_algorithm_run, service, ASG_SIMPLE_ADAPTATION_INTERVAL);
		}
	}
}

/**
 * @brief Start ASG simple algorithm.
 *
 * @param service Pointer to service for which to start this algorithm.
 */
void asg_simple_algorithm_start(SERVICE_service_t* service) {
	asg_simple_algorithm_data_t* placement_data = NULL;
	C_addr_t* address = NULL;

	C_ABORT_ON(service->placement_data != NULL, "ASG/SIMPLE: Trying to start ASG simple algorithm for service \"%s\", but placement data is not NULL.\n", service->id);

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Starting ASG simple algorithm for service \"%s\"...\n", service->id);
	MALLOC(service->placement_data, asg_simple_algorithm_data_t);
	placement_data = (asg_simple_algorithm_data_t*) service->placement_data;
	SERVICE_alloc_service_id(&placement_data->service_id, service->id);
	list_alloc(&placement_data->known_service_instances);
	MALLOC(address, C_addr_t);
	*address = C_LOCAL_ADDRESS;
	list_add(placement_data->known_service_instances, address);
	placement_data->callback_id = callback_add(asg_simple_algorithm_run, placement_data->service_id, ASG_SIMPLE_ADAPTATION_INTERVAL);
}

/**
 * @brief Stop ASG simple algorithm.
 *
 * @param service Pointer to service for which to stop this algorithm.
 */
void asg_simple_algorithm_stop(SERVICE_service_t* service) {
	asg_simple_algorithm_data_t* placement_data = (asg_simple_algorithm_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "ASG/SIMPLE: Trying to stop ASG simple algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Stopping ASG simple algorithm for service \"%s\"...\n", service->id);
	FREE(placement_data->service_id);
	list_free_including_data(&placement_data->known_service_instances, free);
	if(placement_data->callback_id != 0)
		callback_cancel(placement_data->callback_id);
	FREE(service->placement_data);
}

/**
 * @brief Report known service instance to ASG simple algorithm.
 *
 * @param service Pointer to service for which report a known service instance.
 * @param host_address Address of the host of the known service instance.
 */
void asg_simple_algorithm_report_known_service_instance(SERVICE_service_t* service, C_addr_t host_address) {
	asg_simple_algorithm_data_t* placement_data = (asg_simple_algorithm_data_t*) service->placement_data;
	C_addr_t* address = NULL;
	C_addr_t cmp_address;

	C_ABORT_ON(placement_data == NULL, "ASG/SIMPLE: Trying to report known service instance of service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Handling report of known service instance of service \"%s\" at %s...\n", service->id, addr_to_str(host_address));

	cmp_address = host_address;
	if((address = (C_addr_t*) list_find(placement_data->known_service_instances, &cmp_address, matches_address)) == NULL) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Adding %s to list of known service instances of service \"%s\"...\n", addr_to_str(host_address), service->id);
		MALLOC(address, C_addr_t);
		*address = host_address;
		list_add(placement_data->known_service_instances, address);
	}
}

/**
 * @brief Report invalid service instance to ASG simple algorithm.
 *
 * @param service Pointer to service for which report a known service instance.
 * @param host_address Address of the host of the invalid service instance.
 */
void asg_simple_algorithm_report_invalid_service_instance(SERVICE_service_t* service, C_addr_t host_address) {
	asg_simple_algorithm_data_t* placement_data = (asg_simple_algorithm_data_t*) service->placement_data;
	C_addr_t* address = NULL;
	C_addr_t cmp_address;

	C_ABORT_ON(placement_data == NULL, "ASG/SIMPLE: Trying to report invalid service instance of service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Handling report of invalid service instance of service \"%s\" at %s...\n", service->id, addr_to_str(host_address));

	cmp_address = host_address;
	if((address = (C_addr_t*) list_find(placement_data->known_service_instances, &cmp_address, matches_address)) != NULL) {
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_INFO, "ASG/SIMPLE: Removing %s from list of known service instances of service \"%s\"...\n", addr_to_str(host_address), service->id);
		list_remove(placement_data->known_service_instances, address);
		FREE(address);
	}
}

/**
 * @brief Get synchronization hosts from ASG simple algorithm.
 *
 * The synchronization host in the ASG context is called the "primary".
 * Herrmann states that "[t]he initial replica of a specific service type that
 * is installed in an ASG automatically becomes the primary" (p. 205). It is
 * however unclear how this primary service instance is identified if it
 * migrates, or replicates and shuts down afterwards. Hence, we just use the
 * instance with the smallest address as primary.
 *
 * @param service Pointer to service for which to get the synchronization hosts.
 * @param inter_host_update Flag whether this update is an inter-host update.
 * @return Pointer to a list of nodes to which synchronize changes to (must be
 * freed by the caller), or NULL if no synchronization is necessary.
 */
list_t* asg_simple_algorithm_get_synchronization_hosts(const SERVICE_service_t* service, bool inter_host_update) {
	asg_simple_algorithm_data_t* placement_data = (asg_simple_algorithm_data_t*) service->placement_data;
	list_t* synchronization_hosts = NULL;
	C_addr_t primary_address = C_BROADCAST_ADDRESS;
	C_addr_t* instance_address = NULL;
	list_iterator_t instances_iterator;

	C_ABORT_ON(placement_data == NULL, "ASG/SIMPLE: Trying to get synchronization hosts for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Getting synchronization hosts for service \"%s\"...\n", service->id);
	list_iterator_init(placement_data->known_service_instances, &instances_iterator);
	while((instance_address = (C_addr_t*) list_iterator_next(&instances_iterator)) != NULL) {
		if(
			(primary_address == C_BROADCAST_ADDRESS)
			|| (*instance_address < primary_address)
		) {
			primary_address = *instance_address;
		}
	}
	C_ABORT_ON(primary_address == C_BROADCAST_ADDRESS, "ASG/SIMPLE: Failed to establish \"primary\" for service \"%s\"...", service->id);

	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Using %s as \"primary\" for service \"%s\"...\n", addr_to_str(primary_address), service->id);

	list_alloc(&synchronization_hosts);
	if(primary_address == C_LOCAL_ADDRESS) {
		list_iterator_init(placement_data->known_service_instances, &instances_iterator);
		while((instance_address = (C_addr_t*) list_iterator_next(&instances_iterator)) != NULL) {
			if(*instance_address != primary_address)
				list_add(synchronization_hosts, instance_address);
		}
	} else if(!inter_host_update) {
		// Only synchronize regular updates on non-primary hosts.
		list_iterator_init(placement_data->known_service_instances, &instances_iterator);
		while((instance_address = (C_addr_t*) list_iterator_next(&instances_iterator)) != NULL) {
			if(*instance_address == primary_address) {
				// Using the item from the list avoids having
				// to allocate memory (which would be hard to
				// free later on).
				list_add(synchronization_hosts, instance_address);
				break;
			}
		}
	}
	list_sort(synchronization_hosts, node_address_less_than);

	if(inter_host_update)
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Synchronization hosts for inter-host update of service \"%s\" are ", service->id);
	else
		C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/SIMPLE: Synchronization hosts for update of service \"%s\" are ", service->id);
	list_print(ASG_SIMPLE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), synchronization_hosts, print_node_address);
	C_DEBUG(ASG_SIMPLE_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ".\n");

	return synchronization_hosts;
}

#else
	#warning Compiling without support for ASG simple algorithm...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
