/**
 * @file gcmi_cost_calculation.c
 *
 * @brief Graph cost multiple instances (cost calculation).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: gcmi_cost_calculation.c 3426 2010-03-18 15:00:10Z gwitten $
 */

#include <float.h>
#include <math.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/mapping.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "graph_cost_multiple_instances.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Debug level for this component.
 */
#define GCMI_COST_CALCULATION_DEBUG_LEVEL	C_MSG_WARN
// #define GCMI_COST_CALCULATION_DEBUG_LEVEL	C_MSG_VERBOSE

/**
 * @brief Flag whether to use the coordinator node for synchronization.
 */
#define GCMI_USE_COORDINATOR_FOR_SYNCHRONIZATION



/**
 * @brief Calculate service cost between clients and hosts for a multiple
 * service instances.
 *
 * Time complexity:	O(nodes * hosts)
 * Space complexity:	O(1)
 *
 * @param node_stats Pointer to list of node statistics for a service.
 * @param host_stats Pointer to list of service instance host statistics (return value).
 * @return Service cost between clients for given service configuration.
 */
static float calculate_client_level_service_cost(const list_t* node_stats, list_t* host_stats) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	float distance_to_closest_service_host = FLT_MAX;
	SERVICE_host_stat_t* closest_host_stat = NULL;
	SERVICE_host_stat_t* host_stat = NULL;
	list_iterator_t host_stats_iterator;
	float client_cost = 0.0;
	float distance_to_service_host = 0.0;
	float service_cost_between_clients = 0.0;

	list_iterator_init(node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {

		client_cost = SERVICE_node_stat_to_cost(node_stat);

		// Skip clients without service demand.
		if(client_cost < FLT_EPSILON)
			continue;

		// Establish distance to the closest service host.
		distance_to_closest_service_host = FLT_MAX;
		closest_host_stat = NULL;
		list_iterator_init(host_stats, &host_stats_iterator);
		while((host_stat = (SERVICE_host_stat_t*) list_iterator_next(&host_stats_iterator)) != NULL) {
			// Distance from client to potentially closest service host.
			distance_to_service_host = get_distance_between_nodes(node_stat->node->address, host_stat->node->address);
			if(distance_to_service_host < distance_to_closest_service_host) {
				distance_to_closest_service_host = distance_to_service_host;
				closest_host_stat = host_stat;
			}
		}
		C_ABORT_ON(distance_to_closest_service_host == FLT_MAX, "GCMI/COST: Failed to establish distance to closest service host for %s.\n", addr_to_str(node_stat->node->address));

		if(distance_to_closest_service_host >= C_MAXIMAL_NETWORK_DIAMETER) {
			C_DEBUG(GCMI_COST_CALCULATION_DEBUG_LEVEL, C_MSG_WARN, "GCMI/COST: Ignoring cost of client %s (%d, %d, %d = %.1f) since no path to any of the %d host(s) is known...\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat), client_cost, list_size(host_stats));
			continue;
		}

		// Store service metrics for this service host.
		closest_host_stat->request_count += node_stat->request_count;
		closest_host_stat->traffic += node_stat->traffic;
		add_datarate_stats(closest_host_stat->datarate_stat, node_stat->datarate_stat);

		// Service cost is the request traffic times the distance from
		// host to client.
		service_cost_between_clients += client_cost * distance_to_closest_service_host;

		C_DEBUG(GCMI_COST_CALCULATION_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/COST: Added cost of client %s (%d, %d, %d = %.1f) to host %s at a distance of %.1f for a total of %.1f using %d host(s).\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat), client_cost, addr_to_str(closest_host_stat->node->address), distance_to_closest_service_host, service_cost_between_clients, list_size(host_stats));
	}

	C_ABORT_ON(service_cost_between_clients < 0, "GCMI/COST: Service cost between clients is negative (%.1f).\n", service_cost_between_clients);
	C_DEBUG(GCMI_COST_CALCULATION_DEBUG_LEVEL, C_MSG_INFO, "GCMI/COST: Service cost between clients is %.1f using %d host(s).\n", service_cost_between_clients, list_size(host_stats));
	return service_cost_between_clients;
}

/**
 * @brief Calculate service cost between hosts for a multiple service instances.
 *
 * Time complexity:	O(hosts ^ 2)
 * Space complexity:	O(1)
 *
 * @param host_stats Pointer to list of service instance host statistics.
 * @param sync_traffic_fraction Fraction of data traffic that is required for synchronization between service instances.
 * @param coordinator_address Address of the coordinator node.
 * @return Service cost between service instances for given service configuration.
 */
static float calculate_host_level_service_cost(const list_t* host_stats, float sync_traffic_fraction, C_addr_t coordinator_address) {
	SERVICE_host_stat_t* src_host_stat = NULL;
	SERVICE_host_stat_t* dst_host_stat = NULL;
	list_iterator_t host_stats_iterator_1;
	list_iterator_t host_stats_iterator_2;
	float service_cost_between_hosts = 0;

#ifdef GCMI_USE_COORDINATOR_FOR_SYNCHRONIZATION
	// Add cost of inter-instance communication between all
	// instances and the coordinator.

	list_iterator_init(host_stats, &host_stats_iterator_1);
	while((src_host_stat = (SERVICE_host_stat_t*) list_iterator_next(&host_stats_iterator_1)) != NULL) {

		// Add cost from source host to coordinator.
		if(src_host_stat->node->address != coordinator_address) {
			service_cost_between_hosts += SERVICE_host_stat_to_cost(src_host_stat) * get_distance_between_nodes(src_host_stat->node->address, coordinator_address) * sync_traffic_fraction;
		}

		// Add cost from coordinator to all other hosts.
		list_iterator_init(host_stats, &host_stats_iterator_2);
		while((dst_host_stat = (SERVICE_host_stat_t*) list_iterator_next(&host_stats_iterator_2)) != NULL) {
			if(
				(dst_host_stat->node->address != coordinator_address)
				&& (dst_host_stat->node->address != src_host_stat->node->address)
			) {
				service_cost_between_hosts += SERVICE_host_stat_to_cost(src_host_stat) * get_distance_between_nodes(coordinator_address, dst_host_stat->node->address) * sync_traffic_fraction;
			}
		}
	}
#else
	// Add cost of inter-instance communication for all pairs of
	// instances.

	list_iterator_init(host_stats, &host_stats_iterator_1);
	while((src_host_stat = (SERVICE_host_stat_t*) list_iterator_next(&host_stats_iterator_1)) != NULL) {
		list_iterator_init(host_stats, &host_stats_iterator_2);
		while((dst_host_stat = (SERVICE_host_stat_t*) list_iterator_next(&host_stats_iterator_2)) != NULL) {
			if(dst_host_stat->node->address != src_host_stat->node->address) {

				// Service synchronization cost is the service
				// traffic times the distance to another host
				// multiplied by the synchronization fraction.
				service_cost_between_hosts += SERVICE_host_stat_to_cost(src_host_stat) * get_distance_between_nodes(src_host_stat->node->address, dst_host_stat->node->address) * sync_traffic_fraction;

			}
		}
	}
#endif	// GCMI_USE_COORDINATOR_FOR_SYNCHRONIZATION

	C_ABORT_ON(service_cost_between_hosts < 0, "GCMI/COST: Service cost between hosts is negative (%.1f).\n", service_cost_between_hosts);
	C_DEBUG(GCMI_COST_CALCULATION_DEBUG_LEVEL, C_MSG_INFO, "GCMI/COST: Service cost between hosts is %.1f using %d host(s).\n", service_cost_between_hosts, list_size(host_stats));
	return service_cost_between_hosts;
}

/**
 * @brief Calculate control cost between hosts and coordinator.
 *
 * Time complexity:	O(hosts)
 * Space complexity:	O(1)
 *
 * @param host_stats Pointer to list of service instance host statistics.
 * @param coordinator_address Address of the coordinator node.
 * @return Service cost between service instances and coordinator.
 */
static float calculate_control_cost(const list_t* host_stats, C_addr_t coordinator_address) {
	SERVICE_host_stat_t* src_host_stat = NULL;
	list_iterator_t host_stats_iterator;
	float control_cost = 0;

	list_iterator_init(host_stats, &host_stats_iterator);
	while((src_host_stat = (SERVICE_host_stat_t*) list_iterator_next(&host_stats_iterator)) != NULL) {

		// Add cost from source host to coordinator, assuming that
		// a packet making best use of the available MTU is transmitted
		// to the coordinator every average host trigger interval.
		if(src_host_stat->node->address != coordinator_address) {
			control_cost += ((float) C_MTU)
				/ (((float) (GCMI_HOST_TRIGGER_DURATION + GCMI_HOST_TRIGGER_MAX_DURATION)) / 2.0)
				* get_distance_between_nodes(src_host_stat->node->address, coordinator_address);
		}
	}

	C_ABORT_ON(control_cost < 0, "GCMI/COST: Control cost is negative (%.1f).\n", control_cost);
	C_DEBUG(GCMI_COST_CALCULATION_DEBUG_LEVEL, C_MSG_INFO, "GCMI/COST: Control cost is %.1f using %d host(s).\n", control_cost, list_size(host_stats));
	return control_cost;
}

/**
 * @brief Calculate service cost for a multiple service instances.
 *
 * Time complexity:	O(hosts + nodes * hosts + hosts ^ 2 + hosts)
 *			= O(nodes ^ 2)
 * Space complexity:	O(1)
 *
 * @param service Pointer to service.
 * @param host_addresses Pointer to list of service instance host addresses.
 * @param coordinator_address Address of the coordinator node.
 * @return Service cost for given service configuration.
 */
float calculate_service_cost(const SERVICE_service_t* service, const list_t* host_addresses, C_addr_t coordinator_address) {
	C_addr_t* host_address = NULL;
	list_iterator_t host_addresses_iterator;
	list_t* host_stats = NULL;
	SERVICE_host_stat_t* host_stat = NULL;
	float service_cost = 0;

	list_alloc(&host_stats);

	// Initialize list of host statistics.
	list_iterator_init(host_addresses, &host_addresses_iterator);
	while((host_address = (C_addr_t*) list_iterator_next(&host_addresses_iterator)) != NULL) {
		SERVICE_alloc_host_stat(&host_stat, *host_address);
		list_add(host_stats, host_stat);
		host_stat = NULL;
	}

	// Calculate service cost between clients.
	service_cost += calculate_client_level_service_cost(service->node_stats, host_stats);

	// Calculate service cost between service instances.
	service_cost += calculate_host_level_service_cost(host_stats, service->sync_traffic_fraction, coordinator_address);

	// Calculate cost of control traffic.
	service_cost += calculate_control_cost(host_stats, coordinator_address);

	list_free_including_data(&host_stats, SERVICE_free_host_stat_list);

	C_ABORT_ON(service_cost < 0, "GCMI/COST: Service cost is negative (%.1f).\n", service_cost);
	C_DEBUG(GCMI_COST_CALCULATION_DEBUG_LEVEL, C_MSG_NOTE, "GCMI/COST: Service cost is %.1f using %d host(s).\n", service_cost, list_size(host_addresses));
	return service_cost;
}



#endif	// C_SERVICE_PLACEMENT_EVALUATION
