/**
 * @file graph_cost_single_instance.c
 *
 * @brief Graph cost single instance.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: graph_cost_single_instance.c 3408 2010-03-15 18:52:15Z gwitten $
 */

#include <float.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/mapping.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "graph_cost_single_instance.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION

#ifndef C_USE_NETWORK_MAPPING
	#error Trying to compile GCSI without network mapping support.
#endif	// C_USE_NETWORK_MAPPING

#ifndef C_ENABLE_SPI_OPTIMIZATIONS
	#warning Compiling GCSI without SPI optimizations...
#endif	// C_ENABLE_SPI_OPTIMIZATIONS



/**
 * @brief Interface to GCSI algorithm.
 */
PLACEMENT_algorithm_t graph_cost_single_instance_algorithm = {
	C_PLACEMENT_ALGORITHM_GRAPH_COST_SINGLE_INSTANCE,
	graph_cost_single_instance_start,
	graph_cost_single_instance_stop,
	graph_cost_single_instance_trigger,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

/**
 * @brief GCSI algorithm data.
 */
typedef struct {
	C_time_t		last_run_time;			///< Time of the last run of the algorithm.
	C_time_t		found_superior_placement_time;	///< Point in time at which a superior host was found; or 0 if no superior host is known.
	bool			migration_in_progress;		///< Flag whether an migration is in progress.
} graph_cost_single_instance_data_t;

/**
 * @brief Parameter structure for service migration process.
 */
typedef struct {
	SERVICE_id_t		service_id[SERVICE_ID_LENGTH];	///< Service ID.
	SERVICE_node_stat_t*	migration_target_node_stat;	///< Node stat of the migration target.
} GCSI_migration_parameter_t;

/**
 * @brief Debug level for this component.
 */
#define GCSI_DEBUG_LEVEL	C_MSG_INFO

/**
 * @brief Minimal time interval between subsequent runs of the algorithm.
 */
#define GCSI_MINIMAL_RUN_INTERVAL			1.0

/**
 * @brief Hysteresis (multiplier) between estimated saving and migration cost.
 */
#define GCSI_MIGRATION_HYSTERESIS_MULTIPLIER		1.05



/**
 * @brief Calculate service cost for a single service instance at a given node.
 *
 * The cost of clients to which no distance is known is ignored. For some
 * scenarios it might however be beneficial to use some kind of average hop
 * count instead for these nodes. Possibly a TODO.
 *
 * Time complexity:	O(nodes)
 * Space complexity:	O(1)
 *
 * @param node_stats Pointer to list of node statistics for a service.
 * @param service_host_address Address of candidate service instance host.
 * @return Service cost for serivice instance at given node.
 */
float GCSI_calculate_service_cost(const list_t* node_stats, C_addr_t service_host_address) {
	float service_cost = 0.0;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	float client_cost = 0.0;
	float distance_to_service_host = 0.0;

	list_iterator_init(node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {

		client_cost = SERVICE_node_stat_to_cost(node_stat);

		// Skip clients without service demand.
		if(client_cost < FLT_EPSILON)
			continue;

		// Distance from client to potential service host.
		distance_to_service_host = get_distance_between_nodes(node_stat->node->address, service_host_address);

		if(distance_to_service_host >= C_MAXIMAL_NETWORK_DIAMETER) {
			C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_WARN, "GCSI: Ignoring cost of client %s (%d, %d, %d = %.1f) since no path to host %s is known...\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat), client_cost, addr_to_str(service_host_address));
			continue;
		}

		// Service cost is the client cost times the distance
		// from host to client.
		service_cost += client_cost * distance_to_service_host;

		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCSI: Added cost of client %s (%d, %d, %d = %.1f) to host %s at a distance of %.1f for a total of %.1f.\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat), client_cost, addr_to_str(service_host_address), distance_to_service_host, service_cost);
	}

	C_ABORT_ON(service_cost < 0, "GCSI: Service cost is negative (%.1f).\n", service_cost);

	return service_cost;
}

/**
 * @brief Calculate recommended target node.
 *
 * Calculate the recommended target node based on the network map.
 *
 * @param service Pointer to service.
 * @param current_service_cost Pointer to current service cost to be calculated.
 * @param recommended_service_cost Pointer to recommended service cost to be
 * calculated.
 * @return Address of recommended host.
 */
static C_addr_t GCSI_calculate_recommended_host(const SERVICE_service_t* service, float* current_service_cost, float* recommended_service_cost) {
	node_t* service_host_candidate_node = NULL;
	list_iterator_t known_nodes_iterator;
	float service_host_candidate_cost = 0;
	SERVICE_node_stat_t* node_stat = NULL;
	SERVICE_node_stat_t cmp_node_stat;
	node_t cmp_node;
	C_addr_t recommended_target_node_address = C_BROADCAST_ADDRESS;

	*current_service_cost = FLT_MAX;
	*recommended_service_cost = FLT_MAX;

	// Find the node with the smallest service cost, as given by the sum
	// over all nodes of the product of distance and service traffic.
	list_iterator_init(C_known_nodes, &known_nodes_iterator);
	while((service_host_candidate_node = (node_t*) list_iterator_next(&known_nodes_iterator)) != NULL) {

		// Skip nodes to which no path is known.
		if(!exists_known_path(service_host_candidate_node->address, C_LOCAL_ADDRESS))
			continue;

#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
		// Skip nodes which are not direct neighbors. Migrating to more
		// distant nodes has a lower probability of success.
		if(
			(service_host_candidate_node->address != C_LOCAL_ADDRESS)
			&& !exists_known_link(service_host_candidate_node->address, C_LOCAL_ADDRESS)
		)
			continue;
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

		// Calculate the service cost for the current service host of
		// this loop iteration.
		service_host_candidate_cost = GCSI_calculate_service_cost(service->node_stats, service_host_candidate_node->address);

		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_DEBUG, "GCSI: Estimated service cost for service \"%s\" at %s is %.1f.\n", service->id, addr_to_str(service_host_candidate_node->address), service_host_candidate_cost);

		// Set the local service cost to calculate the estimated
		// savings later on.
		if(service_host_candidate_node->address == C_LOCAL_ADDRESS)
			*current_service_cost = service_host_candidate_cost;

		// Set recommended target node and service cost if the cost for
		// the current service host in this loop iteration is lower.
		memset(&cmp_node_stat, 0, sizeof(SERVICE_node_stat_t));
		memset(&cmp_node, 0, sizeof(node_t));
		cmp_node_stat.node = &cmp_node;
		cmp_node_stat.node->address = service_host_candidate_node->address;
		if(
			(*recommended_service_cost > service_host_candidate_cost)
			&& (	// Only consider nodes that have not refused a migration in the past.
				((node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, SERVICE_matches_node_stat_address)) == NULL)
				|| (!node_stat->refused_migration)	// We know node_stat != NULL.
			)
		) {
			recommended_target_node_address = service_host_candidate_node->address;
			*recommended_service_cost = service_host_candidate_cost;
		}
	}

	C_ABORT_ON(*current_service_cost == FLT_MAX, "GCSI: Failed to calculate local service cost.\n");
	C_ABORT_ON(*recommended_service_cost == FLT_MAX, "GCSI: Failed to calculate recommended service cost.\n");
	C_ABORT_ON(recommended_target_node_address == C_BROADCAST_ADDRESS, "GCSI: Failed to establish recommended target node address.\n");

	return recommended_target_node_address;
}

/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void graph_cost_single_instance_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	GCSI_migration_parameter_t* migration_parameter = (GCSI_migration_parameter_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_NOTE, "GCSI: Migration of service \"%s\" was successful.\n", migration_parameter->service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_WARN, "GCSI: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		SERVICE_shutdown_service(migration_parameter->service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_WARN, "GCSI: Migration of service \"%s\" timed out.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_WARN, "GCSI: Migration of service \"%s\" has been aborted.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_ERROR, "GCSI: Migration of service \"%s\" failed because the local system is busy.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_ERROR, "GCSI: Migration of service \"%s\" is already in progress.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_WARN, "GCSI: Migration of service \"%s\" failed due to a communication error.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_ERROR, "GCSI: Migration of service \"%s\" failed because the remote system refused to host the service.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		migration_parameter->migration_target_node_stat->refused_migration = true;
		break;
	default:
		C_ABORT_ON(true, "GCSI: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(migration_parameter);
	FREE(result);
}

/**
 * @brief Graph cost single instance algorithm.
 *
 * @param service Pointer to service for which to run this algorithm.
 */
static void graph_cost_single_instance_run(SERVICE_service_t* service) {
	graph_cost_single_instance_data_t* placement_data = (graph_cost_single_instance_data_t*) service->placement_data;
	float current_service_cost = FLT_MAX;
	float recommended_service_cost = FLT_MAX;
	C_addr_t recommended_host_address = C_BROADCAST_ADDRESS;
	C_time_t migration_amortization_time = 0;
	float estimated_migration_cost = 0;
	float estimated_savings = 0;
	SERVICE_node_stat_t* migration_target_node_stat = NULL;
	SERVICE_node_stat_t cmp_node_stat;
	node_t cmp_node;
	SERVICE_migration_result_t* result = NULL;
	GCSI_migration_parameter_t* migration_parameter = NULL;

	// Calculate recommended service host and cost.
	recommended_host_address = GCSI_calculate_recommended_host(service, &current_service_cost, &recommended_service_cost);
	if(recommended_host_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_ERROR, "GCSI: No potential target for the migration of service \"%s\" could be established.\n", service->id);
		return;
	} else if(recommended_host_address == C_LOCAL_ADDRESS) {
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_INFO, "GCSI: Placement of service \"%s\" remains unchanged, recommended placement matches current placement.\n", service->id);
		return;
	}
	C_ABORT_ON(!exists_known_path(recommended_host_address, C_LOCAL_ADDRESS), "GCSI: No path known to recommended host %s.\n", addr_to_str(recommended_host_address));

	// Store current and optimal cost in statistics variables.
	C_PLACEMENT_current_service_cost = current_service_cost;
	C_PLACEMENT_currently_optimal_service_cost = recommended_service_cost;

	// Abort if the the recommended placement is more costly than the
	// current placement.
	if(current_service_cost <= (recommended_service_cost * GCSI_MIGRATION_HYSTERESIS_MULTIPLIER)) {
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_INFO, "GCSI: Placement of service \"%s\" remains unchanged, recommended placement on host %s doesn't substantially reduce the cost (%.1f, %.1f\%).\n", service->id, addr_to_str(recommended_host_address), (recommended_service_cost - current_service_cost), ((recommended_service_cost - current_service_cost) / current_service_cost * 100));
		placement_data->found_superior_placement_time = 0.0;
		return;
	}

	// If this is the first time that a superior placement has been found,
	// keep a note of the time (required below) and abort for now.
	if(placement_data->found_superior_placement_time == 0.0) {
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_NOTE, "GCSI: Placement of service \"%s\" remains unchanged, but superior host %s has been found (%.1f, %.1f\%).\n", service->id, addr_to_str(recommended_host_address), (recommended_service_cost - current_service_cost), ((recommended_service_cost - current_service_cost) / current_service_cost * 100));
		placement_data->found_superior_placement_time = C_time();
		return;
	}

	// Calculate adaptation plan and cost.
	migration_amortization_time = C_time() - placement_data->found_superior_placement_time;
	estimated_migration_cost = ((float) service->data_size) * get_distance_between_nodes(C_LOCAL_ADDRESS, recommended_host_address);
	estimated_migration_cost /= ((float) migration_amortization_time);

	// Store estimated adaptation cost in statistics variable.
	C_PLACEMENT_currently_estimated_adaptation_cost = estimated_migration_cost;

	// Check whether it's worth the trouble to migrate the service.
	estimated_savings = current_service_cost - recommended_service_cost;
	if(estimated_savings <= estimated_migration_cost) {
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_INFO, "GCSI: Placement of service \"%s\" remains unchanged, migration to host %s is too expensive (%.1f savings <= %.1f cost).\n", service->id, addr_to_str(recommended_host_address), estimated_savings, estimated_migration_cost);
		return;
	}

	C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_NOTE, "GCSI: Migrating service \"%s\" from %s to %s (savings %.1f > cost %.1f)...\n", service->id, addr_to_str(C_LOCAL_ADDRESS), addr_to_str(recommended_host_address), estimated_savings, estimated_migration_cost);
	memset(&cmp_node_stat, 0, sizeof(SERVICE_node_stat_t));
	memset(&cmp_node, 0, sizeof(node_t));
	cmp_node_stat.node = &cmp_node;
	cmp_node_stat.node->address = recommended_host_address;
	if((migration_target_node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, SERVICE_matches_node_stat_address)) == NULL) {
		SERVICE_alloc_node_stat(&migration_target_node_stat, recommended_host_address);
		list_add(service->node_stats, migration_target_node_stat);
	}

	SERVICE_lock_service(service->id);
	MALLOC(result, SERVICE_migration_result_t);
	MALLOC(migration_parameter, GCSI_migration_parameter_t);
	result->parameter = migration_parameter;
	SERVICE_copy_service_id(migration_parameter->service_id, service->id);
	migration_parameter->migration_target_node_stat = migration_target_node_stat;
	SERVICE_migrate_service(service->id, recommended_host_address, graph_cost_single_instance_callback, result);

	// Reset statistics variables, if the host is about to change.
	C_PLACEMENT_current_service_cost = 0.0;
	C_PLACEMENT_currently_optimal_service_cost = 0.0;
	C_PLACEMENT_currently_optimal_number_of_hosts = 0;
	C_PLACEMENT_currently_estimated_adaptation_cost = 0.0;
}

/**
 * @brief Start graph cost single instance algorithm.
 *
 * @param service Pointer to service for which to start this algorithm.
 */
void graph_cost_single_instance_start(SERVICE_service_t* service) {
	graph_cost_single_instance_data_t* placement_data = NULL;

	C_ABORT_ON(service->placement_data != NULL, "GCSI: Trying to start algorithm for service \"%s\", but placement data is not NULL.\n", service->id);

	C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_DEBUG, "GCSI: Starting algorithm for service \"%s\"...\n", service->id);
	MALLOC(service->placement_data, graph_cost_single_instance_data_t);
	placement_data = (graph_cost_single_instance_data_t*) service->placement_data;
	placement_data->last_run_time = C_time();	// Initialize to current time to gather some data before the first run.
	placement_data->found_superior_placement_time = 0.0;
	placement_data->migration_in_progress = false;

	// The number of service hosts remains constant.
	C_PLACEMENT_currently_optimal_number_of_hosts = 1;
}

/**
 * @brief Stop graph cost single instance algorithm.
 *
 * @param service Pointer to service for which to stop this algorithm.
 */
void graph_cost_single_instance_stop(SERVICE_service_t* service) {
	graph_cost_single_instance_data_t* placement_data = (graph_cost_single_instance_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "GCSI: Trying to stop algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_DEBUG, "GCSI: Stopping algorithm for service \"%s\"...\n", service->id);
	FREE(service->placement_data);
}

/**
 * @brief Trigger graph cost single instance algorithm.
 *
 * @param service Pointer to service for which to trigger this algorithm.
 * @param reason_id Numeric ID for the reason why the algorithm was triggered.
 * @param reason_parameter Pointer to optional, reason-specific parameter of
 * the reason.
 */
void graph_cost_single_instance_trigger(SERVICE_service_t* service, uint8_t reason_id, const void* reason_parameter) {
	graph_cost_single_instance_data_t* placement_data = (graph_cost_single_instance_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "GCSI: Trying to trigger algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	// Only run algorithm in certain minimal intervals.
	if((C_time() - placement_data->last_run_time) > GCSI_MINIMAL_RUN_INTERVAL) {
		C_DEBUG(GCSI_DEBUG_LEVEL, C_MSG_DEBUG, "GCSI: Triggering algorithm for service \"%s\"...\n", service->id);
		graph_cost_single_instance_run(service);
		placement_data->last_run_time = C_time();
	}
}

#else
	#warning Compiling without support for GCSI...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
