/**
 * @file linkpull_migration.c
 *
 * @brief LinkPull migration algorithm.
 *
 * This is an implementation of the LinkPull service placement algorithm
 * proposed in:
 * Hongzhou Liu, Tom Roeder, Kevin Walsh, Rimon Barr, and Emin Guen Sirer.
 * Design and Implementation of a Single System Image Operating System for Ad
 * Hoc Networks. In Proceedings of the Third International Conference on Mobile
 * Systems, Applications, and Services (MobiSys '05), Seattle, Washington, USA,
 * June 2005.
 *
 * At the end of an epoch, this algorithm moves the service one hop along the
 * link with the greatest communication, i.e. the most traffic.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: linkpull_migration.c 3408 2010-03-15 18:52:15Z gwitten $
 */

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/callback.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../placement_types.h"
#include "linkpull_migration.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Interface to LinkPull algorithm.
 */
PLACEMENT_algorithm_t linkpull_migration_algorithm = {
	C_PLACEMENT_ALGORITHM_LINKPULL_MIGRATION,
	linkpull_migration_start,
	linkpull_migration_stop,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

/**
 * @brief LinkPull algorithm data.
 */
typedef struct {
	SERVICE_id_t*	service_id;	///< Service ID.
	callback_id_t	callback_id;	///< ID of the periodic callback.
} linkpull_migration_data_t;

/**
 * @brief Epoch of the LinkPull algorithm in seconds.
 */
#define PLACEMENT_LINKPULL_EPOCH	60



static void linkpull_migration_run(callback_parameter_t parameter);

/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void linkpull_migration_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(service_id);
	linkpull_migration_data_t* placement_data = NULL;

	if(result->error_code != C_ERROR_OK) {
		C_ABORT_ON(service == NULL, "PLACEMENT: Trying to call migration callback of LinkPull algorithm for service \"%s\", but service is NULL.\n", service_id);
		if(service->placement_data == NULL) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed (%d), but LinkPull algorithm is not active anymore.\n", service_id, result->error_code);
			goto free_parameters;
		}
		placement_data = (linkpull_migration_data_t*) service->placement_data;
	}

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" was successful.\n", service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", service_id);
		SERVICE_unlock_service(service_id);
		SERVICE_shutdown_service(service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" timed out.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" has been aborted.\n", service_id);
		SERVICE_unlock_service(service_id);
		// Don't re-add the callback if aborted by the user.
		break;
	case C_ERROR_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the local system is busy.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" is already in progress.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed due to a communication error.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the remote system refused to host the service.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

free_parameters:
	FREE(service_id);
	FREE(result);
}

/**
 * @brief LinkPull migration algorithm.
 *
 * @param parameter Pointer to service for which to run this algorithm.
 */
static void linkpull_migration_run(callback_parameter_t parameter) {
	SERVICE_id_t* service_id = (SERVICE_id_t*) parameter;
	SERVICE_service_t* service = SERVICE_get_running_service(service_id);
	linkpull_migration_data_t* placement_data = NULL;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	unsigned int max_last_hop_request_count = 0;
	node_t* node_with_max_last_hop_request_count = NULL;

	if(service == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to run LinkPull migration algorithm for service \"%s\", but service is not running.\n", service->id);
		return;
	}

	C_ABORT_ON(service->placement_data == NULL, "PLACEMENT: Trying to run LinkPull algorithm for service \"%s\", but placement data is NULL.\n", service->id);
	placement_data = (linkpull_migration_data_t*) service->placement_data;
	placement_data->callback_id = 0;

	if(service->host_stat->request_count == 0) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to run LinkPull migration algorithm for service \"%s\", but no service requests have been received.\n", service->id);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
		return;
	}

	// Look for node that was the last hop to forward most of traffic to
	// this node.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		if(node_stat->last_hop_traffic > max_last_hop_request_count) {
			max_last_hop_request_count = node_stat->last_hop_request_count;
			node_with_max_last_hop_request_count = node_stat->node;
		}
	}
	C_ABORT_ON(node_with_max_last_hop_request_count == NULL, "PLACEMENT: %d service request(s) for service \"%s\" but no requests from any last hop.\n", service->host_stat->request_count, service->id);

	if(node_with_max_last_hop_request_count->address == C_LOCAL_ADDRESS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Service \"%s\" is placed optimally on the current node (%d requests of %d total requests).\n", service->id, max_last_hop_request_count, service->host_stat->request_count);
		placement_data->callback_id = callback_add(linkpull_migration_run, service, PLACEMENT_LINKPULL_EPOCH);
	} else {
		SERVICE_migration_result_t* result = NULL;

		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migrating service \"%s\" to %s (%d requests of %d total requests)...\n", service->id, addr_to_str(node_with_max_last_hop_request_count->address), max_last_hop_request_count, service->host_stat->request_count);

		SERVICE_lock_service(service->id);
		MALLOC(result, SERVICE_migration_result_t);
		SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, service->id);
		SERVICE_migrate_service(service->id, node_with_max_last_hop_request_count->address, linkpull_migration_callback, result);
	}
}

/**
 * @brief Start LinkPull migration algorithm.
 *
 * @param service Pointer to service for which to start this algorithm.
 */
void linkpull_migration_start(SERVICE_service_t* service) {
	linkpull_migration_data_t* placement_data = NULL;

	C_ABORT_ON(service->placement_data != NULL, "PLACEMENT: Trying to start LinkPull algorithm for service \"%s\", but placement data is not NULL.\n", service->id);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Starting LinkPull algorithm for service \"%s\"...\n", service->id);
	MALLOC(service->placement_data, linkpull_migration_data_t);
	placement_data = (linkpull_migration_data_t*) service->placement_data;
	SERVICE_alloc_service_id(&placement_data->service_id, service->id);
	placement_data->callback_id = callback_add(linkpull_migration_run, placement_data->service_id, PLACEMENT_LINKPULL_EPOCH);
}

/**
 * @brief Stop LinkPull migration algorithm.
 *
 * @param service Pointer to service for which to stop this algorithm.
 */
void linkpull_migration_stop(SERVICE_service_t* service) {
	linkpull_migration_data_t* placement_data = (linkpull_migration_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "PLACEMENT: Trying to stop LinkPull algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Stopping LinkPull algorithm for service \"%s\"...\n", service->id);
	FREE(placement_data->service_id);
	if(placement_data->callback_id != 0)
		callback_cancel(placement_data->callback_id);
	FREE(service->placement_data);
}

#else
	#warning Compiling without support for LinkPull migration...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
