/**
 * @file placement_types.h
 *
 * @brief Placement (types).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: placement_types.h 3424 2010-03-18 14:45:09Z gwitten $
 */

#ifndef PLACEMENT_TYPES_H
#define PLACEMENT_TYPES_H

#include "../types.h"
#include "../service/service_types.h"



#define PLACEMENT_COMMAND_NOP			0x0	///< Placement NOP command.
#define PLACEMENT_COMMAND_COORDINATOR		0x1	///< Placement (set) COORDINATOR command.
#define PLACEMENT_COMMAND_REPLICATE		0x2	///< Placement REPLICATE command.
#define PLACEMENT_COMMAND_MIGRATE		0x3	///< Placement MIGRATE command.
#define PLACEMENT_COMMAND_SHUTDOWN		0x4	///< Placement SHUTDOWN command.
#define PLACEMENT_COMMAND_DELAYED_SHUTDOWN	0x5	///< Placement DELAYED SHUTDOWN command.

/**
 * @brief Placement command.
 */
typedef struct {
	C_addr_t	handler_address;		///< Address of the node that should handle this command.
	uint8_t		id;				///< ID of the command.
	C_addr_t	destination_address;		///< Destination address of a replication or migration command, or coordinator address.
	C_time_t	delay;				///< Delay before executing a DELAYED SHUTDOWN command.
} PLACEMENT_command_t;

/**
 * @brief Placement adaptation result.
 */
typedef struct {
	SERVICE_id_t			service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_errno_t			error_code;			///< Error code.
	callback_parameter_t		parameter;			///< Parameter for callback function.
} PLACEMENT_adaptation_result_t;

/**
 * @brief Outgoing placement adaptation plan.
 */
typedef struct {
	SERVICE_id_t			service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_addr_t			coordinator_address;		///< Address of the coordinator resulting from this adaptation.
	C_addr_t			coordinator_handler_address;	///< Address of the handler for the above coordinator.
	callback_id_t			timeout_callback_id;		///< ID of the timeout callback.
	callback_function_t		callback;			///< Callback function.
	PLACEMENT_adaptation_result_t*	result;				///< Pointer to adaptation result.
} PLACEMENT_outgoing_adaptation_t;

/**
 * @brief Incoming placement adaptation plan.
 */
typedef struct {
	SERVICE_id_t			service_id[SERVICE_ID_LENGTH];	///< Service ID.
	list_t*				commands;			///< List of commands.
	C_addr_t			coordinator_address;		///< Address of the coordinator node.
	C_addr_t			adaptation_source_address;	///< Address of the node that requested this adaptation.
	bool				current_command_handles_coordinator;	///< Flag whether the current command (while implementing an adaptation plan) handles the coordinator node.
} PLACEMENT_incoming_adaptation_t;

/**
 * @brief Service placement algorithm.
 */
typedef struct PLACEMENT_algorithm {
	uint8_t id;							///< ID of service placement algorithm.
	void (*start)(SERVICE_service_t*);				///< Pointer to start function of service placement algorithm.
	void (*stop)(SERVICE_service_t*);				///< Pointer to stop function of service placement algorithm.
	void (*trigger)(SERVICE_service_t*, uint8_t reason_id, const void* reason_parameter);	///< Pointer to trigger function of service placement algorithm.
	size_t (*get_data)(const SERVICE_service_t*, char**);		///< Pointer to get data function of service placement algorithm.
	void (*set_data)(SERVICE_service_t*, const char*, size_t);	///< Pointer to set data function of service placement algorithm.
	bool (*update_data)(SERVICE_service_t*, const PLACEMENT_incoming_adaptation_t*);	///< Pointer to update data function of service placement algorithm.
	void (*report_known_service_instance)(SERVICE_service_t*, C_addr_t);			///< Pointer to report known service instance of service placement algorithm.
	void (*report_invalid_service_instance)(SERVICE_service_t*, C_addr_t);			///< Pointer to report invalid service instance of service placement algorithm.
	list_t* (*get_synchronization_hosts)(const SERVICE_service_t*, bool);			///< Pointer to get synchronization hosts function of service placement algorithm.
} PLACEMENT_algorithm_t;

/**
 * @brief Parameter to trigger call due to STATS.
 */
typedef struct {
	C_addr_t			source_address;			///< Address of the stats source node.
	bool				valid_instance;			///< Flag whether this service instance should (in doubt) be considered valid by the coordinator node.
	C_addr_t			last_failed_migration_destination;	///< Address of the node to which the last migration attempt has failed, or broadcast if not applicable.
} PLACEMENT_trigger_reason_stats_parameter_t;



/*
 * Packet types.
 */

/**
 * @brief PLACEMENT ADAPTATION REQUEST packet.
 */
typedef struct {
	SERVICE_id_t		service_id[SERVICE_ID_LENGTH];			///< Service ID.
} PLACEMENT_ADAPTATION_REQUEST_packet_t;

/**
 * @brief PLACEMENT ADAPTATION COMPLETE packet.
 */
typedef struct {
	SERVICE_id_t		service_id[SERVICE_ID_LENGTH];			///< Service ID.
} PLACEMENT_ADAPTATION_COMPLETE_packet_t;

#endif	// PLACEMENT_TYPES_H
