/**
 * @file simple_heuristic.c
 *
 * @brief Simple heuristic algorithm.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: simple_heuristic.c 3408 2010-03-15 18:52:15Z gwitten $
 */

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/mapping.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../placement_types.h"
#include "simple_heuristic.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Interface to simple heuristic algorithm.
 */
PLACEMENT_algorithm_t simple_heuristic_algorithm = {
	C_PLACEMENT_ALGORITHM_SIMPLE_HEURISTIC,
	NULL,
	NULL,
	simple_heuristic_trigger,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

/**
 * @brief Parameter structure for service migration process.
 */
typedef struct {
	SERVICE_id_t		service_id[SERVICE_ID_LENGTH];	///< Service ID.
	SERVICE_node_stat_t*	migration_target_node_stat;	///< Node stat of the migration target.
} simple_heuristic_migration_parameter_t;

/**
 * @brief Minimal number of data points to collect before running the algorithm.
 */
#define PLACEMENT_SIMPLE_HEURISTIC_MIN_DATA_POINTS		100

/**
 * @brief Purely heuristic factor to adjust the threshold traffic metric that
 * controls whether to migrate or not.
 */
#define PLACEMENT_SIMPLE_HEURISTIC_THRESHOLD_TRAFFIC_FACTOR	2.0



/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void simple_heuristic_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	simple_heuristic_migration_parameter_t* migration_parameter = (simple_heuristic_migration_parameter_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" was successful.\n", migration_parameter->service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		SERVICE_shutdown_service(migration_parameter->service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" timed out.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" has been aborted.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the local system is busy.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" is already in progress.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed due to a communication error.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the remote system refused to host the service.\n", migration_parameter->service_id);
		SERVICE_unlock_service(migration_parameter->service_id);
		migration_parameter->migration_target_node_stat->refused_migration = true;
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(migration_parameter);
	FREE(result);
}

/**
 * @brief Simple heuristic algorithm.
 *
 * @param service Pointer to service for which to run this algorithm.
 */
static void simple_heuristic_run(SERVICE_service_t* service) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;

	// Initialize aggregate statistics.
	unsigned int direct_neighbors = 0;
	unsigned int max_last_hop_traffic = 0;
	SERVICE_node_stat_t* node_stat_with_max_last_hop_traffic = NULL;
	unsigned int threshold_traffic;

	// Calculate aggregate statistics.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {

		if(exists_known_link(C_LOCAL_ADDRESS, node_stat->node->address)) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: %d data request(s) for service \"%s\" via %s at 1 hop.\n", node_stat->last_hop_request_count, service->id, addr_to_str(node_stat->node->address));
			direct_neighbors++;
		}

		// Only consider nodes that have not refused a migration in the past.
		if(!node_stat->refused_migration && (node_stat->last_hop_traffic > max_last_hop_traffic)) {
			max_last_hop_traffic = node_stat->last_hop_traffic;
			node_stat_with_max_last_hop_traffic = node_stat;
		}
	}

	// This basically can't happen but it can't hurt to check.
	C_ABORT_ON(direct_neighbors == 0, "PLACEMENT: %d service request(s) for service \"%s\" but no direct neighbors.\n", service->host_stat->request_count, service->id);
	C_ABORT_ON(node_stat_with_max_last_hop_traffic == NULL, "PLACEMENT: %d service request(s) for service \"%s\" but no traffic.\n", service->host_stat->request_count, service->id);

	// Make a decision whether to migrate or not.
	// NOTE: This is just a simple heuristic. Nothing more, nothing less.
	threshold_traffic = (unsigned int) (((float) service->host_stat->traffic) / ((float) direct_neighbors) * PLACEMENT_SIMPLE_HEURISTIC_THRESHOLD_TRAFFIC_FACTOR);
	if(
		(max_last_hop_traffic > threshold_traffic)
		&& (node_stat_with_max_last_hop_traffic->node->address != C_LOCAL_ADDRESS)
	) {
		SERVICE_migration_result_t* result = NULL;
		simple_heuristic_migration_parameter_t* migration_parameter = NULL;

		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migrating service \"%s\" to %s (%d bytes after %d total requests, %d neighbors, %d bytes threshold traffic)...\n", service->id, addr_to_str(node_stat_with_max_last_hop_traffic->node->address), max_last_hop_traffic, service->host_stat->request_count, direct_neighbors, threshold_traffic);

		SERVICE_lock_service(service->id);
		MALLOC(result, SERVICE_migration_result_t);
		MALLOC(migration_parameter, simple_heuristic_migration_parameter_t);
		result->parameter = migration_parameter;
		SERVICE_copy_service_id(migration_parameter->service_id, service->id);
		migration_parameter->migration_target_node_stat = node_stat_with_max_last_hop_traffic;
		SERVICE_migrate_service(service->id, node_stat_with_max_last_hop_traffic->node->address, simple_heuristic_callback, result);
	} else if(node_stat_with_max_last_hop_traffic->node->address == C_LOCAL_ADDRESS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Service \"%s\" is placed optimally on the current node (%d bytes after %d total requests, %d neighbors, %d bytes threshold traffic)...\n", service->id, max_last_hop_traffic, service->host_stat->request_count, direct_neighbors, threshold_traffic);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Migration of service \"%s\" is not necessary (%d bytes after %d total requests, %d neighbors, %d bytes threshold traffic)...\n", service->id, max_last_hop_traffic, service->host_stat->request_count, direct_neighbors, threshold_traffic);
	}
}

/**
 * @brief Trigger simple heuristic algorithm.
 *
 * @param service Pointer to service for which to trigger this algorithm.
 * @param reason_id Numeric ID for the reason why the algorithm was triggered.
 * @param reason_parameter Pointer to optional, reason-specific parameter of
 * the reason.
 */
void simple_heuristic_trigger(SERVICE_service_t* service, uint8_t reason_id, const void* reason_parameter) {
	// Don't do anything if we don't have enough data points.
	if(service->host_stat->request_count < PLACEMENT_SIMPLE_HEURISTIC_MIN_DATA_POINTS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Only %d of %d required data points in statistics of service \"%s\", not running algorithm...\n", service->host_stat->request_count, PLACEMENT_SIMPLE_HEURISTIC_MIN_DATA_POINTS, service->id);
		return;
	}

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Triggering simple heuristic algorithm for service \"%s\"...\n", service->id);
	simple_heuristic_run(service);
}

#else
	#warning Compiling without support for simple heuristic...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
