/**
 * @file tree-topology_migration.c
 *
 * @brief Tree-topology migration algorithm.
 *
 * This is an implementation of the service placement algorithm proposed in:
 * Konstantinos Oikonomou and Ioannis Stavrakakis. Scalable Service Migration:
 * The Tree Topology Case. In Proceedings of the Fifth Annual Mediterranean Ad
 * Hoc Networking Workshop (Med-Hoc-Net '06), Lipari, Italy, June 2006.
 *
 * It is applicable to a single service with only one service instance. This
 * service is iteratively migrated to a neighboring node over which more than
 * half of the service requests are forwarded.
 *
 * In our implementation, we estimate aggregate service demand by looking at
 * the past service traffic forwarded by direct (one-hop) neighbors of the
 * node that currently hosts the service (see Section 6 of the aforementioned
 * paper). We omit implementing the concept of a service monitoring entity (SME)
 * for simplicity.
 *
 * To avoid wrong migration decisions due to a small data set, we further
 * require a minimum number of service requests to be processed before running
 * the algorithm.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: tree-topology_migration.c 3719 2010-08-23 13:00:36Z gwitten $
 */

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../placement_types.h"
#include "tree-topology_migration.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Interface to tree topology migration algorithm.
 */
PLACEMENT_algorithm_t tree_topology_migration_algorithm = {
	C_PLACEMENT_ALGORITHM_TREE_TOPOLOGY_MIGRATION,
	NULL,
	NULL,
	tree_topology_migration_trigger,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

/**
 * @brief Minimal number of data points to collect before running the algorithm.
 */
#define PLACEMENT_TREE_TOPOLOGY_MIGRATION_MIN_DATA_POINTS	100



/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void tree_topology_migration_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" was successful.\n", service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", service_id);
		SERVICE_unlock_service(service_id);
		SERVICE_shutdown_service(service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" timed out.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" has been aborted.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the local system is busy.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" is already in progress.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed due to a communication error.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the remote system refused to host the service.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(service_id);
	FREE(result);
}

/**
 * @brief Tree-topology migration algorithm.
 *
 * @param service Pointer to service for which to run this algorithm.
 */
static void tree_topology_migration_run(SERVICE_service_t* service) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	uint32_t max_last_hop_datarate = 0;
	node_t* node_with_max_last_hop_datarate = NULL;
	uint32_t total_datarate_via_other_last_hops = 0;

	if(SERVICE_get_running_service(service->id) == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to run Tree-topology migration algorithm for service \"%s\", but service is not running.\n", service->id);
		return;
	}
	if(service->host_stat->request_count == 0) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to run Tree-topology migration algorithm for service \"%s\", but no service requests have been received.\n", service->id);
		return;
	}

	// Look for node that was the last hop to forward most of the service
	// demand to this node.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		uint32_t datarate = calculate_datarate(node_stat->last_hop_datarate_stat);
		if(datarate > max_last_hop_datarate) {
			max_last_hop_datarate = datarate;
			node_with_max_last_hop_datarate = node_stat->node;
		}
	}
	C_ABORT_ON(node_with_max_last_hop_datarate == NULL, "PLACEMENT: %d service request(s) for service \"%s\" but no service demand.\n", service->host_stat->request_count, service->id);

	// Calculate the sum of the service demand forwarded to this node
	// via all other neighbors.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		if(node_stat->node != node_with_max_last_hop_datarate) {
			total_datarate_via_other_last_hops += calculate_datarate(node_stat->last_hop_datarate_stat);
		}
	}

	// Migrate to the last hop that forwarded most of the service demand if
	// this service demand is greater than the total service demand via all
	// other neighbors.
	if(node_with_max_last_hop_datarate->address == C_LOCAL_ADDRESS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Service \"%s\" is placed optimally on the current node (%d B/s service demand after %d requests, %d B/s from all other neighbors).\n", service->id, max_last_hop_datarate, service->host_stat->request_count, total_datarate_via_other_last_hops);
	} else if(max_last_hop_datarate > total_datarate_via_other_last_hops) {
		SERVICE_migration_result_t* result = NULL;

		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migrating service \"%s\" to %s (%d B/s service demand after %d requests, %d B/s from all other neighbors)...\n", service->id, addr_to_str(node_with_max_last_hop_datarate->address), max_last_hop_datarate, service->host_stat->request_count, total_datarate_via_other_last_hops);

		SERVICE_lock_service(service->id);
		MALLOC(result, SERVICE_migration_result_t);
		SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, service->id);
		SERVICE_migrate_service(service->id, node_with_max_last_hop_datarate->address, tree_topology_migration_callback, result);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Migration of service \"%s\" is not necessary (%d B/s service demand after %d requests, %d B/s from all other neighbors).\n", service->id, max_last_hop_datarate, service->host_stat->request_count, total_datarate_via_other_last_hops);
	}
}

/**
 * @brief Trigger tree-topology migration algorithm.
 *
 * @param service Pointer to service for which to trigger this algorithm.
 * @param reason_id Numeric ID for the reason why the algorithm was triggered.
 * @param reason_parameter Pointer to optional, reason-specific parameter of
 * the reason.
 */
void tree_topology_migration_trigger(SERVICE_service_t* service, uint8_t reason_id, const void* reason_parameter) {
	// Don't do anything if we don't have enough data points.
	if(service->host_stat->request_count < PLACEMENT_TREE_TOPOLOGY_MIGRATION_MIN_DATA_POINTS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Only %d of %d required data points in statistics of service \"%s\", not running algorithm...\n", service->host_stat->request_count, PLACEMENT_TREE_TOPOLOGY_MIGRATION_MIN_DATA_POINTS, service->id);
		return;
	}

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Triggering tree-topology migration algorithm for service \"%s\"...\n", service->id);
	tree_topology_migration_run(service);
}

#else
	#warning Compiling without support for tree-topology migration...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
