/**
 * @file service.c
 *
 * @brief Service main file.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service.c 3494 2010-05-05 22:06:16Z gwitten $
 */

#include <stdlib.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/callback.h"
#include "../util/debug.h"
#include "../util/net.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_client.h"
#include "service_server.h"
#include "service_management.h"
#include "service_replication.h"
#include "service_stats.h"
#include "service_util.h"



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the service module.
 */
void SERVICE_init() {
	// Client variables.
	C_SERVICE_total_data_requests = 0;
	C_SERVICE_successful_data_requests = 0;
	C_SERVICE_failed_data_requests = 0;
	C_SERVICE_total_successful_data_request_delay = 0;
	C_SERVICE_total_data_updates = 0;
	C_SERVICE_successful_data_updates = 0;
	C_SERVICE_failed_data_updates = 0;
	C_SERVICE_total_stream_requests = 0;
	C_SERVICE_successful_stream_requests = 0;
	C_SERVICE_failed_stream_requests = 0;
	C_SERVICE_expected_stream_packets = 0;
	C_SERVICE_received_stream_packets = 0;
	C_SERVICE_total_requests = 0;
	C_SERVICE_successful_requests = 0;
	C_SERVICE_failed_requests = 0;
	C_SERVICE_transmitted_bytes = 0;
	C_SERVICE_received_bytes = 0;
	C_SERVICE_transmitted_datarate_stat = NULL;
	alloc_datarate_stat(&C_SERVICE_transmitted_datarate_stat);
	C_SERVICE_transmitted_datarate = 0;
	C_SERVICE_received_datarate_stat = NULL;
	alloc_datarate_stat(&C_SERVICE_received_datarate_stat);
	C_SERVICE_received_datarate = 0;
	memset(C_SERVICE_request_packets_per_hop_count, 0, C_MAXIMAL_NETWORK_DIAMETER * sizeof(uint32_t));
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	C_SERVICE_indicate_request_callback_id = 0;
	C_SERVICE_indicate_activity_callback_id = 0;
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

	// Server variables.
	C_SERVICE_hosted_services = NULL;
	list_alloc(&C_SERVICE_hosted_services);
	C_SERVICE_served_streams = NULL;
	list_alloc(&C_SERVICE_served_streams);
	C_SERVICE_outgoing_replications = NULL;
	list_alloc(&C_SERVICE_outgoing_replications);
	C_SERVICE_incoming_replications = NULL;
	list_alloc(&C_SERVICE_incoming_replications);
	C_SERVICE_currently_hosted_services = 0;
	C_SERVICE_hosted_services_running_sum = 0;
	C_SERVICE_average_hosted_services = 0.0;
	C_SERVICE_available_until_time = 0.0;
	C_SERVICE_sync_overhead_bytes = 0;
	C_SERVICE_total_replications = 0;
	C_SERVICE_successful_replications = 0;
	C_SERVICE_failed_replications = 0;
	C_SERVICE_total_migrations = 0;
	C_SERVICE_successful_migrations = 0;
	C_SERVICE_failed_migrations = 0;
}

/**
 * @brief Start the service module.
 */
void SERVICE_start() {
}

/**
 * @brief Stop the service module.
 */
void SERVICE_stop() {
	SERVICE_service_t* service = NULL;

	while(!list_empty(C_SERVICE_hosted_services)) {
		service = (SERVICE_service_t*) list_head(C_SERVICE_hosted_services);
		SERVICE_shutdown_service(service->id);
	}
}

/**
 * @brief Finalize the service module.
 */
void SERVICE_finalize() {
	if((C_SERVICE_total_requests > 0) && (C_SERVICE_successful_requests == 0)) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: %d service request(s), but none of them was successful.\n", C_SERVICE_total_requests);
	}

	if(C_check_consistency_on_shutdown) {

		/// Client variables.

		// Data requests are not acknowledged, hence we can't tell whether they failed.
		/*
		C_ABORT_ON(C_SERVICE_total_data_requests != (C_SERVICE_successful_data_requests + C_SERVICE_failed_data_requests), "SERVICE: Data request counters don't add up (%d != (%d + %d)).\n", C_SERVICE_total_data_requests, C_SERVICE_successful_data_requests, C_SERVICE_failed_data_requests);
		*/
#ifdef SERVICE_CLIENT_USE_RUDP_FOR_UPDATES
		C_ABORT_ON(C_SERVICE_total_data_updates != (C_SERVICE_successful_data_updates + C_SERVICE_failed_data_updates), "SERVICE: Data update counters don't add up (%d != (%d + %d)).\n", C_SERVICE_total_data_updates, C_SERVICE_successful_data_updates, C_SERVICE_failed_data_updates);
#else
		// Data updates are not acknowledged, hence we can't tell whether they failed.
#endif	// SERVICE_CLIENT_USE_RUDP_FOR_UPDATES
		C_ABORT_ON(C_SERVICE_total_stream_requests != (C_SERVICE_successful_stream_requests + C_SERVICE_failed_stream_requests), "SERVICE: Stream request counters don't add up (%d != (%d + %d)).\n", C_SERVICE_total_stream_requests, C_SERVICE_successful_stream_requests, C_SERVICE_failed_stream_requests);

		/// Server variables.

		C_ABORT_ON(C_SERVICE_total_replications != (C_SERVICE_successful_replications + C_SERVICE_failed_replications), "SERVICE: Replication counters don't add up (%d != (%d + %d)).\n", C_SERVICE_total_replications, C_SERVICE_successful_replications, C_SERVICE_failed_replications);
		C_ABORT_ON(C_SERVICE_total_migrations != (C_SERVICE_successful_migrations + C_SERVICE_failed_migrations), "SERVICE: Migration counters don't add up (%d != (%d + %d)).\n", C_SERVICE_total_migrations, C_SERVICE_successful_migrations, C_SERVICE_failed_migrations);
	}

	C_ABORT_ON(!list_empty(C_SERVICE_outgoing_replications), "SERVICE: Trying to finalize component, but %d outgoing replication(s) active.\n", list_size(C_SERVICE_outgoing_replications));
	C_ABORT_ON(!list_empty(C_SERVICE_incoming_replications), "SERVICE: Trying to finalize component, but %d incoming replication(s) active.\n", list_size(C_SERVICE_incoming_replications));

	free_datarate_stat(&C_SERVICE_transmitted_datarate_stat);
	free_datarate_stat(&C_SERVICE_received_datarate_stat);
	list_free_including_data(&C_SERVICE_hosted_services, free_service_list);
	list_free_including_data(&C_SERVICE_served_streams, free);
	list_free(&C_SERVICE_outgoing_replications);
	list_free(&C_SERVICE_incoming_replications);
}

/**
 * @brief Do statistics for an incoming client/server packet.
 *
 * @param header Pointer to header of an incoming packet.
 */
static void stat_request_packet(const net_header_t* header) {
	C_ABORT_ON(header->destination_address == C_BROADCAST_ADDRESS, "SERVICE: Trying to stat request packet from %s sent to the broadcast address.\n", addr_to_str(header->source_address));
	C_ABORT_ON(header->hop_count > C_MAXIMAL_NETWORK_DIAMETER, "SERVICE: Trying to stat request packet from %s with a hop count that exceeds the maximal network diameter (%d > %d).\n", addr_to_str(header->source_address), header->hop_count, C_MAXIMAL_NETWORK_DIAMETER);

	C_SERVICE_request_packets_per_hop_count[header->hop_count]++;
}

/**
 * @brief Service packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void SERVICE_recv(const net_header_t* header, const void* payload) {
	char* data = NULL;
	size_t data_size = 0;

	C_ABORT_ON(header->app != APP_TYPE_SERVICE, "SERVICE: Received non-SERVICE packet.\n");

	switch(header->type) {
	case SERVICE_PACKET_TYPE_CLIENT_DATA_REQUEST:
		C_ABORT_ON(header->payload_length != sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t), "SERVICE: Received CLIENT DATA REQUEST packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t));
		stat_request_packet(header);
		handle_SERVICE_CLIENT_DATA_REQUEST(header, (SERVICE_CLIENT_DATA_REQUEST_packet_t*) payload);
		break;
	case SERVICE_PACKET_TYPE_CLIENT_DATA_UPDATE:
		C_ABORT_ON(header->payload_length < sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t), "SERVICE: Received CLIENT DATA UPDATE packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t));
		stat_request_packet(header);
		data = (((char*) payload) + sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t));
		data_size = header->payload_length - sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t);
		handle_SERVICE_CLIENT_DATA_UPDATE(header, (SERVICE_CLIENT_DATA_UPDATE_packet_t*) payload, data, data_size);
		break;
	case SERVICE_PACKET_TYPE_CLIENT_STREAM_REQUEST:
		C_ABORT_ON(header->payload_length != sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t), "SERVICE: Received CLIENT STREAM REQUST packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t));
		stat_request_packet(header);
		handle_SERVICE_CLIENT_STREAM_REQUEST(header, (SERVICE_CLIENT_STREAM_REQUEST_packet_t*) payload);
		break;
	case SERVICE_PACKET_TYPE_SERVER_DATA_ITEM:
		C_ABORT_ON(header->payload_length < sizeof(SERVICE_SERVER_DATA_ITEM_packet_t), "SERVICE: Received SERVER DATA ITEM packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SERVICE_SERVER_DATA_ITEM_packet_t));
		stat_request_packet(header);
		data = (((char*) payload) + sizeof(SERVICE_SERVER_DATA_ITEM_packet_t));
		data_size = header->payload_length - sizeof(SERVICE_SERVER_DATA_ITEM_packet_t);
		handle_SERVICE_SERVER_DATA_ITEM(header, (SERVICE_SERVER_DATA_ITEM_packet_t*) payload, data, data_size);
		break;
	case SERVICE_PACKET_TYPE_SERVER_STREAM_ITEM:
		C_ABORT_ON(header->payload_length < sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t), "SERVICE: Received SERVER STREAM ITEM packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t));
		stat_request_packet(header);
		data = (((char*) payload) + sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t));
		data_size = header->payload_length - sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t);
		handle_SERVICE_SERVER_STREAM_ITEM(header, (SERVICE_SERVER_STREAM_ITEM_packet_t*) payload, data, data_size);
		break;
	case SERVICE_PACKET_TYPE_SERVER_NOT_AVAILABLE:
		C_ABORT_ON(header->payload_length != sizeof(SERVICE_SERVER_NOT_AVAILABLE_packet_t), "SERVICE: Received SERVER NOT AVAILABLE packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SERVICE_SERVER_NOT_AVAILABLE_packet_t));
		handle_SERVICE_SERVER_NOT_AVAILABLE(header, (SERVICE_SERVER_NOT_AVAILABLE_packet_t*) payload);
		break;
	case SERVICE_PACKET_TYPE_STATS:
		C_ABORT_ON(header->payload_length < sizeof(SERVICE_STATS_packet_t), "SERVICE: Received STATS packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SERVICE_STATS_packet_t));
		data = (((char*) payload) + sizeof(SERVICE_STATS_packet_t));
		data_size = header->payload_length - sizeof(SERVICE_STATS_packet_t);
		handle_SERVICE_STATS(header, (SERVICE_STATS_packet_t*) payload, data, data_size);
		break;
	case SERVICE_PACKET_TYPE_REPLICATION_REQUEST:
		C_ABORT_ON(header->payload_length < sizeof(SERVICE_REPLICATION_REQUEST_packet_t), "SERVICE: Received REPLICATION REQUEST packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SERVICE_REPLICATION_REQUEST_packet_t));
		data = (((char*) payload) + sizeof(SERVICE_REPLICATION_REQUEST_packet_t));
		data_size = header->payload_length - sizeof(SERVICE_REPLICATION_REQUEST_packet_t);
		handle_SERVICE_REPLICATION_REQUEST(header, (SERVICE_REPLICATION_REQUEST_packet_t*) payload, data, data_size);
		break;
	case SERVICE_PACKET_TYPE_REPLICATION_DATA:
		C_ABORT_ON(header->payload_length < sizeof(SERVICE_REPLICATION_DATA_packet_t), "SERVICE: Received REPLICATION DATA packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SERVICE_REPLICATION_DATA_packet_t));
		data = (((char*) payload) + sizeof(SERVICE_REPLICATION_DATA_packet_t));
		data_size = header->payload_length - sizeof(SERVICE_REPLICATION_DATA_packet_t);
		handle_SERVICE_REPLICATION_DATA(header, (SERVICE_REPLICATION_DATA_packet_t*) payload, data, data_size);
		break;
	case SERVICE_PACKET_TYPE_REPLICATION_ACK:
		C_ABORT_ON(header->payload_length != sizeof(SERVICE_REPLICATION_ACK_packet_t), "SERVICE: Received REPLICATION ACK packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SERVICE_REPLICATION_ACK_packet_t));
		handle_SERVICE_REPLICATION_ACK(header, (SERVICE_REPLICATION_ACK_packet_t*) payload);
		break;
	case SERVICE_PACKET_TYPE_REPLICATION_NACK:
		C_ABORT_ON(header->payload_length != sizeof(SERVICE_REPLICATION_NACK_packet_t), "SERVICE: Received REPLICATION NACK packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SERVICE_REPLICATION_NACK_packet_t));
		handle_SERVICE_REPLICATION_NACK(header, (SERVICE_REPLICATION_NACK_packet_t*) payload);
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Received packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Request specific service data of a service.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 */
void SERVICE_request_data(const SERVICE_id_t* service_id, uint16_t offset, size_t size) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	request_data(service_id, offset, size);
}

/**
 * @brief Update specific service data of a service.
 *
 * @param service_id ID of the service of which to update data.
 * @param offset Offset in service data.
 * @param size Size of service data to update.
 * @param data Pointer to new service data (which is freed internally).
 */
void SERVICE_update_data(const SERVICE_id_t* service_id, uint16_t offset, size_t size, char* data) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	if(data == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: No data provided.\n");
		return;
	}

	update_data(service_id, offset, size, data);
}

/**
 * @brief Request a stream of specific service data of a service.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 * @param interval Interval at which to send stream packets in seconds.
 * @param duration Time during which the stream is to be active in seconds.
 */
void SERVICE_request_stream(const SERVICE_id_t* service_id, uint16_t offset, size_t size, C_time_t interval, C_time_t duration) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	if(interval < 0) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Interval has negative value (%f).\n", interval);
		return;
	} else if(duration < 0) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Duration has negative value (%f).\n", duration);
		return;
	}

	request_stream(service_id, offset, size, interval, duration);
}

/**
 * @brief Create a new service on this node.
 *
 * @param service_id ID of the new service.
 * @param service_data_size Size of the service data in bytes.
 * @param sync_traffic_fraction Fraction of data traffic that is required for synchronization between service instances.
 * @param placement_algorithm_id ID of service placement algorithm.
 */
void SERVICE_create_service(const SERVICE_id_t* service_id, size_t service_data_size, float sync_traffic_fraction, uint8_t placement_algorithm_id) {
	SERVICE_service_t* service = NULL;
#ifdef C_ENABLE_DEBUG
	char* p = NULL;
#endif	// C_ENABLE_DEBUG

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	if(sync_traffic_fraction < 0) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Fraction of synchronization traffic for new service \"%s\" is negative.\n", service_id);
		return;
	} else if(get_service(service_id) != NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to create service \"%s\", but a service with that ID is already hosted on this node...\n", service_id);
		return;
	} else if(list_size(C_SERVICE_hosted_services) >= SERVICE_MAX_HOSTED_SERVICES) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: No space to host additional service.\n");
		return;
	}

	alloc_service(&service, service_id, service_data_size, sync_traffic_fraction);
#ifdef C_ENABLE_DEBUG
	MALLOC_ARRAY(p, char, service_data_size);
	memset(p, 0xDA, service_data_size);
	update_service_data(service, 0x00, service_data_size, p);	// Fill service data memory with a known pattern.
	FREE(p);
#endif	// C_ENABLE_DEBUG
	start_service(service, placement_algorithm_id);
}

/**
 * @brief Shutdown a service on this node.
 *
 * @param service_id ID of the service to shutdown.
 */
void SERVICE_shutdown_service(const SERVICE_id_t* service_id) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't shutdown unknown service \"%s\".\n", service_id);
		return;
	}
	stop_service(service);

	// If we were dealing with an incoming replication, stopping the
	// service also freed the resources. Hence we have to check again.
	service = get_service(service_id);
	if(service != NULL)
		free_service(&service);
}

/**
 * @brief Retrieve a pointer to a service on this node.
 *
 * @param service_id ID of the service to be retrieved.
 * @return Pointer to the service if it exists on this node, NULL otherwise.
 */
SERVICE_service_t* SERVICE_get_service(const SERVICE_id_t* service_id) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	return get_service(service_id);
}

/**
 * @brief Retrieve a pointer to a running service on this node.
 *
 * @param service_id ID of the service to be retrieved.
 * @return Pointer to the service if it is running on this node, NULL otherwise.
 */
SERVICE_service_t* SERVICE_get_running_service(const SERVICE_id_t* service_id) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	return get_running_service(service_id);
}

/**
 * @brief Check whether a specific service is still hosted on this node.
 *
 * @param service Pointer to a service structure.
 * @return <code>true</code> if the service exists, <code>false</code>
 * otherwise.
 */
bool SERVICE_exists_service(const SERVICE_service_t* service) {
	return exists_service(service);
}

/**
 * @brief Lock a service on this node.
 *
 * @param service_id ID of the service to lock.
 */
void SERVICE_lock_service(const SERVICE_id_t* service_id) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't lock unknown service \"%s\".\n", service_id);
		return;
	}
	lock_service(service);
}

/**
 * @brief Unlock a service on this node.
 *
 * @param service_id ID of the service to unlock.
 */
void SERVICE_unlock_service(const SERVICE_id_t* service_id) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't unlock unknown service \"%s\".\n", service_id);
		return;
	}
	unlock_service(service);
}

/**
 * @brief Check whether a  service on this node is locked.
 *
 * @param service_id ID of the service to replicate.
 * @return <code>true</code> if the service is locked, <code>false</code>
 * otherwise.
 */
bool SERVICE_is_locked_service(const SERVICE_id_t* service_id) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't check for lock on unknown service \"%s\".\n", service_id);
		return false;
	}
	return is_locked_service(service);
}

/**
 * @brief Replicate a service from this node to another node.
 *
 * @param service_id ID of the service to replicate.
 * @param destination_address Address of the destination node.
 * @param callback Callback function.
 * @param result Pointer to replication result.
 */
void SERVICE_replicate_service(const SERVICE_id_t* service_id, C_addr_t destination_address, callback_function_t callback, SERVICE_replication_result_t* result) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't replicate service \"%s\", service not known.\n", service_id);
		return;
	}

	replicate_service(service, destination_address, false, callback, result);
}

/**
 * @brief Migrate a service from this node to another node.
 *
 * @param service_id ID of the service to migrate.
 * @param destination_address Address of the destination node.
 * @param callback Callback function.
 * @param result Pointer to migration result.
 */
void SERVICE_migrate_service(const SERVICE_id_t* service_id, C_addr_t destination_address, callback_function_t callback, SERVICE_migration_result_t* result) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't migrate service \"%s\", service not known.\n", service_id);
		return;
	}

	migrate_service(service, destination_address, callback, result);
}

/**
 * @brief Set the placement algorithm for a service.
 *
 * @param service_id ID of the service to migrate.
 * @param placement_algorithm_str String ID of service placement algorithm.
 */
void SERVICE_set_placement_algorithm(const SERVICE_id_t* service_id, const char* placement_algorithm_str) {
	SERVICE_service_t* service = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = get_service(service_id);
	if(service == NULL) {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Can't set placement algorithm of service \"%s\", service not running.\n", service_id);
		return;
	}

	PLACEMENT_stop_algorithm(service);
	if(strcmp(placement_algorithm_str, "none") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_NONE);
	} else if(strcmp(placement_algorithm_str, "simple_heuristic") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_SIMPLE_HEURISTIC);
	} else if(strcmp(placement_algorithm_str, "graph_cost_single_instance") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_GRAPH_COST_SINGLE_INSTANCE);
	} else if(strcmp(placement_algorithm_str, "graph_cost_multiple_instances") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_GRAPH_COST_MULTIPLE_INSTANCES);
	} else if(strcmp(placement_algorithm_str, "linkpull_migration") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_LINKPULL_MIGRATION);
	} else if(strcmp(placement_algorithm_str, "peerpull_migration") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_PEERPULL_MIGRATION);
	} else if(strcmp(placement_algorithm_str, "topocenter_migration") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_TOPOCENTER_MIGRATION);
	} else if(strcmp(placement_algorithm_str, "tree-topology_migration") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_TREE_TOPOLOGY_MIGRATION);
	} else if(strcmp(placement_algorithm_str, "asg_simple") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_ASG_SIMPLE);
	} else if(strcmp(placement_algorithm_str, "asg_event_flow_tree") == 0) {
		PLACEMENT_start_algorithm(service, C_PLACEMENT_ALGORITHM_ASG_EVENT_FLOW_TREE);
	} else {
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Invalid algorithm ID \"%s\" while setting placement algorithm for service \"%s\"...\n", placement_algorithm_str, service_id);
		return;
	}
}

/**
 * @brief Check whether the node addresses of two host statistics data
 * structures match.
 *
 * @param item Host statistic item to check.
 * @param cmp_item Reference host statistic item.
 * @return <code>true</code> if the address of the two host statistic items
 * matches, <code>false</code> otherwise.
 */
bool SERVICE_matches_host_stat_address(const void* item, const void* cmp_item) {
	return matches_host_stat_address(item, cmp_item);
}

/**
 * @brief Allocate and initialize statistics for a service host / instance.
 *
 * @param host_stat Pointer to pointer to host statistics structure to
 * initialize.
 * @param address Address of host node.
 */
void SERVICE_alloc_host_stat(SERVICE_host_stat_t** host_stat, C_addr_t address) {
	alloc_host_stat(host_stat, address);
}

/**
 * @brief Initialize statistics for a service host / instance.
 *
 * @param host_stat Pointer to host statistics structure to initialize.
 * @param address Address of host node.
 */
void SERVICE_init_host_stat(SERVICE_host_stat_t* host_stat, C_addr_t address) {
	init_host_stat(host_stat, address);
}

/**
 * @brief Free host statistics.
 *
 * @param host_stat Pointer to pointer to host statistics structure to free.
 */
void SERVICE_free_host_stat(SERVICE_host_stat_t** host_stat) {
	free_host_stat(host_stat);
}

/**
 * @brief Free host statistics (list interface).
 *
 * @param p Pointer to host statistics structure to free.
 */
void SERVICE_free_host_stat_list(void* p) {
	free_host_stat_list(p);
}

/**
 * @brief Calculate service cost based on host statistics.
 *
 * @param host_stat Pointer to host statistics data structure.
 * @return Cost of providing a service given the host statistics.
 */
float SERVICE_host_stat_to_cost(SERVICE_host_stat_t* host_stat) {
	return host_stat_to_cost(host_stat);
}

/**
 * @brief Check whether the node addresses of two node statistics data
 * structures match.
 *
 * @param item Node statistic item to check.
 * @param cmp_item Reference node statistic item.
 * @return <code>true</code> if the address of the two node statistic items
 * matches, <code>false</code> otherwise.
 */
bool SERVICE_matches_node_stat_address(const void* item, const void* cmp_item) {
	return matches_node_stat_address(item, cmp_item);
}

/**
 * @brief Allocate and initialize node statistics.
 *
 * @param node_stat Pointer to pointer to node statistics structure to
 * initialize.
 * @param address Address of the corresponding node.
 */
void SERVICE_alloc_node_stat(SERVICE_node_stat_t** node_stat, C_addr_t address) {
	alloc_node_stat(node_stat, address);
}

/**
 * @brief Free node statistics.
 *
 * @param node_stat Pointer to pointer to node statistics structure to free.
 */
void SERVICE_free_node_stat(SERVICE_node_stat_t** node_stat) {
	free_node_stat(node_stat);
}

/**
 * @brief Free node statistics (list interface).
 *
 * @param p Pointer to node statistics structure to free.
 */
void SERVICE_free_node_stat_list(void* p) {
	free_node_stat_list(p);
}

/**
 * @brief Calculate service cost based on node statistics.
 *
 * @param node_stat Pointer to node statistics data structure.
 * @return Cost of providing a service given the node statistics.
 */
float SERVICE_node_stat_to_cost(SERVICE_node_stat_t* node_stat) {
	return node_stat_to_cost(node_stat);
}

/**
 * @brief Add a node statistics data structure to a host statistics data
 * structure.
 *
 * @param host_stat Pointer to host statistics struct.
 * @param node_stat Pointer to node statistics struct.
 */
void SERVICE_add_node_stat_to_host_stat(SERVICE_host_stat_t* host_stat, SERVICE_node_stat_t* node_stat) {
	add_node_stat_to_host_stat(host_stat, node_stat);
}

/**
 * @brief Subtract a node statistics data structure to a host statistics data
 * structure.
 *
 * @param host_stat Pointer to host statistics struct.
 * @param node_stat Pointer to node statistics struct.
 */
void SERVICE_subtract_node_stat_to_host_stat(SERVICE_host_stat_t* host_stat, SERVICE_node_stat_t* node_stat) {
	subtract_node_stat_to_host_stat(host_stat, node_stat);
}

/**
 * @brief Reset node statistics.
 *
 * @param node_stat Pointer to node statistics data structure.
 * @param reset_time_independant_fields Flag whether to also reset
 * time-independant fields, or not.
 */
void SERVICE_reset_node_stat(SERVICE_node_stat_t* node_stat, bool reset_time_independant_fields) {
	init_node_stat(node_stat, reset_time_independant_fields);
}

/**
 * @brief Reset statistics of a service.
 *
 * @param service_id Service ID of the service whose stats are to be reset.
 */
void SERVICE_reset_service_stats(SERVICE_id_t* service_id) {
	reset_service_stats(service_id);
}

/**
 * @brief Send the local statistics to another node.
 *
 * @param service_id Service ID of the service whose stats are to be send.
 * @param destination_address Address of the destination node.
 */
void SERVICE_send_statistics(SERVICE_id_t* service_id, C_addr_t destination_address) {
	send_statistics(service_id, destination_address);
}

/**
 * @brief Allocate a service ID.
 *
 * @param new_service_id Pointer to new service ID.
 * @param service_id Service ID used for initialization.
 */
void SERVICE_alloc_service_id(SERVICE_id_t** new_service_id, const SERVICE_id_t* service_id) {
	alloc_service_id(new_service_id, service_id);
}

#ifdef C_ENABLE_CONSISTENCY_CHECKS
/**
 * @brief Check a service ID.
 *
 * @param service_id Service ID to be checked.
 */
void SERVICE_check_service_id(const SERVICE_id_t* service_id) {
	check_service_id(service_id);
}
#endif	// C_ENABLE_CONSISTENCY_CHECKS

/**
 * @brief Copy a service ID to another.
 *
 * @param dest_service_id Destination service ID.
 * @param src_service_id Source service ID.
 */
void SERVICE_copy_service_id(SERVICE_id_t* dest_service_id, const SERVICE_id_t* src_service_id) {
	copy_service_id(dest_service_id, src_service_id);
}

/**
 * @brief Check whether two service IDs match.
 *
 * @param service_id_1 First service ID.
 * @param service_id_2 Seconds service ID.
 * @return <code>true</code> if the two service IDs are equal,
 * <code>false</code> otherwise.
 */
bool SERVICE_same_service_id(const SERVICE_id_t* service_id_1, const SERVICE_id_t* service_id_2) {
	return same_service_id(service_id_1, service_id_2);
}

/**
 * @brief Check whether two services have the same service ID (list interface).
 *
 * @param item Service ID to check.
 * @param cmp_item Reference service ID.
 * @return <code>true</code> if the service IDs of the two services are equal,
 * <code>false</code> otherwise.
 */
bool SERVICE_matches_service_id(const void* item, const void* cmp_item) {
	return matches_service_id(item, cmp_item);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Service command replication callback function.
 *
 * @param parameter Replication result.
 */
static void SERVICE_command_replicate_service_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* result = (SERVICE_replication_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Replication of service \"%s\" was successful.\n", service_id);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" failed due to an invalid parameter.\n", service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" timed out.\n", service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" has been aborted.\n", service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" failed because system is busy.\n", service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" failed because a replication is already in progress.\n", service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" failed due to communication errors.\n", service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Replication of service \"%s\" failed because the remote system refused.\n", service_id);
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Replication callback with unknown error code 0x%X.\n", result->error_code);
	}

	SERVICE_unlock_service(service_id);
	FREE(service_id);
	FREE(result);
}

/**
 * @brief Service command migration callback function.
 *
 * @param parameter Migration result.
 */
static void SERVICE_command_migrate_service_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Migration of service \"%s\" was successful.\n", service_id);
		// No need to unlock the service, it's gone.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" failed due to an invalid parameter.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" timed out.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" has been aborted.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" failed because system is busy.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" failed because a migration is already in progress.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" failed due to communication errors.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Migration of service \"%s\" failed because the remote system refused.\n", service_id);
		SERVICE_unlock_service(service_id);
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(service_id);
	FREE(result);
}

/**
 * @brief Service command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool SERVICE_command(int argc, const char*const* argv) {
	if(argc == 3) {
		if(strcmp(argv[1], "SERVICE_shutdown_service") == 0) {
			SERVICE_shutdown_service(argv[2]);
			return true;
		} else if(strcmp(argv[1], "SERVICE_is_local_service") == 0) {
			if(SERVICE_get_running_service(argv[2]) != NULL) {
				C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Service \"%s\" is hosted on this node.\n", argv[2]);
			} else {
				C_DEBUG(SERVICE_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Service \"%s\" is not hosted on this node.\n", argv[2]);
			}
			return true;
		} else if(strcmp(argv[1], "SERVICE_set_placement_algorithm") == 0) {
			if(C_SERVICE_hosted_services != NULL) {
				SERVICE_service_t* service = NULL;
				list_iterator_t services_iterator;
				list_iterator_init(C_SERVICE_hosted_services, &services_iterator);
				while((service = (SERVICE_service_t*) list_iterator_next(&services_iterator)) != NULL)
					SERVICE_set_placement_algorithm(service->id, argv[2]);
			}
			return true;
		}
	} else if(argc == 4) {
		if(strcmp(argv[1], "SERVICE_replicate_service") == 0) {
			SERVICE_replication_result_t* result = NULL;
			SERVICE_lock_service(argv[2]);
			MALLOC(result, SERVICE_replication_result_t);
			SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, argv[2]);
			SERVICE_replicate_service(argv[2], atoi(argv[3]), SERVICE_command_replicate_service_callback, result);
			return true;
		} else if(strcmp(argv[1], "SERVICE_migrate_service") == 0) {
			SERVICE_migration_result_t* result = NULL;
			SERVICE_lock_service(argv[2]);
			MALLOC(result, SERVICE_migration_result_t);
			SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, argv[2]);
			SERVICE_migrate_service(argv[2], atoi(argv[3]), SERVICE_command_migrate_service_callback, result);
			return true;
		} else if(strcmp(argv[1], "SERVICE_set_placement_algorithm") == 0) {
			SERVICE_set_placement_algorithm(argv[2], argv[3]);
			return true;
		}
	} else if(argc == 5) {
		if(strcmp(argv[1], "SERVICE_create_service") == 0) {
			SERVICE_create_service(argv[2], atoi(argv[3]), atof(argv[4]), C_PLACEMENT_ALGORITHM_NONE);
			return true;
		} else if(strcmp(argv[1], "SERVICE_request_data") == 0) {
			SERVICE_request_data(argv[2], atoi(argv[3]), atoi(argv[4]));
			return true;
		} else if(strcmp(argv[1], "SERVICE_update_data") == 0) {
			size_t size = atoi(argv[4]);
			char* data = NULL;
			MALLOC_ARRAY(data, char, size);
			memset(data, 0xDB, size);	// Some byte pattern.
			SERVICE_update_data(argv[2], atoi(argv[3]), size, data);
			FREE(data);
			return true;
		}
	} else if(argc == 7) {
		if(strcmp(argv[1], "SERVICE_request_stream") == 0) {
			SERVICE_request_stream(argv[2], atoi(argv[3]), atoi(argv[4]), atof(argv[5]), atof(argv[6]));
			return true;
		}
	}
	return false;
}

/**
 * @brief Service calculate statistics function.
 *
 * Call periodically to gather statistics about the service module.
 */
void SERVICE_calculate_statistics() {
	SERVICE_service_t* service = NULL;
	list_iterator_t services_iterator;

	C_SERVICE_transmitted_datarate = calculate_datarate(C_SERVICE_transmitted_datarate_stat);
	C_SERVICE_received_datarate = calculate_datarate(C_SERVICE_received_datarate_stat);

	C_SERVICE_currently_hosted_services = 0;
	list_iterator_init(C_SERVICE_hosted_services, &services_iterator);
	while((service = (SERVICE_service_t*) list_iterator_next(&services_iterator)) != NULL) {
		if(is_running_service(service) || is_locked_service(service)) {
			C_SERVICE_currently_hosted_services++;
			C_SERVICE_hosted_services_running_sum++;
		}
	}

	if((C_state == C_STATE_INITIALIZED) || (C_state == C_STATE_RUNNING)) {
		C_SERVICE_average_hosted_services = ((double) C_SERVICE_hosted_services_running_sum) / ((double) C_periodic_statistics_measurement_samples);
	}

	if(C_SERVICE_currently_hosted_services > 0)
		C_SERVICE_available_until_time = (C_time() - C_init_time);
}
