/**
 * @file service.h
 *
 * @brief Service main file (header).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service.h 3413 2010-03-15 22:15:01Z gwitten $
 */

#ifndef SERVICE_H
#define SERVICE_H

#include "../types.h"
#include "service_types.h"
#include "service_vars.h"



#define SERVICE_STATE_INIT		0x0	///< Service is initializing.
#define SERVICE_STATE_RUNNING		0x1	///< Service is running.
#define SERVICE_STATE_LOCKED		0x2	///< Service is locked, i.e., read-only.
#define SERVICE_STATE_REPLICATING	0x3	///< Service is replicating to another node.
#define SERVICE_STATE_REDIRECTING	0x4	///< Service is redirecting requests to another node.
#define SERVICE_STATE_STOPPED		0x5	///< Service is stopped.

#define SERVICE_PACKET_TYPE_CLIENT_DATA_REQUEST		0x0	///< SERVICE DATA REQUEST packet type.
#define SERVICE_PACKET_TYPE_CLIENT_DATA_UPDATE		0x1	///< SERVICE DATA UPDATE packet type.
#define SERVICE_PACKET_TYPE_CLIENT_STREAM_REQUEST	0x2	///< SERVICE STREAM REQUEST packet type.
#define SERVICE_PACKET_TYPE_SERVER_DATA_ITEM		0x3	///< SERVICE DATA ITEM packet type.
#define SERVICE_PACKET_TYPE_SERVER_STREAM_ITEM		0x4	///< SERVICE STREAM ITEM packet type.
#define SERVICE_PACKET_TYPE_SERVER_NOT_AVAILABLE	0x5	///< SERVICE NOT AVAILABLE packet type.
#define SERVICE_PACKET_TYPE_STATS			0x6	///< SERVICE STATS packet type.
#define SERVICE_PACKET_TYPE_REPLICATION_REQUEST		0x7	///< SERVICE REPLICATION REQUEST packet type.
#define SERVICE_PACKET_TYPE_REPLICATION_DATA		0x8	///< SERVICE REPLICATION DATA packet type.
#define SERVICE_PACKET_TYPE_REPLICATION_ACK		0x9	///< SERVICE REPLICATION ACK packet type.
#define SERVICE_PACKET_TYPE_REPLICATION_NACK		0xa	///< SERVICE REPLICATION NACK packet type.



/**
 * @brief Use RUDP when sending update packets from the client.
 */
//#define SERVICE_CLIENT_USE_RUDP_FOR_UPDATES

/**
 * @brief Enable redirection of clients after an instance was migrated.
 */
#if defined (C_ENABLE_SPI_OPTIMIZATIONS) || defined (C_ENABLE_ASG_OPTIMIZATIONS)
	#define SERVICE_MIGRATION_USE_REDIRECTION
#endif	// (C_ENABLE_SPI_OPTIMIZATIONS || C_ENABLE_ASG_OPTIMIZATIONS)



// Interface to main module.
void SERVICE_init();
void SERVICE_start();
void SERVICE_stop();
void SERVICE_finalize();
void SERVICE_recv(const net_header_t* header, const void* payload);
bool SERVICE_command(int argc, const char*const* argv);
void SERVICE_calculate_statistics();

// Interface provided by this module (client).
void SERVICE_request_data(const SERVICE_id_t* service_id, uint16_t offset, size_t size);
void SERVICE_update_data(const SERVICE_id_t* service_id, uint16_t offset, size_t size, char* data);
void SERVICE_request_stream(const SERVICE_id_t* service_id, uint16_t offset, size_t size, C_time_t interval, C_time_t duration);
// Interface provided by this module (server).
void SERVICE_create_service(const SERVICE_id_t* service_id, size_t service_data_size, float sync_traffic_fraction, uint8_t placement_algorithm_id);
void SERVICE_shutdown_service(const SERVICE_id_t* service_id);
SERVICE_service_t* SERVICE_get_service(const SERVICE_id_t* service_id);
SERVICE_service_t* SERVICE_get_running_service(const SERVICE_id_t* service_id);
bool SERVICE_exists_service(const SERVICE_service_t* service);
void SERVICE_lock_service(const SERVICE_id_t* service_id);
void SERVICE_unlock_service(const SERVICE_id_t* service_id);
bool SERVICE_is_locked_service(const SERVICE_id_t* service_id);
void SERVICE_replicate_service(const SERVICE_id_t* service_id, C_addr_t destination_address, callback_function_t callback, SERVICE_replication_result_t* result);
void SERVICE_migrate_service(const SERVICE_id_t* service_id, C_addr_t destination_address, callback_function_t callback, SERVICE_migration_result_t* result);
void SERVICE_set_placement_algorithm(const SERVICE_id_t* service_id, const char* placement_algorithm_str);
// Statistics functions.
bool SERVICE_matches_host_stat_address(const void* item, const void* cmp_item);
void SERVICE_alloc_host_stat(SERVICE_host_stat_t** host_stat, C_addr_t address);
void SERVICE_init_host_stat(SERVICE_host_stat_t* host_stat, C_addr_t address);
void SERVICE_free_host_stat(SERVICE_host_stat_t** host_stat);
void SERVICE_free_host_stat_list(void* p);
float SERVICE_host_stat_to_cost(SERVICE_host_stat_t* host_stat);
bool SERVICE_matches_node_stat_address(const void* item, const void* cmp_item);
void SERVICE_alloc_node_stat(SERVICE_node_stat_t** node_stat, C_addr_t address);
void SERVICE_free_node_stat(SERVICE_node_stat_t** node_stat);
void SERVICE_free_node_stat_list(void* p);
float SERVICE_node_stat_to_cost(SERVICE_node_stat_t* node_stat);
void SERVICE_add_node_stat_to_host_stat(SERVICE_host_stat_t* host_stat, SERVICE_node_stat_t* node_stat);
void SERVICE_subtract_node_stat_to_host_stat(SERVICE_host_stat_t* host_stat, SERVICE_node_stat_t* node_stat);
void SERVICE_reset_node_stat(SERVICE_node_stat_t* node_stat, bool reset_time_independant_fields);
void SERVICE_reset_service_stats(SERVICE_id_t* service_id);
void SERVICE_send_statistics(SERVICE_id_t* service_id, C_addr_t destination_address);
// Utility functions.
void SERVICE_alloc_service_id(SERVICE_id_t** new_service_id, const SERVICE_id_t* service_id);
#ifdef C_ENABLE_CONSISTENCY_CHECKS
void SERVICE_check_service_id(const SERVICE_id_t* service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
void SERVICE_copy_service_id(SERVICE_id_t* dest_service_id, const SERVICE_id_t* src_service_id);
bool SERVICE_same_service_id(const SERVICE_id_t* service_id_1, const SERVICE_id_t* service_id_2);
bool SERVICE_matches_service_id(const void* item, const void* cmp_item);

#endif	// SERVICE_H
