/**
 * @file service_management.c
 *
 * @brief Service management.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_management.c 3475 2010-04-26 15:07:35Z gwitten $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/callback.h"
#include "../util/debug.h"
#include "../util/net.h"
#include "../slp-light/slp-light.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_management.h"
#include "service_replication.h"
#include "service_stats.h"
#include "service_util.h"



/**
 * @brief Debug level for the management component.
 */
#define SERVICE_MANAGEMENT_DEBUG_LEVEL	C_MSG_NOTE

/**
 * @brief Allocate and initialize a new service on this node.
 *
 * @param service Pointer to pointer to the service that is to be allocated.
 * @param service_id ID of the new service.
 * @param service_data_size Size of the service data in bytes.
 * @param sync_traffic_fraction Fraction of data traffic that is required for synchronization between service instances.
 */
void alloc_service(SERVICE_service_t** service, const SERVICE_id_t* service_id, size_t service_data_size, float sync_traffic_fraction) {
	SERVICE_node_stat_t* node_stat = NULL;

	C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Allocating service \"%s\" with %d byte(s) of data...\n", service_id, service_data_size);

	MALLOC(*service, SERVICE_service_t);
	copy_service_id((*service)->id, service_id);
	(*service)->state = SERVICE_STATE_INIT;
	MALLOC_ARRAY((*service)->data, char, service_data_size);
	(*service)->data_size = service_data_size;
	(*service)->sync_traffic_fraction = sync_traffic_fraction;
	(*service)->start_time = C_time();
	alloc_host_stat(&(*service)->host_stat, C_LOCAL_ADDRESS);
	list_alloc(&(*service)->node_stats);
	alloc_node_stat(&node_stat, C_LOCAL_ADDRESS);
	list_add((*service)->node_stats, node_stat);
	(*service)->placement_algorithm = NULL;
	(*service)->placement_data = NULL;
	(*service)->outgoing_replication = NULL;
	(*service)->incoming_replication = NULL;
	(*service)->redirect_addr = C_BROADCAST_ADDRESS;
	(*service)->replace_addr = C_BROADCAST_ADDRESS;
	(*service)->last_failed_migration_destination = C_BROADCAST_ADDRESS;
	(*service)->failed_stat_transmissions = 0;

	list_add(C_SERVICE_hosted_services, *service);
}

/**
 * @brief Start a service on this node.
 *
 * @param service Pointer to the service that is to be started.
 * @param placement_algorithm_id ID of service placement algorithm.
 */
void start_service(SERVICE_service_t* service, uint8_t placement_algorithm_id) {
	C_ABORT_ON(service->state != SERVICE_STATE_INIT, "SERVICE: Trying to start service \"%s\" service that is not initializing (%d)...\n", service->id, service->state);
	C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Starting service \"%s\"...\n", service->id);
	service->state = SERVICE_STATE_RUNNING;
	service->start_time = C_time();
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	C_trace("m -t %f -s %d -n %s -h circle -c black", C_time(), C_LOCAL_ADDRESS, service->id);
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
	PLACEMENT_start_algorithm(service, placement_algorithm_id);
	SLP_announce_service(service->id, service->replace_addr);
}

/**
 * @brief Stop service on this node.
 *
 * @param service Pointer to the service that is to be stopped.
 */
void stop_service(SERVICE_service_t* service) {
	C_ABORT_ON(service->state == SERVICE_STATE_STOPPED, "SERVICE: Trying to stop service \"%s\" that has already been stopped...\n", service->id);

	C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Stopping service \"%s\"...\n", service->id);
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	C_trace("m -t %f -s %d -n %s -X", C_time(), C_LOCAL_ADDRESS, service->id);
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

	// Abort outgoing and incoming replications.
	if(service->outgoing_replication != NULL) {
		C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Aborting outgoing replication while stopping service \"%s\"...\n", service->id);
		abort_outgoing_replication(service->outgoing_replication, C_ERROR_ABORTED);
		C_SERVICE_total_replications--;
		if(is_locked_service(service))
			unlock_service(service);
	}
	if(service->incoming_replication != NULL) {
		C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Aborting incoming replication while stopping service \"%s\"...\n", service->id);
		abort_incoming_replication(service->incoming_replication);
		free_service(&service);
		return;
	}

	// Stop placement algorithm and notify clients.
	PLACEMENT_stop_algorithm(service);
	SLP_purge_service(service->id, service->redirect_addr);

	// TODO: Shutdown streams when stopping a service (including cancelling their current callbacks) and notify clients.

	// Stop service.
	service->state = SERVICE_STATE_STOPPED;
}

/**
 * @brief Free a service on this node.
 *
 * @param service Pointer to pointer to the service that is to be freed.
 */
void free_service(SERVICE_service_t** service) {
	C_ABORT_ON(!list_contains(C_SERVICE_hosted_services, *service), "SERVICE: Trying to free service at %p which is not contained in list of hosted services.\n", *service);

	C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Freeing service \"%s\"...\n", (*service)->id);
	list_remove(C_SERVICE_hosted_services, *service);
	free_service_list(*service);
	*service = NULL;
}

/**
 * @brief Free a service on this node (list interface).
 *
 * @param p Pointer to the service that is to be freed.
 */
void free_service_list(void* p) {
	SERVICE_service_t* service = (SERVICE_service_t*) p;

	if(
		(service->state != SERVICE_STATE_INIT)
		&& (service->state != SERVICE_STATE_STOPPED)
	) {
		C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Freeing service \"%s\" that has not been stopped (%d)...\n", service->id, service->state);
	}

	if(service->placement_data != NULL) {
		C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Freeing service \"%s\", but placement data has not been freed before...\n", service->id);
		FREE(service->placement_data);
	}

	FREE(service->data);
	free_host_stat(&service->host_stat);
	list_free_including_data(&service->node_stats, free_node_stat_list);
	FREE(service);
}

/**
 * @brief Update service data of a service hosted on this node.
 *
 * @param service Pointer to the service.
 * @param offset Offset in service data.
 * @param size Size of the new service data in bytes.
 * @param data Pointer to new service data.
 */
void update_service_data(SERVICE_service_t* service, uint16_t offset, size_t size, const char* data) {
	if((offset + size) > service->data_size) {
		C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Trying to update %d byte(s) at offset 0x%X of service \"%s\", but service data size is only %d...\n", size, offset, service->id, service->data_size);
		return;
	}

	C_DEBUG(SERVICE_MANAGEMENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Updating %d byte(s) at offset 0x%X of service \"%s\"...\n", size, offset, service->id);
	memcpy(service->data + offset, data, size);
}

/**
 * @brief Check whether a specific service is still hosted on this node.
 *
 * @param service Pointer to a service structure.
 * @return <code>true</code> if the service exists, <code>false</code>
 * otherwise.
 */
bool exists_service(const SERVICE_service_t* service) {
	return list_contains(C_SERVICE_hosted_services, service);
}

/**
 * @brief Check whether a service on this node is running.
 *
 * @param service Pointer to the service.
 * @return <code>true</code> if the service is running, <code>false</code>
 * otherwise.
 */
bool is_running_service(const SERVICE_service_t* service) {
	return (service->state == SERVICE_STATE_RUNNING);
}

/**
 * @brief Check whether a service on this node is redirecting clients.
 *
 * @param service Pointer to the service.
 * @return <code>true</code> if the service is redirecting, <code>false</code>
 * otherwise.
 */
bool is_redirecting_service(const SERVICE_service_t* service) {
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	bool is_redirecting = (
		(service->state == SERVICE_STATE_REDIRECTING)
		|| (service->state == SERVICE_STATE_STOPPED)
	);

	if(is_redirecting)
		C_ABORT_ON(service->redirect_addr == C_BROADCAST_ADDRESS, "SERVICE: Service \"%s\" is redirecting, but the redirect address is set to broadcast.\n", service->id);

	return is_redirecting;
#else
	return false;
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
}

/**
 * @brief Lock a service on this node.
 *
 * @param service Pointer to the service.
 */
void lock_service(SERVICE_service_t* service) {
	C_ABORT_ON(service->state != SERVICE_STATE_RUNNING, "SERVICE: Trying to lock service \"%s\", but service is not running (%d).\n", service->id, service->state);
	service->state = SERVICE_STATE_LOCKED;
}

/**
 * @brief Unlock a service on this node.
 *
 * @param service Pointer to the service.
 */
void unlock_service(SERVICE_service_t* service) {
	C_ABORT_ON(service->state != SERVICE_STATE_LOCKED, "SERVICE: Trying to unlock service \"%s\", but service is not locked (%d).\n", service->id, service->state);
	service->state = SERVICE_STATE_RUNNING;
}

/**
 * @brief Check whether a service on this node is locked.
 *
 * @param service Pointer to the service.
 * @return <code>true</code> if the service is locked, <code>false</code>
 * otherwise.
 */
bool is_locked_service(const SERVICE_service_t* service) {
	// NOTE: A replicating or redirecting service has been locked before.
	return (
		(service->state == SERVICE_STATE_LOCKED)
		|| (service->state == SERVICE_STATE_REPLICATING)
		|| (service->state == SERVICE_STATE_REDIRECTING)
	);
}

/**
 * @brief Retrieve a service by service ID.
 *
 * @param service_id ID of the service to retrieve.
 * @return Pointer to the service if it exists, NULL otherwise.
 */
SERVICE_service_t* get_service(const SERVICE_id_t* service_id) {
	SERVICE_service_t cmp_service;
	copy_service_id(cmp_service.id, service_id);
	return (SERVICE_service_t*) list_find(C_SERVICE_hosted_services, &cmp_service, matches_service_id);
}

/**
 * @brief Check whether two running services have the same service ID.
 *
 * @param item Service to check.
 * @param cmp_item Reference service.
 * @return <code>true</code> if the service is running and the service IDs of
 * the two services are equal, <code>false</code> otherwise.
 */
static bool matches_running_service_id(const void* item, const void* cmp_item) {
	SERVICE_service_t* service = (SERVICE_service_t*) item;
	SERVICE_service_t* cmp_service = (SERVICE_service_t*) cmp_item;
	return (
		(service->state == SERVICE_STATE_RUNNING)
		&& same_service_id(service->id, cmp_service->id)
	);
}

/**
 * @brief Retrieve a pointer to a running service on this node.
 *
 * @param service_id ID of the service to be retrieved.
 * @return Pointer to the service if it is running on this node, NULL otherwise.
 */
SERVICE_service_t* get_running_service(const SERVICE_id_t* service_id) {
	SERVICE_service_t cmp_service;
	copy_service_id(cmp_service.id, service_id);
	return (SERVICE_service_t*) list_find(C_SERVICE_hosted_services, &cmp_service, matches_running_service_id);
}
