/**
 * @file service_server.c
 *
 * @brief Service - Server.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_server.c 3425 2010-03-18 14:48:58Z gwitten $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_client.h"
#include "service_server.h"
#include "service_management.h"
#include "service_stats.h"
#include "service_util.h"



/**
 * @brief Debug level for the server component.
 */
#define SERVICE_SERVER_DEBUG_LEVEL	C_MSG_WARN

/**
 * @brief Maximal number of served streams.
 */
#define SERVICE_MAX_SERVED_STREAMS	64



/*
 * Packet handling functions.
 */

/**
 * @brief Send a SERVICE DATA ITEM packet.
 *
 * Also send redirection information if the service is redirecting.
 *
 * @param to Address of destination node.
 * @param service Pointer to requested service.
 * @param offset Offset in service data.
 * @param size Requested service data size in bytes.
 * @param start_time Start time of this data request in seconds.
 */
static void send_SERVICE_SERVER_DATA_ITEM(C_addr_t to, const SERVICE_service_t* service, uint16_t offset, size_t size, C_time_t start_time) {
	SERVICE_SERVER_DATA_ITEM_packet_t data_item;

	if(size > service->data_size) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Requested too much data (%d > %d), request size reduced to %d.\n", size, service->data_size, service->data_size);
		size = service->data_size;
	}

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Sending %d byte(s) SERVER DATA ITEM for service \"%s\" (0x%X...) to %s...\n", size, service->id, (uint8_t) *service->data, addr_to_str(to));
	memset(&data_item, 0, sizeof(SERVICE_SERVER_DATA_ITEM_packet_t));
	copy_service_id(data_item.service_id, service->id);
	data_item.start_time = start_time;
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	data_item.redirect_addr = is_redirecting_service(service) ? service->redirect_addr : C_BROADCAST_ADDRESS;
#else
	data_item.redirect_addr = C_BROADCAST_ADDRESS;
#endif	// SERVICE_MIGRATION_USE_REDIRECTION

	C_sendto2(to, PACKET_PRIORITY_DATA, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_SERVER_DATA_ITEM, &data_item, sizeof(SERVICE_SERVER_DATA_ITEM_packet_t), (service->data + offset), size);

	C_SERVICE_transmitted_bytes += sizeof(SERVICE_SERVER_DATA_ITEM_packet_t) + size;
	add_bytes_to_datarate_stat(C_SERVICE_transmitted_datarate_stat, (sizeof(SERVICE_SERVER_DATA_ITEM_packet_t) + size));
}

/**
 * @brief Send a SERVICE STREAM ITEM packet.
 *
 * Also send redirection information if the service is redirecting.
 *
 * @param to Address of destination node.
 * @param service Pointer to requested service.
 * @param offset Offset in service data.
 * @param size Requested service data size in bytes.
 */
static void send_SERVICE_SERVER_STREAM_ITEM(C_addr_t to, const SERVICE_service_t* service, uint16_t offset, size_t size) {
	SERVICE_SERVER_STREAM_ITEM_packet_t stream_item;

	if(size > service->data_size) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Requested too much data (%d > %d), request size reduced to %d.\n", size, service->data_size, service->data_size);
		size = service->data_size;
	}

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending %d byte(s) SERVER STREAM ITEM for service \"%s\" (0x%X...) to %s...\n", size, service->id, (uint8_t) *service->data, addr_to_str(to));
	memset(&stream_item, 0, sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t));
	copy_service_id(stream_item.service_id, service->id);
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	stream_item.redirect_addr = is_redirecting_service(service) ? service->redirect_addr : C_BROADCAST_ADDRESS;
#else
	stream_item.redirect_addr = C_BROADCAST_ADDRESS;
#endif	// SERVICE_MIGRATION_USE_REDIRECTION

	C_sendto2(to, PACKET_PRIORITY_DATA, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_SERVER_STREAM_ITEM, &stream_item, sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t), (service->data + offset), size);

	C_SERVICE_transmitted_bytes += sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t) + size;
	add_bytes_to_datarate_stat(C_SERVICE_transmitted_datarate_stat, (sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t) + size));
}

/**
 * @brief Send SERVICE NOT AVAILABLE packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of the requested service.
 * @param redirect_addr Address to redirect future service requests to.
 */
static void send_SERVICE_SERVER_NOT_AVAILABLE(C_addr_t to, const SERVICE_id_t* service_id, C_addr_t redirect_addr) {
	SERVICE_SERVER_NOT_AVAILABLE_packet_t not_available;

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending SERVER NOT AVAILABLE for service \"%s\" to %s...\n", service_id, addr_to_str(to));
	memset(&not_available, 0, sizeof(SERVICE_SERVER_NOT_AVAILABLE_packet_t));
	copy_service_id(not_available.service_id, service_id);
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	not_available.redirect_addr = redirect_addr;
#else
	not_available.redirect_addr = C_BROADCAST_ADDRESS;
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_SERVER_NOT_AVAILABLE, &not_available, sizeof(SERVICE_SERVER_NOT_AVAILABLE_packet_t), NULL, NULL);
}

/**
 * @brief Handle incoming SERVICE DATA REQUEST packet.
 *
 * Send a SERVICE DATA ITEM packet if the service is available or a SERVICE NOT
 * AVAILABLE packet otherwise.
 *
 * @param header Header of incoming packet.
 * @param request Data request information.
 */
void handle_SERVICE_CLIENT_DATA_REQUEST(const net_header_t* header, const SERVICE_CLIENT_DATA_REQUEST_packet_t* request) {
	SERVICE_service_t* service = get_service(request->service_id);

	C_SERVICE_received_bytes += sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t);
	add_bytes_to_datarate_stat(C_SERVICE_received_datarate_stat, sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t));

	if(service == NULL) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT DATA REQUEST from %s, but service \"%s\" is not available. Notifying client...\n", addr_to_str(header->source_address), request->service_id);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, request->service_id, C_BROADCAST_ADDRESS);
		return;
	} else if(!is_running_service(service) && !is_locked_service(service)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT DATA REQUEST from %s, but service \"%s\" is not active (%d). Notifying client...\n", addr_to_str(header->source_address), request->service_id, service->state);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, request->service_id, C_BROADCAST_ADDRESS);
		return;
	} else if(service->data_size < (request->offset + request->size)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Received CLIENT DATA REQUEST for %d byte(s) at 0x%X of service \"%s\" from %s, but request size exceeds service data (%d < (%d + %d)). Dropping invalid request...\n", request->size, request->offset, request->service_id, addr_to_str(header->source_address), service->data_size, request->offset, request->size);
		return;
	}

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received CLIENT DATA REQUEST for %d byte(s) at 0x%X of service \"%s\" from %s.\n", request->size, request->offset, request->service_id, addr_to_str(header->source_address));
	send_SERVICE_SERVER_DATA_ITEM(header->source_address, service, request->offset, request->size, request->start_time);
	stat_service_request(service, request->size, header);
}

/**
 * @brief Handle incoming SERVICE DATA UPDATE packet.
 *
 * Update the service data if the srevice is available or send a SERVICE NOT
 * AVAILABLE packet otherwise.
 *
 * @param header Header of incoming packet.
 * @param update CLIENT DATA UPDATE header.
 * @param data Data to update.
 * @param data_size Size of data.
 */
void handle_SERVICE_CLIENT_DATA_UPDATE(const net_header_t* header, const SERVICE_CLIENT_DATA_UPDATE_packet_t* update, char* data, size_t data_size) {
	SERVICE_service_t* service = get_service(update->service_id);
	list_t* host_addresses = NULL;
	C_addr_t* host_address = NULL;
	list_iterator_t hosts_iterator;

	C_SERVICE_received_bytes += sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t) + data_size;
	add_bytes_to_datarate_stat(C_SERVICE_received_datarate_stat, (sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t) + data_size));

	if(service == NULL) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT DATA UPDATE from %s, but service \"%s\" is not available. Notifying client...\n", addr_to_str(header->source_address), update->service_id);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, update->service_id, C_BROADCAST_ADDRESS);
		return;
	} else if(!is_running_service(service)) {
		if(is_locked_service(service)) {
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
			if(is_redirecting_service(service)) {
				C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received CLIENT DATA UPDATE from %s, but service \"%s\" is locked. Forwarding update and redirecting client to %s...\n", addr_to_str(header->source_address), update->service_id, addr_to_str(service->redirect_addr));
				update_data(update->service_id, update->offset, data_size, data);
				send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, update->service_id, service->redirect_addr);
				return;
			} else {
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
				C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT DATA UPDATE from %s, but service \"%s\" is locked, discarding update...\n", addr_to_str(header->source_address), update->service_id, service->state);
				// NOTE: We ignore updated data that arrives
				// while a service is replicating.
				return;
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
			}
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
		} else {
			C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT DATA UPDATE from %s, but service \"%s\" is not active (%d). Notifying client...\n", addr_to_str(header->source_address), update->service_id, service->state);
			send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, update->service_id, C_BROADCAST_ADDRESS);
			return;
		}
	} else if(service->data_size < (update->offset + data_size)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Received CLIENT DATA UPDATE for %d byte(s) at 0x%X of service \"%s\" from %s, but update size exceeds service data (%d < (%d + %d)). Dropping invalid update...\n", data_size, update->offset, update->service_id, addr_to_str(header->source_address), service->data_size, update->offset, data_size);
		return;
	}

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received CLIENT DATA UPDATE for %d byte(s) at 0x%X of service \"%s\" (0x%X...) from %s.\n", data_size, update->offset, update->service_id, (uint8_t) *data, addr_to_str(header->source_address));
	update_service_data(service, update->offset, data_size, data);

	// Synchronize with other service instances.
	if((host_addresses = PLACEMENT_get_synchronization_hosts(service, update->inter_host_update)) != NULL) {
		list_iterator_init(host_addresses, &hosts_iterator);
		while((host_address = (C_addr_t*) list_iterator_next(&hosts_iterator)) != NULL)
			send_SERVICE_CLIENT_DATA_UPDATE(*host_address, service->id, update->offset, data_size, data, true);
		list_free(&host_addresses);
	}

	if(update->inter_host_update) {
		PLACEMENT_report_known_service_instance(service, header->source_address);
	} else {

#ifndef SERVICE_CLIENT_USE_RUDP_FOR_UPDATES
		C_SERVICE_successful_data_updates++;
		C_SERVICE_successful_requests++;
#endif	// SERVICE_CLIENT_USE_RUDP_FOR_UPDATES

		// Add update to service statistics.
		stat_service_request(service, data_size, header);
	}
}

/**
 * @brief Process an active stream.
 *
 * Send a SERVICE STREAM ITEM if the service is still active, possibly
 * reschedule until the next packet is due.
 *
 * @param parameter Pointer to stream to be processed.
 */
static void process_SERVICE_CLIENT_STREAM_REQUEST(callback_parameter_t parameter) {
	SERVICE_stream_t* stream = (SERVICE_stream_t*) parameter;
	SERVICE_service_t* service = get_service(stream->service_id);

	if(service == NULL) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to stream service \"%s\" to %s, but service is not available. Notifying client...\n", stream->service_id, addr_to_str(stream->to));
		send_SERVICE_SERVER_NOT_AVAILABLE(stream->to, stream->service_id, C_BROADCAST_ADDRESS);
		list_remove(C_SERVICE_served_streams, stream);
		FREE(stream);
		return;
	} else if(!is_running_service(service) && !is_locked_service(service)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to stream service \"%s\" to %s, but service is not active (%d). Notifying client...\n", stream->service_id, addr_to_str(stream->to), service->state);
		send_SERVICE_SERVER_NOT_AVAILABLE(stream->to, stream->service_id, C_BROADCAST_ADDRESS);
		list_remove(C_SERVICE_served_streams, stream);
		FREE(stream);
		return;
	}

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Processing stream of service \"%s\" to %s...\n", stream->service_id, addr_to_str(stream->to));
	send_SERVICE_SERVER_STREAM_ITEM(stream->to, service, stream->offset, stream->size);
	if((C_time() + stream->interval) <= stream->end_time) {
		callback_add(process_SERVICE_CLIENT_STREAM_REQUEST, stream, stream->interval);
	} else {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Stream of service \"%s\" to %s has ended.\n", stream->service_id, addr_to_str(stream->to));
		list_remove(C_SERVICE_served_streams, stream);
		FREE(stream);
	}
}

/**
 * @brief Check whether a two streams match.
 *
 * @param item Stream to check.
 * @param cmp_item Reference stream.
 * @return <code>true</code> if the two streams match, <code>false</code>
 * otherwise.
 */
static bool matches_stream(const void* item, const void* cmp_item) {
	SERVICE_stream_t* stream = (SERVICE_stream_t*) item;
	SERVICE_stream_t* cmp_stream = (SERVICE_stream_t*) cmp_item;
	return (
		(stream->to == cmp_stream->to)
		&& (strcmp(stream->service_id, cmp_stream->service_id) == 0)
		&& (stream->offset == cmp_stream->offset)
		&& (stream->size == cmp_stream->size)
		&& (stream->interval == cmp_stream->interval)
	);
}

/**
 * @brief Handle incoming SERVICE STREAM REQUEST packet.
 *
 * Schedule a stream for processing if the service is available and additional
 * streams can be supported, send a SERVICE NOT AVAILABLE packet otherwise.
 *
 * @param header Header of incoming packet.
 * @param request Stream request information.
 */
void handle_SERVICE_CLIENT_STREAM_REQUEST(const net_header_t* header, const SERVICE_CLIENT_STREAM_REQUEST_packet_t* request) {
	SERVICE_service_t* service = get_service(request->service_id);
	SERVICE_stream_t cmp_stream;
	SERVICE_stream_t* stream = NULL;
	unsigned int expected_stream_size;

	C_SERVICE_received_bytes += sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t);
	add_bytes_to_datarate_stat(C_SERVICE_received_datarate_stat, sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t));

	if(service == NULL) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT STREAM REQUEST from %s, but service \"%s\"is not available. Notifying client...\n", addr_to_str(header->source_address), request->service_id);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, request->service_id, C_BROADCAST_ADDRESS);
		return;
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	} else if(is_redirecting_service(service)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT STREAM REQUEST from %s, but service \"%s\" is not active. Redirecting client to %s...\n", addr_to_str(header->source_address), request->service_id, addr_to_str(service->redirect_addr));
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, request->service_id, service->redirect_addr);
		return;
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
	} else if(!is_running_service(service) && !is_locked_service(service)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received CLIENT STREAM REQUEST from %s, but service \"%s\" is not active (%d). Notifying client...\n", addr_to_str(header->source_address), request->service_id, service->state);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, request->service_id, C_BROADCAST_ADDRESS);
		return;
	} else if(list_size(C_SERVICE_served_streams) >= SERVICE_MAX_SERVED_STREAMS) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE:  Received CLIENT STREAM REQUEST, but there are no empty slots for stream processing.\n");
		send_SERVICE_SERVER_NOT_AVAILABLE(header->source_address, request->service_id, C_BROADCAST_ADDRESS);
		return;
	} else if(service->data_size < (request->offset + request->size)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Received CLIENT STREAM REQUEST for %d byte(s) at 0x%X of service \"%s\" from %s, but request size exceeds service data (%d < (%d + %d)). Dropping invalid stream request...\n", request->size, request->offset, request->service_id, addr_to_str(header->source_address), service->data_size, request->offset, request->size);
		return;
	}

	C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received CLIENT STREAM REQUEST for %d byte(s) at 0x%X of service \"%s\" at interval %.2fs for a duration of %.3fs from %s.\n", request->size, request->offset, request->service_id, request->interval, request->duration, addr_to_str(header->source_address));
	memset(&cmp_stream, 0, sizeof(SERVICE_stream_t));
	cmp_stream.to = header->source_address;
	copy_service_id(cmp_stream.service_id, request->service_id);
	cmp_stream.offset = request->offset;
	cmp_stream.size = request->size;
	cmp_stream.interval = request->interval;
	if(list_find(C_SERVICE_served_streams, &cmp_stream, matches_stream)) {
		C_DEBUG(SERVICE_SERVER_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: A matching CLIENT STREAM REQUEST is already being served, ignoring request.\n");
		return;
	}

	// Initialize stream.
	MALLOC(stream, SERVICE_stream_t);
	stream->to = header->source_address;
	copy_service_id(stream->service_id, request->service_id);
	stream->offset = request->offset;
	stream->size = request->size;
	stream->interval = request->interval;
	stream->end_time = request->duration + C_time();
	list_add(C_SERVICE_served_streams, stream);
	callback_add(process_SERVICE_CLIENT_STREAM_REQUEST, stream, 0);

	expected_stream_size = request->size * ((unsigned int) (request->duration / request->interval)) + 1;
	stat_service_request(service, expected_stream_size, header);
}
