/**
 * @file service_stats.c
 *
 * @brief Service statistics functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_stats.c 3475 2010-04-26 15:07:35Z gwitten $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/net.h"
#include "../util/mapping.h"
#include "../util/trigger.h"
#include "../rudp-light/rudp-light.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_management.h"
#include "service_stats.h"
#include "service_util.h"



/**
 * @brief Debug level for the stats component.
 */
#define SERVICE_STATS_DEBUG_LEVEL	C_MSG_WARN

//#define COST_METRIC_REQUEST_COUNT	///< Use request count (data, update or stream) as cost metric.
//#define COST_METRIC_TRAFFIC		///< Use traffic as cost metric.
#define COST_METRIC_DATARATE		///< Use datarate as cost metric.

/**
 * @brief Maximal number of transmission failures of statistics to the
 * coordinator.
 */
#define SERVICE_STATS_MAX_TRANSMISSION_FAILURES		3



/**
 * @brief Check whether the node addresses of two host statistics data
 * structures match.
 *
 * @param item Host statistic item to check.
 * @param cmp_item Reference host statistic item.
 * @return <code>true</code> if the address of the two host statistic items
 * matches, <code>false</code> otherwise.
 */
bool matches_host_stat_address(const void* item, const void* cmp_item) {
	SERVICE_host_stat_t* host_stat = (SERVICE_host_stat_t*) item;
	SERVICE_host_stat_t* cmp_host_stat = (SERVICE_host_stat_t*) cmp_item;
	return matches_node_address(host_stat->node, cmp_host_stat->node);
}

/**
 * @brief Allocate and initialize statistics for a service host / instance.
 *
 * @param host_stat Pointer to pointer to host statistics structure to
 * initialize.
 * @param address Address of host node.
 */
void alloc_host_stat(SERVICE_host_stat_t** host_stat, C_addr_t address) {
	MALLOC(*host_stat, SERVICE_host_stat_t);
	alloc_datarate_stat(&(*host_stat)->datarate_stat);
	// Set host stat to support aggregation of other stats.
	(*host_stat)->datarate_stat->may_aggregate_other_stats = true;
	init_host_stat(*host_stat, address);
}

/**
 * @brief Initialize statistics for a service host / instance.
 *
 * @param host_stat Pointer to host statistics structure to initialize.
 * @param address Address of host node.
 */
void init_host_stat(SERVICE_host_stat_t* host_stat, C_addr_t address) {
	host_stat->node = get_known_node(address);
	host_stat->request_count = 0;
	host_stat->traffic = 0;
	init_datarate_stat(host_stat->datarate_stat);
}

/**
 * @brief Free host statistics.
 *
 * @param host_stat Pointer to pointer to host statistics structure to free.
 */
void free_host_stat(SERVICE_host_stat_t** host_stat) {
	free_host_stat_list(*host_stat);
	*host_stat = NULL;
}

/**
 * @brief Free host statistics (list interface).
 *
 * @param p Pointer to host statistics structure to free.
 */
void free_host_stat_list(void* p) {
	SERVICE_host_stat_t* host_stat = (SERVICE_host_stat_t*) p;
	// Don't free the node; it's contained in C_known_nodes.
	free_datarate_stat(&host_stat->datarate_stat);
	FREE(host_stat);
}

/**
 * @brief Calculate service cost based on host statistics.
 *
 * @param host_stat Pointer to host statistics data structure.
 * @return Cost of providing a service given the host statistics.
 */
float host_stat_to_cost(SERVICE_host_stat_t* host_stat) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	uint32_t datarate = calculate_datarate(host_stat->datarate_stat);

	if(
		((host_stat->request_count == 0)
			&& (host_stat->traffic != 0))
		|| ((host_stat->request_count != 0)
			&& (host_stat->traffic == 0))
// No need to check against the data rate. It will necessarily get out of sync
// as stats are sent to the coordinator.
//		|| ((host_stat->request_count == 0)
//			&& (host_stat->traffic == 0)
//			&& (datarate != 0))
//		|| ((host_stat->request_count != 0)
//			&& (host_stat->traffic != 0)
//			&& (datarate == 0))
	)
			C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Host statistics of %s might be inconsistent (%d, %d, %d).\n", addr_to_str(host_stat->node->address), host_stat->request_count, host_stat->traffic, datarate);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#if defined(COST_METRIC_REQUEST_COUNT)
	return (float) host_stat->request_count;
#elif defined(COST_METRIC_TRAFFIC)
	return (float) host_stat->traffic;
#elif defined(COST_METRIC_DATARATE)
	return (float) calculate_datarate(host_stat->datarate_stat);
#else
	#error No cost metric defined (available cost metrics are: COST_METRIC_REQUEST_COUNT, COST_METRIC_TRAFFIC and COST_METRIC_DATARATE).
#endif	// COST_METRIC_*
}


/**
 * @brief Check whether the node addresses of two node statistics data
 * structures match.
 *
 * @param item Node statistic item to check.
 * @param cmp_item Reference node statistic item.
 * @return <code>true</code> if the address of the two node statistic items
 * matches, <code>false</code> otherwise.
 */
bool matches_node_stat_address(const void* item, const void* cmp_item) {
	SERVICE_node_stat_t* node_stat = (SERVICE_node_stat_t*) item;
	SERVICE_node_stat_t* cmp_node_stat = (SERVICE_node_stat_t*) cmp_item;
	return matches_node_address(node_stat->node, cmp_node_stat->node);
}

/**
 * @brief Initialize node statistics.
 *
 * NOTE: Optionally, this function does NOT initialize the entire note
 * statistics structure, but rather only those fields, that have a
 * time-dependant average value, e.g., fields that increase monotonously over
 * time. These fields need to be reset after sending stats to the coordinator
 * node, as otherwise they would be counted multiple times.
 *
 * @param node_stat Pointer to pointer to node statistics structure to
 * initialize.
 * @param init_time_independant_fields Flag whether to initialize
 * time-independant fields, or not.
 */
void init_node_stat(SERVICE_node_stat_t* node_stat, bool init_time_independant_fields) {
	node_stat->request_count = 0;
	node_stat->traffic = 0;
	if(init_time_independant_fields)
		init_datarate_stat(node_stat->datarate_stat);
	node_stat->last_hop_request_count = 0;
	node_stat->last_hop_traffic = 0;
	if(init_time_independant_fields)
		init_datarate_stat(node_stat->last_hop_datarate_stat);
	node_stat->last_hop_hop_count_running_sum = 0;
	node_stat->requests_forwarded = 0;
	node_stat->last_hop_of_last_request = C_BROADCAST_ADDRESS;
	node_stat->refused_migration = false;
}

/**
 * @brief Allocate and initialize node statistics.
 *
 * @param node_stat Pointer to pointer to node statistics structure to
 * initialize.
 * @param address Address of the corresponding node.
 */
void alloc_node_stat(SERVICE_node_stat_t** node_stat, C_addr_t address) {
	MALLOC(*node_stat, SERVICE_node_stat_t);
	(*node_stat)->node = register_known_node(address);
	alloc_datarate_stat(&(*node_stat)->datarate_stat);
	alloc_datarate_stat(&(*node_stat)->last_hop_datarate_stat);
	init_node_stat(*node_stat, true);
}

/**
 * @brief Free node statistics.
 *
 * @param node_stat Pointer to pointer to node statistics structure to free.
 */
void free_node_stat(SERVICE_node_stat_t** node_stat) {
	free_node_stat_list(*node_stat);
	*node_stat = NULL;
}

/**
 * @brief Free node statistics (list interface).
 *
 * @param p Pointer to node statistics structure to free.
 */
void free_node_stat_list(void* p) {
	SERVICE_node_stat_t* node_stat = (SERVICE_node_stat_t*) p;
	// Don't free the node; it's contained in C_known_nodes.
	free_datarate_stat(&node_stat->datarate_stat);
	free_datarate_stat(&node_stat->last_hop_datarate_stat);
	FREE(node_stat);
}

/**
 * @brief Check whether the node statistics appear to be consistent.
 *
 * @param node_stat Pointer to node statistics data structure.
 * @return <code>true</code> if the node statistics appear to be consistent,
 * <code>false</code> otherwise.
 */
#ifdef C_ENABLE_CONSISTENCY_CHECKS
static bool check_node_stat(SERVICE_node_stat_t* node_stat) {
// No need to check against the data rate. It will necessarily get out of sync
// as stats are sent to the coordinator.
//	uint32_t datarate = calculate_datarate(node_stat->datarate_stat);

	if(
		((node_stat->request_count == 0)
			&& (node_stat->traffic != 0))
		|| ((node_stat->request_count != 0)
			&& (node_stat->traffic == 0))
//		|| ((node_stat->request_count == 0)
//			&& (node_stat->traffic == 0)
//			&& (datarate != 0))
//		|| ((node_stat->request_count != 0)
//			&& (node_stat->traffic != 0)
//			&& (datarate == 0))
	) {
		return false;
	}
	return true;
}
#endif	// C_ENABLE_CONSISTENCY_CHECKS

/**
 * @brief Calculate service cost based on node statistics.
 *
 * @param node_stat Pointer to node statistics data structure.
 * @return Cost of providing a service given the node statistics.
 */
float node_stat_to_cost(SERVICE_node_stat_t* node_stat) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	if(!check_node_stat(node_stat))
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to calculate cost for node statistics of %s, but the data might be inconsistent (%d, %d, %d).\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat));
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#if defined(COST_METRIC_REQUEST_COUNT)
	return (float) node_stat->request_count;
#elif defined(COST_METRIC_TRAFFIC)
	return (float) node_stat->traffic;
#elif defined(COST_METRIC_DATARATE)
	return (float) calculate_datarate(node_stat->datarate_stat);
#else
	#error No cost metric defined (available cost metrics are: COST_METRIC_REQUEST_COUNT, COST_METRIC_TRAFFIC and COST_METRIC_DATARATE).
#endif	// COST_METRIC_*
}

/**
 * @brief Add a node statistics data structure to a host statistics data
 * structure.
 *
 * @param host_stat Pointer to host statistics struct.
 * @param node_stat Pointer to node statistics struct.
 */
void add_node_stat_to_host_stat(SERVICE_host_stat_t* host_stat, SERVICE_node_stat_t* node_stat) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	float cost_before_adding_node_stat = SERVICE_host_stat_to_cost(host_stat);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	host_stat->request_count += node_stat->request_count;
	host_stat->traffic += node_stat->traffic;
	add_datarate_stats(host_stat->datarate_stat, node_stat->datarate_stat);

#ifdef C_ENABLE_CONSISTENCY_CHECKS
#ifdef ARCH_NS2
	C_ABORT_ON(SERVICE_host_stat_to_cost(host_stat) < cost_before_adding_node_stat, "SERVICE: Cost for host %s has decreased while adding %s (%.1f < %.1f)...\n", addr_to_str(host_stat->node->address), addr_to_str(node_stat->node->address), SERVICE_host_stat_to_cost(host_stat), cost_before_adding_node_stat);
#else
	// The cost of a cluster may decrease due to processing time, because
	// the data rate is measured using a running average.
	if(SERVICE_host_stat_to_cost(host_stat) < cost_before_adding_node_stat)
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Cost for host %s has decreased while adding %s (%.1f < %.1f)...\n", addr_to_str(host_stat->node->address), addr_to_str(node_stat->node->address), SERVICE_host_stat_to_cost(host_stat), cost_before_adding_node_stat);
#endif	// ARCH_NS2
#endif	// C_ENABLE_CONSISTENCY_CHECKS
}

/**
 * @brief Subtract a node statistics data structure to a host statistics data
 * structure.
 *
 * @param host_stat Pointer to host statistics struct.
 * @param node_stat Pointer to node statistics struct.
 */
void subtract_node_stat_to_host_stat(SERVICE_host_stat_t* host_stat, SERVICE_node_stat_t* node_stat) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	float cost_before_removing_node_stat = SERVICE_host_stat_to_cost(host_stat);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	host_stat->request_count -= node_stat->request_count;
	host_stat->traffic -= node_stat->traffic;
	subtract_datarate_stats(host_stat->datarate_stat, node_stat->datarate_stat);

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(SERVICE_host_stat_to_cost(host_stat) > cost_before_removing_node_stat, "SERVICE: Cost for host %s has increased while removing %s (%.1f > %.1f)...\n", addr_to_str(host_stat->node->address), addr_to_str(node_stat->node->address), SERVICE_host_stat_to_cost(host_stat), cost_before_removing_node_stat);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
}

/**
 * @brief Reset statistics of a service.
 *
 * @param service_id Service ID of the service whose stats are to be reset.
 */
void reset_service_stats(SERVICE_id_t* service_id) {
	SERVICE_service_t* service = get_service(service_id);
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;

	if(service == NULL) {
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to reset statistics of service \"%s\", but service not active.\n", service_id);
		return;
	}

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Resetting statistics of service \"%s\"...\n", service_id);
	service->host_stat->request_count = 0;
	service->host_stat->traffic = 0;
	init_datarate_stat(service->host_stat->datarate_stat);

	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL)
		init_node_stat(node_stat, true);
}



/**
 * @brief Do statistics for a service request.
 *
 * @param service Pointer to service for which statistics are to be counted.
 * @param request_size Size of the data request in bytes.
 * @param header Pointer to header of service request packet.
 */
void stat_service_request(SERVICE_service_t* service, size_t request_size, const net_header_t* header) {
	SERVICE_node_stat_t* node_stat;
	SERVICE_node_stat_t cmp_node_stat;
	node_t cmp_node;
	list_iterator_t node_stats_iterator;
#ifdef C_USE_NETWORK_MAPPING
	unsigned int i = 0;
#endif	// C_USE_NETWORK_MAPPING
	unsigned int clients = 0;

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Processing statistics for request from %s over %d hop(s) via %s for %d byte(s) of \"%s\"...\n", addr_to_str(header->source_address), header->hop_count, addr_to_str(header->last_hop_address), request_size, service->id);

	// Update statistics of the source node.
	memset(&cmp_node_stat, 0, sizeof(SERVICE_node_stat_t));
	memset(&cmp_node, 0, sizeof(node_t));
	cmp_node_stat.node = &cmp_node;
	cmp_node_stat.node->address = header->source_address;
	if((node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, matches_node_stat_address)) == NULL) {
		alloc_node_stat(&node_stat, header->source_address);
		list_add(service->node_stats, node_stat);
	}
	node_stat->request_count++;
	node_stat->traffic += request_size;
	add_bytes_to_datarate_stat(node_stat->datarate_stat, request_size);
	node_stat->last_hop_of_last_request = header->last_hop_address;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	if(!check_node_stat(node_stat))
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to update node statistics of %s, but the data might be inconsistent (%d, %d, %d).\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat));
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Service data rate from %s is %d B/s after %d B in %d requests.\n", addr_to_str(node_stat->node->address), calculate_datarate(node_stat->datarate_stat), node_stat->traffic, node_stat->request_count);

	// Update statistics of the last forwarding hop.
	cmp_node_stat.node->address = header->last_hop_address;
	if((node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, matches_node_stat_address)) == NULL) {
		alloc_node_stat(&node_stat, header->last_hop_address);
		list_add(service->node_stats, node_stat);
	}
	node_stat->last_hop_request_count++;
	node_stat->last_hop_traffic += request_size;
	add_bytes_to_datarate_stat(node_stat->last_hop_datarate_stat, request_size);
	node_stat->last_hop_hop_count_running_sum += header->hop_count;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	if(!check_node_stat(node_stat))
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to update node statistics of last hop %s, but the data might be inconsistent (%d, %d, %d).\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat));
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Service data rate from last hop %s is %d B/s after %d B in %d requests.\n", addr_to_str(node_stat->node->address), calculate_datarate(node_stat->last_hop_datarate_stat), node_stat->last_hop_traffic, node_stat->last_hop_request_count);

#ifdef C_USE_NETWORK_MAPPING
	// Update statistics of nodes on the path.
	for(i = 0; i < header->hop_count; i++) {
		cmp_node_stat.node->address = header->hops[i];
		if((node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, matches_node_stat_address)) == NULL) {
			alloc_node_stat(&node_stat, header->hops[i]);
			list_add(service->node_stats, node_stat);
		}
		node_stat->requests_forwarded++;
	}
#endif	// C_USE_NETWORK_MAPPING

	// Update statistics of the service host.
	service->host_stat->request_count++;
	service->host_stat->traffic += request_size;
	add_bytes_to_datarate_stat(service->host_stat->datarate_stat, request_size);

	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		if(node_stat_to_cost(node_stat) > 0)
			clients++;
	}

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Service data rate on this host is %d B/s after %d B in %d requests from %d clients.\n", calculate_datarate(service->host_stat->datarate_stat), service->host_stat->traffic, service->host_stat->request_count, clients);

	// Trigger placement algorithm if service is running.
	if(is_running_service(service))
		PLACEMENT_trigger_algorithm(service, C_PLACEMENT_TRIGGER_REASON_DATA, &header->source_address);
}



/*
 * Packet handling functions.
 */

/**
 * @brief RUDP callback for a SERVICE STATS packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_SERVICE_STATS_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) result->parameter;
	SERVICE_service_t* service = get_service(service_id);

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: SERVICE STATS for service \"%s\" reached the destination node, resetting node stats...\n", service_id);
		if(service != NULL)
			service->failed_stat_transmissions = 0;
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: SERVICE STATS of service \"%s\" timed out.\n", service_id);
		if(service != NULL) {
			service->failed_stat_transmissions++;
			if(service->failed_stat_transmissions >= SERVICE_STATS_MAX_TRANSMISSION_FAILURES) {
				C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Transmission of STATS for service \"%s\" to coordinator failed for %d consecutive times. Is something wrong with the coordinator?\n", service_id, service->failed_stat_transmissions);
				// TODO: Establish a new coordinator if the old one is missing.
			}
		}
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Sending SERVICE STATS of service \"%s\" aborted.\n", service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: SERVICE STATS of service \"%s\" hit full queue.\n", service_id);
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Unhandled error while sending SERVICE STATS with RUDP (%d).\n", result->error_code);
	}

	FREE(service_id);
	FREE(result);
}

/**
 * @brief Send a SERVICE STATS packet.
 *
 * The size of a STATS packet is a follows (note that sizeof(C_addr_t) is
 * architecture dependant):
 *
 *	packet_size	= sizeof(net_header_t)
 *			+ ((C_MAXIMAL_NETWORK_DIAMETER + 1 + C_MAXIMAL_NODE_DEGREE) * sizeof(C_addr_t))
 *			+ sizeof(RUDP_DATA_packet_t)
 *			+ (number_of_nodes * sizeof(SERVICE_condensed_node_stat_t))
 *			+ ((number_of_nodes / 8 + 1) * number_of_nodes)
 *			= 36 + 164 + 2
 *			+ (number_of_nodes * 16)
 *			+ ((number_of_nodes / 8 + 1) * number_of_nodes)
 *			= 1/8 * number_of_nodes ^ 2 + 17 * number_of_nodes + 202
 *			= [Excel =((1/8) * $A1 ^ 2) + (17 * $A1) + 202 ]
 *
 *	nodes	packet_size
 *	16	506
 *	25	705,125
 *	36	976
 *	49	1.335,125
 *	50	1.364,5
 *	51	1.394,125
 *	52	1.424
 *	53	1.454,125
 *	54	1.484,5		<= Maximal number of nodes for MTU = 1.500
 *	55	1.515,125
 *	64	1.802
 *	81	2.399,125
 *	100	3.152
 *	120	4.042
 *	121	4.089,125	<= Maximal number of nodes for MTU = 4.096
 *	122	4.136,5
 *
 * If the node stats and the network map were sent seperately, this calculation
 * would change as follows:
 *
 *	stats_size	= sizeof(net_header_t)
 *			+ ((C_MAXIMAL_NETWORK_DIAMETER + 1 + C_MAXIMAL_NODE_DEGREE) * sizeof(C_addr_t))
 *			+ sizeof(RUDP_DATA_packet_t)
 *			+ (number_of_nodes * sizeof(SERVICE_condensed_node_stat_t))
 *			= 36 + 164 + 2
 *			+ (number_of_nodes * 16)
 *			= 16 * number_of_nodes + 202
 *			= [Excel =(16 * $A1) + 202 ]
 *
 *	map_size	= sizeof(net_header_t)
 *			+ ((C_MAXIMAL_NETWORK_DIAMETER + 1 + C_MAXIMAL_NODE_DEGREE) * sizeof(C_addr_t))
 *			+ sizeof(RUDP_DATA_packet_t)
 *			+ ((number_of_nodes / 8 + 1) * number_of_nodes)
 *			= 36 + 164 + 2
 *			+ ((number_of_nodes / 8 + 1) * number_of_nodes)
 *			= 1/8 * number_of_nodes ^ 2 + number_of_nodes + 202
 *			= [Excel =((1/8) * $A1 ^ 2) + $A1 + 202 ]
 *
 *	nodes	stats_size	map_size
 *	16	458		250
 *	25	602		305,125
 *	36	778		400
 *	49	986		551,125
 *	50	1.002		564,5
 *	51	1.018		578,125
 *	52	1.034		592
 *	53	1.050		606,125
 *	54	1.066		620,5
 *	55	1.082		635,125
 *	64	1.226		778
 *	80	1.482		1.082
 *	81	1.498		1.103,125	<= Maximal number of nodes for MTU = 1.500
 *	82	1.514		1.124,5
 *	100	1.802		1.552
 *	170	2.922		3.984,5
 *	171	2.938		4.028,125
 *	172	2.954		4.072		<= Maximal number of nodes for MTU = 4.096
 *	173	2.970		4.116,125
 *
 * @param service Pointer to service whose stats are to be send.
 * @param destination_address Address of the destination node.
 */
static void send_SERVICE_STATS(SERVICE_service_t* service, C_addr_t destination_address) {
	SERVICE_STATS_packet_t stats;
	uint16_t number_of_nodes = get_number_of_known_nodes();
#ifdef C_USE_NETWORK_MAPPING
	size_t condensed_network_map_size = (number_of_nodes / 8 + 1) * number_of_nodes;
#else
	size_t condensed_network_map_size = 0;
#endif	// C_USE_NETWORK_MAPPING
	SERVICE_condensed_node_stat_t* node_stat_array = NULL;
	char* condensed_network_map = NULL;
	char* payload = NULL;
	uint16_t node_id = 0;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
#ifdef C_USE_NETWORK_MAPPING
	uint16_t src_id, dst_id;
	char value = 0;
#endif	// C_USE_NETWORK_MAPPING
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Sending STATS of service \"%s\" to %s...\n", service->id, addr_to_str(destination_address));
	memset(&stats, 0, sizeof(SERVICE_STATS_packet_t));
	copy_service_id(stats.service_id, service->id);
	stats.node_stat_array_size = number_of_nodes * sizeof(SERVICE_condensed_node_stat_t);
	stats.number_of_nodes = number_of_nodes;
	if(
		(service->last_failed_migration_destination != C_BROADCAST_ADDRESS)
		|| ((C_time() - service->start_time) < SERVICE_INSTANCE_VALIDITY_TIME)
	) {
		stats.valid_instance = true;
		stats.last_failed_migration_destination = service->last_failed_migration_destination;	// We omit the check for (last_failed_migration_destination == C_BROADCAST_ADDRESS) since the result is the same.
	} else {
		stats.valid_instance = false;
		stats.last_failed_migration_destination = C_BROADCAST_ADDRESS;
	}

	MALLOC_ARRAY(payload, char, (stats.node_stat_array_size + condensed_network_map_size));
	memset(payload, 0, (stats.node_stat_array_size + condensed_network_map_size));
	node_stat_array = (SERVICE_condensed_node_stat_t*) payload;
	condensed_network_map = payload + stats.node_stat_array_size;

	// Write node statististics.
	// TODO: Reduce size of stats packets.
	for(node_id = 0; node_id < number_of_nodes; node_id++) {
		node_stat_array[node_id].node_address = node_id_to_address(node_id);
		node_stat_array[node_id].request_count = 0;
		node_stat_array[node_id].traffic = 0;
		node_stat_array[node_id].datarate = 0;

		list_iterator_init(service->node_stats, &node_stats_iterator);
		while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
			if(node_stat->node->address == node_stat_array[node_id].node_address) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
				if(!check_node_stat(node_stat))
					C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to add node statistics of %s to array in STATS packet, but the data might be inconsistent (%d, %d, %d).\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat));
#endif	// C_ENABLE_CONSISTENCY_CHECKS

				node_stat_array[node_id].request_count = node_stat->request_count;
				node_stat_array[node_id].traffic = node_stat->traffic;
				node_stat_array[node_id].datarate = calculate_datarate(node_stat->datarate_stat);
				break;
			}
		}

		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Added statistics for %s (%d, %d, %d)...\n", addr_to_str(node_id_to_address(node_id)), node_stat_array[node_id].request_count, node_stat_array[node_id].traffic, node_stat_array[node_id].datarate);
	}

#ifdef C_USE_NETWORK_MAPPING
	// Write network map.
	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_VERBOSE, "SERVICE: Condensed network map (sent) is:\n");
	for(src_id = 0; src_id < number_of_nodes; src_id++) {
		value = (1 << (src_id % 8));
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_VERBOSE, "SERVICE: %s\t", addr_to_str(node_id_to_address(src_id)));
		for(dst_id = 0; dst_id < number_of_nodes; dst_id++) {
			if(exists_directly_known_link(node_id_to_address(src_id), node_id_to_address(dst_id))) {
				C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, (C_MSG_VERBOSE | C_MSG_CONT), " 1");
				condensed_network_map[((src_id / 8) * number_of_nodes) + dst_id] |= value;
			} else {
				C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, (C_MSG_VERBOSE | C_MSG_CONT), " 0");
			}
		}
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, (C_MSG_VERBOSE | C_MSG_CONT), "\n");
	}
#endif	// C_USE_NETWORK_MAPPING

	MALLOC(result, RUDP_transmission_result_t);
	SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, service->id);
	RUDP_sendto2(destination_address, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_STATS, &stats, sizeof(SERVICE_STATS_packet_t), payload, (stats.node_stat_array_size + condensed_network_map_size), send_SERVICE_STATS_callback, result);
	FREE(payload);
}

/**
 * @brief Handle an incoming SERVICE STATS packet.
 *
 * @param header Pointer to the packet header.
 * @param stats Pointer to the placement stats.
 * @param payload Pointer to array of commands.
 * @param payload_size Size of command array in bytes.
 */
void handle_SERVICE_STATS(const net_header_t* header, const SERVICE_STATS_packet_t* stats, char* payload, size_t payload_size) {
	SERVICE_service_t* service = get_service(stats->service_id);
	SERVICE_condensed_node_stat_t* node_stat_array = (SERVICE_condensed_node_stat_t*) payload;
	uint16_t remote_node_id = 0;
	SERVICE_node_stat_t* node_stat;
	SERVICE_node_stat_t cmp_node_stat;
	node_t cmp_node;
#ifdef C_USE_NETWORK_MAPPING
	char* condensed_network_map = payload + stats->node_stat_array_size;
	uint16_t remote_src_id, remote_dst_id;
	char value = 0;
	C_addr_t src_address, dst_address;
#endif	// C_USE_NETWORK_MAPPING
	PLACEMENT_trigger_reason_stats_parameter_t trigger_reason_stats_parameter;

	C_ABORT_ON((stats->node_stat_array_size / sizeof(SERVICE_condensed_node_stat_t)) != stats->number_of_nodes, "Received STATS is inconsistent about the number of nodes (%d != %d).\n", (stats->node_stat_array_size / sizeof(SERVICE_condensed_node_stat_t)), stats->number_of_nodes);

	if(service == NULL) {
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received STATS of service \"%s\" hosted on %s, but service is not hosted on this node. Shutting down apparently left-over service instance...\n", stats->service_id, addr_to_str(header->source_address));
		PLACEMENT_shutdown_remote_service_host(stats->service_id, header->source_address);
		return;
	}

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Received STATS of service \"%s\" hosted on %s.\n", stats->service_id, addr_to_str(header->source_address));

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Updating statistics of service \"%s\" based on data received from %s...\n", stats->service_id, addr_to_str(header->source_address));

	// Update node statististics.
	for(remote_node_id = 0; remote_node_id < stats->number_of_nodes; remote_node_id++) {

		// Add node to network mapping.
		register_known_node(node_stat_array[remote_node_id].node_address);

		// Add node to service statistics.
		if(
			(node_stat_array[remote_node_id].request_count > 0)
			|| (node_stat_array[remote_node_id].traffic > 0)
			|| (node_stat_array[remote_node_id].datarate > 0)
		) {
			C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Adding statistics data (%d, %d, %d) to %s...\n", node_stat_array[remote_node_id].request_count, node_stat_array[remote_node_id].traffic, node_stat_array[remote_node_id].datarate, addr_to_str(node_stat_array[remote_node_id].node_address));

			memset(&cmp_node_stat, 0, sizeof(SERVICE_node_stat_t));
			memset(&cmp_node, 0, sizeof(node_t));
			cmp_node_stat.node = &cmp_node;
			cmp_node_stat.node->address = node_stat_array[remote_node_id].node_address;
			if((node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, matches_node_stat_address)) == NULL) {
				alloc_node_stat(&node_stat, node_stat_array[remote_node_id].node_address);
				list_add(service->node_stats, node_stat);
			}
			node_stat->request_count += node_stat_array[remote_node_id].request_count;
			node_stat->traffic += node_stat_array[remote_node_id].traffic;
			set_additional_datarate_for_datarate_stats(node_stat->datarate_stat, node_stat_array[remote_node_id].datarate);
#ifdef C_ENABLE_CONSISTENCY_CHECKS
			if(!check_node_stat(node_stat))
				C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to read node statistics of %s from array in STATS packet, but the data might be inconsistent (%d, %d, %d).\n", addr_to_str(node_stat->node->address), node_stat->request_count, node_stat->traffic, calculate_datarate(node_stat->datarate_stat));
#endif	// C_ENABLE_CONSISTENCY_CHECKS
			node_stat = NULL;
		}
	}

#ifdef C_USE_NETWORK_MAPPING
	// Update network map.
	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_VERBOSE, "SERVICE: Condensed network map (received) is:\n");
	for(remote_src_id = 0; remote_src_id < stats->number_of_nodes; remote_src_id++) {

		src_address = node_stat_array[remote_src_id].node_address;
		value = (1 << (remote_src_id % 8));

		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_VERBOSE, "SERVICE: %s\t", addr_to_str(src_address));
		for(remote_dst_id = 0; remote_dst_id < stats->number_of_nodes; remote_dst_id++) {
			if(condensed_network_map[((remote_src_id / 8) * stats->number_of_nodes) + remote_dst_id] & value) {
				C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, (C_MSG_VERBOSE | C_MSG_CONT), " 1");

				dst_address = node_stat_array[remote_dst_id].node_address;
				// Make sure that overly fresh network map entries are not redistributed.
				if(!exists_known_link(src_address, dst_address)) {
					register_reportedly_known_link(src_address, dst_address);
				}
			} else {
				C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, (C_MSG_VERBOSE | C_MSG_CONT), " 0");
			}
		}
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, (C_MSG_VERBOSE | C_MSG_CONT), "\n");
	}
#endif	// C_USE_NETWORK_MAPPING

	// Trigger placement algorithm.
	trigger_reason_stats_parameter.source_address = header->source_address;
	trigger_reason_stats_parameter.valid_instance = stats->valid_instance;
	trigger_reason_stats_parameter.last_failed_migration_destination = stats->last_failed_migration_destination;
	PLACEMENT_trigger_algorithm(service, C_PLACEMENT_TRIGGER_REASON_STATS, &trigger_reason_stats_parameter);
}

/**
 * @brief Send the local statistics to another node.
 *
 * @param service_id Service ID of the service whose stats are to be send.
 * @param destination_address Address of the destination node.
 */
void send_statistics(SERVICE_id_t* service_id, C_addr_t destination_address) {
	SERVICE_service_t* service = get_service(service_id);

	if(service == NULL) {
		C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to send statistics of service \"%s\" to %s, but service not active.\n", service_id, addr_to_str(destination_address));
		return;
	}

	C_DEBUG(SERVICE_STATS_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending statistics of service \"%s\" to %s...\n", service_id, addr_to_str(destination_address));
	send_SERVICE_STATS(service, destination_address);
}
