/**
 * @file service_util.c
 *
 * @brief Service utility functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_util.c 3177 2010-01-05 16:43:45Z wittenbu $
 */

#include <string.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../slp-light/slp-light_types.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_util.h"



/**
 * @brief Allocate a service ID.
 *
 * @param new_service_id Pointer to new service ID.
 * @param service_id Service ID used for initialization.
 */
void alloc_service_id(SERVICE_id_t** new_service_id, const SERVICE_id_t* service_id) {
	MALLOC_ARRAY(*new_service_id, char, SERVICE_ID_LENGTH);
#ifdef ARCH_WIN32
	strncpy_s(*new_service_id, SERVICE_ID_LENGTH, service_id, SERVICE_ID_LENGTH);
#else
	strncpy(*new_service_id, service_id, SERVICE_ID_LENGTH);
#endif	// ARCH_WIN32
}

#ifdef C_ENABLE_CONSISTENCY_CHECKS
/**
 * @brief Check a service ID.
 *
 * @param service_id Service ID to be checked.
 */
void check_service_id(const SERVICE_id_t* service_id) {
	C_ABORT_ON(service_id == NULL, "SERVICE: Service ID is NULL.\n");
	C_ABORT_ON(strlen(service_id) == 0, "SERVICE: Service ID \"%s\" has length of zero.\n");
	C_ABORT_ON((strlen(service_id) + 1) > SERVICE_ID_LENGTH, "SERVICE: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SERVICE_ID_LENGTH);
}
#endif	// C_ENABLE_CONSISTENCY_CHECKS

/**
 * @brief Copy a service ID to another.
 *
 * @param dest_service_id Destination service ID.
 * @param src_service_id Source service ID.
 */
void copy_service_id(SERVICE_id_t* dest_service_id, const SERVICE_id_t* src_service_id) {
#ifdef ARCH_WIN32
	strncpy_s(dest_service_id, SERVICE_ID_LENGTH, src_service_id, SERVICE_ID_LENGTH);
#else
	strncpy(dest_service_id, src_service_id, SERVICE_ID_LENGTH);
#endif	// ARCH_WIN32
}

/**
 * @brief Check whether two service IDs match.
 *
 * @param service_id_1 First service ID.
 * @param service_id_2 Seconds service ID.
 * @return <code>true</code> if the two service IDs are equal,
 * <code>false</code> otherwise.
 */
bool same_service_id(const SERVICE_id_t* service_id_1, const SERVICE_id_t* service_id_2) {
	if(strncmp(service_id_1, service_id_2, SERVICE_ID_LENGTH) == 0)
		return true;
	else
		return false;
}

/**
 * @brief Check whether two services have the same service ID (list interface).
 *
 * @param item Service ID to check.
 * @param cmp_item Reference service ID.
 * @return <code>true</code> if the service IDs of the two services are equal,
 * <code>false</code> otherwise.
 */
bool matches_service_id(const void* item, const void* cmp_item) {
	SERVICE_service_t* service = (SERVICE_service_t*) item;
	SERVICE_service_t* cmp_service = (SERVICE_service_t*) cmp_item;
	return same_service_id(service->id, cmp_service->id);
}
