/**
 * @file service_vars.h
 *
 * @brief Service (variables).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_vars.h 3354 2010-02-26 13:30:48Z gwitten $
 */

#ifndef SERVICE_VARS_H
#define SERVICE_VARS_H

#include "../types.h"
#include "service_types.h"



/**
 * @brief Debug level for this component.
 */
#define SERVICE_DEBUG_LEVEL	C_MSG_WARN

/**
 * @brief Maximal number of hosted services.
 */
#define SERVICE_MAX_HOSTED_SERVICES	4

/**
 * @brief Service validity time in seconds.
 *
 * Time during which a service instance will be considered valid, i.e. it won't
 * be shutdown by the coordinator even if it's not part of the current service
 * configuration.
 */
#define SERVICE_INSTANCE_VALIDITY_TIME	60



// Client variables.

/**
 * @brief Statistics counter for total number of data requests.
 */
uint32_t C_SERVICE_total_data_requests;

/**
 * @brief Statistics counter for number of successful data requests.
 */
uint32_t C_SERVICE_successful_data_requests;

/**
 * @brief Statistics counter for number of failed data requests.
 */
uint32_t C_SERVICE_failed_data_requests;

/**
 * @brief Statistics counter for total delay of successful data requests.
 */
double C_SERVICE_total_successful_data_request_delay;

/**
 * @brief Statistics counter for total number of data updates.
 */
uint32_t C_SERVICE_total_data_updates;

/**
 * @brief Statistics counter for number of successful data updates.
 */
uint32_t C_SERVICE_successful_data_updates;

/**
 * @brief Statistics counter for number of failed data updates.
 */
uint32_t C_SERVICE_failed_data_updates;

/**
 * @brief Statistics counter for total number of stream requests.
 */
uint32_t C_SERVICE_total_stream_requests;

/**
 * @brief Statistics counter for number of successful stream requests.
 */
uint32_t C_SERVICE_successful_stream_requests;

/**
 * @brief Statistics counter for number of failed stream requests.
 */
uint32_t C_SERVICE_failed_stream_requests;

/**
 * @brief Statistics counter for number of expected stream packets.
 */
uint32_t C_SERVICE_expected_stream_packets;

/**
 * @brief Statistics counter for number of received stream packets.
 */
uint32_t C_SERVICE_received_stream_packets;

/**
 * @brief Statistics counter for total number of overall service requests.
 */
uint32_t C_SERVICE_total_requests;

/**
 * @brief Statistics counter for number of successful overall service requests.
 */
uint32_t C_SERVICE_successful_requests;

/**
 * @brief Statistics counter for number of failed overall service requests.
 */
uint32_t C_SERVICE_failed_requests;

/**
 * @brief Statistics counter for number of transmitted bytes.
 */
uint32_t C_SERVICE_transmitted_bytes;

/**
 * @brief Statistics counter for number of received bytes.
 */
uint32_t C_SERVICE_received_bytes;

/**
 * @brief Internal statistics struct for transmitted datarate.
 */
datarate_stat_t* C_SERVICE_transmitted_datarate_stat;

/**
 * @brief Statistics counter for transmitted datarate.
 */
uint32_t C_SERVICE_transmitted_datarate;

/**
 * @brief Internal statistics struct for transmitted datarate.
 */
datarate_stat_t* C_SERVICE_received_datarate_stat;

/**
 * @brief Statistics counter for transmitted datarate.
 */
uint32_t C_SERVICE_received_datarate;

/**
 * @brief Array to count received request packets depending on their hop count.
 *
 * The size of the array is C_MAXIMAL_NETWORK_DIAMETER + 1 to make the
 * processing more natural for packets with hop_count = 0.
 */
uint32_t C_SERVICE_request_packets_per_hop_count[C_MAXIMAL_NETWORK_DIAMETER + 1];

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
/**
 * @brief Callback ID for indicating pending requests.
 */
callback_id_t C_SERVICE_indicate_request_callback_id;

/**
 * @brief Callback ID for indicating overall node activity.
 */
callback_id_t C_SERVICE_indicate_activity_callback_id;
#endif	// C_ENABLE_NAM_TRACE_OUTPUT


// Server variables.

/**
 * @brief List of currently hosted services.
 */
list_t* C_SERVICE_hosted_services;

/**
 * @brief List of currently served stream.
 */
list_t* C_SERVICE_served_streams;

/**
 * @brief List of outgoing service replications.
 */
list_t* C_SERVICE_outgoing_replications;

/**
 * @brief List of incoming service replications.
 */
list_t* C_SERVICE_incoming_replications;

/**
 * @brief Number of currently hosted services on this node.
 */
uint32_t C_SERVICE_currently_hosted_services;

/**
 * @brief Running sum of the number of hosted services on this node.
 */
uint32_t C_SERVICE_hosted_services_running_sum;

/**
 * @brief Average (over time) number of hosted services on this node.
 */
double C_SERVICE_average_hosted_services;

/**
 * @brief Time until which a service instace was available on this node.
 */
C_time_t C_SERVICE_available_until_time;

/**
 * @brief Statistics counter for inter-host synchronization overhead in bytes.
 */
uint32_t C_SERVICE_sync_overhead_bytes;

/**
 * @brief Statistics counter for total number of replications.
 */
uint32_t C_SERVICE_total_replications;

/**
 * @brief Statistics counter for number of successful replications.
 */
uint32_t C_SERVICE_successful_replications;

/**
 * @brief Statistics counter for number of failed replications.
 */
uint32_t C_SERVICE_failed_replications;

/**
 * @brief Statistics counter for total number of migrations.
 */
uint32_t C_SERVICE_total_migrations;

/**
 * @brief Statistics counter for number of successful migrations.
 */
uint32_t C_SERVICE_successful_migrations;

/**
 * @brief Statistics counter for number of failed migrations.
 */
uint32_t C_SERVICE_failed_migrations;

#endif	// SERVICE_VARS_H
