/**
 * @file slp-light.c
 *
 * @brief Service Location Protocol (SLP) - minimal implemenation.
 *
 * This is a very basic implementation of the Service Location Protocol (SLP).
 * It covers searching for services by expanding-ring broadcasts. It does not
 * include service naming conventions or Directory Agents (DA).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: slp-light.c 3514 2010-05-13 15:44:29Z gwitten $
 */

#include <math.h>
#include <stdlib.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../service/service.h"
#include "../placement/placement.h"
#include "slp-light.h"
#include "slp-light_types.h"
#include "slp-light_vars.h"



#define SLP_PACKET_TYPE_QUERY		0x0	///< SLP QUERY packet type.
#define SLP_PACKET_TYPE_REPLY		0x1	///< SLP REPLY packet type.
#define SLP_PACKET_TYPE_ANNOUNCE	0x2	///< SLP ANNOUNCE packet type.
#define SLP_PACKET_TYPE_PURGE		0x3	///< SLP PURGE packet type.

/**
 * @brief Maximal supported number of active queries.
 */
#define SLP_MAX_ACTIVE_QUERIES		(C_MAX_SERVICES * 128)

/**
 * @brief Number of retries of SLP queries.
 */
#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	#define SLP_QUERY_TRIES			3
#else
	#define SLP_QUERY_TRIES			16
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

/**
 * @brief Timeout for SLP queries in seconds.
 */
#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	#define SLP_QUERY_TIMEOUT		(2 * C_NETWORK_TRAVERSAL_TIME + C_PATH_DISCOVERY_TIME)
#else
	#define SLP_QUERY_TIMEOUT		(0.2 * C_NETWORK_TRAVERSAL_TIME + C_PATH_DISCOVERY_TIME)
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

/**
 * @brief Timeout for entries to unavailable services list in seconds.
 */
#define SLP_UNAVAILABLE_SERVICE_TIMEOUT	60

/**
 * @brief Use expanding ring search for queries.
 *
 * With each retry of the query the TTL of the query packet is increased, thus
 * implementing an expanding ring search.
 */
#ifndef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	#define SLP_USE_EXPANDING_RING_SEARCH
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

/**
 * @brief Use expanding ring search for queries.
 *
 * With each retry of the query the TTL of the query packet is increased, thus
 * implementing an expanding ring search.
 */
#define SLP_EXPANDING_RING_SEARCH_INITIAL_TTL 2

/**
 * @brief Use cache to store information about services.
 */
#define SLP_USE_CACHE

#ifdef SLP_USE_CACHE
#include "slp_cache.h"
#endif	// SLP_USE_CACHE



/**
 * @brief Use pro-active announcements and purges.
 *
 * Requires SLP_USE_CACHE to be defined.
 */
#if defined (C_ENABLE_SPI_OPTIMIZATIONS) || defined (C_ENABLE_ASG_OPTIMIZATIONS)
	#define SLP_USE_ANNOUNCEMENTS
#endif	// (C_ENABLE_SPI_OPTIMIZATIONS || C_ENABLE_ASG_OPTIMIZATIONS)

#ifdef SLP_USE_ANNOUNCEMENTS
	#ifndef SLP_USE_CACHE
		#error Trying to use SLP announcements without caching.
	#endif	// SLP_USE_CACHE
#endif	// SLP_USE_ANNOUNCEMENTS



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the SLP module.
 */
void SLP_init() {
	C_SLP_active_queries = NULL;
	list_alloc(&C_SLP_active_queries);
	C_SLP_total_queries = 0;
	C_SLP_successful_queries = 0;
	C_SLP_failed_queries = 0;
	C_SLP_service_cache = NULL;
	C_SLP_service_cache = NULL;
	C_SLP_unavailable_services = NULL;
#ifdef SLP_USE_CACHE
	list_alloc(&C_SLP_service_cache);
	list_alloc(&C_SLP_unavailable_services);
#endif	// SLP_USE_CACHE
}

/**
 * @brief Start the SLP module.
 */
void SLP_start() {
}

/**
 * @brief Stop the SLP module.
 */
void SLP_stop() {
	SLP_query_t* query = NULL;

	while(!list_empty(C_SLP_active_queries)) {
		query = (SLP_query_t*) list_head(C_SLP_active_queries);
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Aborting QUERY for service \"%s\"...\n", query->service_id);
		if(query->timeout_callback_id != 0) {
			callback_cancel(query->timeout_callback_id);
			C_SLP_failed_queries++;		// Queries with a callback have not been queued and hence count in the stats.
		}
		query->result->error_code = C_ERROR_ABORTED;
		query->function(query->result);
		list_remove(C_SLP_active_queries, query);
		FREE(query);
	}
}

/**
 * @brief Finalize the SLP module.
 */
void SLP_finalize() {
#ifdef SLP_USE_CACHE
	SLP_service_instance_t* service_instance = NULL;
#endif	// SLP_USE_CACHE

	if(C_check_consistency_on_shutdown) {
		C_ABORT_ON(C_SLP_total_queries != (C_SLP_successful_queries + C_SLP_failed_queries), "SLP: Query counters don't add up (%d != (%d + %d)).\n", C_SLP_total_queries, C_SLP_successful_queries, C_SLP_failed_queries);
	}
	list_free_including_data(&C_SLP_active_queries, free);
#ifdef SLP_USE_CACHE
	list_free_including_data(&C_SLP_service_cache, free);
	while(!list_empty(C_SLP_unavailable_services)) {
		service_instance = (SLP_service_instance_t*) list_head(C_SLP_unavailable_services);
		callback_cancel(service_instance->delayed_removal_callback_id);
		list_remove(C_SLP_unavailable_services, service_instance);
		FREE(service_instance);
	}
	list_free(&C_SLP_unavailable_services);
#endif	// SLP_USE_CACHE
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send a SLP QUERY packet.
 *
 * @param service_id ID of the service to query for.
 * @param current_try Current try of this query.
 */
static void send_SLP_QUERY(const SERVICE_id_t* service_id, uint8_t current_try) {
	net_header_t* header = NULL;
	SLP_QUERY_packet_t query;
#ifdef SLP_USE_EXPANDING_RING_SEARCH
	uint32_t new_ttl = 0;
#endif	// SLP_USE_EXPANDING_RING_SEARCH

	alloc_header(&header, C_BROADCAST_ADDRESS, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SLP, SLP_PACKET_TYPE_QUERY, sizeof(SLP_QUERY_packet_t));
#ifdef SLP_USE_EXPANDING_RING_SEARCH
	new_ttl = SLP_EXPANDING_RING_SEARCH_INITIAL_TTL * pow(2, (current_try - 1));
	if((new_ttl <= C_DEFAULT_TTL) && (new_ttl != 0))
		header->ttl = new_ttl;
	else 
		header->ttl = C_DEFAULT_TTL;
#endif	// SLP_USE_EXPANDING_RING_SEARCH
	memset(&query, 0, sizeof(SLP_QUERY_packet_t));
	SERVICE_copy_service_id(query.service_id, service_id);
	query.current_try = current_try;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Broadcasting QUERY for service \"%s\" with TTL %d (try #%d)...\n", service_id, header->ttl, current_try);
	C_route(header, &query);
	free_header(&header);
}

/**
 * @brief Send a SLP REPLY packet.
 *
 * @param to Destination address of the reply.
 * @param service_id ID of the service that has been queried for.
 */
static void send_SLP_REPLY(C_addr_t to, const SERVICE_id_t* service_id) {
	SLP_REPLY_packet_t reply;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Sending REPLY to query for service \"%s\" to %s...\n", service_id, addr_to_str(to));
	memset(&reply, 0, sizeof(SLP_REPLY_packet_t));
	SERVICE_copy_service_id(reply.service_id, service_id);
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SLP, SLP_PACKET_TYPE_REPLY, &reply, sizeof(SLP_REPLY_packet_t), NULL, NULL);
}

#ifdef SLP_USE_ANNOUNCEMENTS
/**
 * @brief Send a SLP ANNOUNCE packet.
 *
 * @param service_id ID of the service that is to be announced.
 * @param server Server address.
 * @param replace_address Address of node that the server replaces.
 * @param hop_count Distance to the server in hops.
 */
static void send_SLP_ANNOUNCE(const SERVICE_id_t* service_id, C_addr_t server, C_addr_t replace_address, uint8_t hop_count) {
	SLP_ANNOUNCE_packet_t announcement;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Broadcasting ANNOUNCE for service \"%s\"...\n", service_id);
	memset(&announcement, 0, sizeof(SLP_ANNOUNCE_packet_t));
	SERVICE_copy_service_id(announcement.service_id, service_id);
	announcement.replace_address = replace_address;
	C_sendto(C_BROADCAST_ADDRESS, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SLP, SLP_PACKET_TYPE_ANNOUNCE, &announcement, sizeof(SLP_ANNOUNCE_packet_t));
}

/**
 * @brief Send a SLP PURGE packet.
 *
 * @param service_id ID of the service that is to be purged.
 * @param server Server address.
 * @param redirect_address Address of node to redirect requests to.
 */
static void send_SLP_PURGE(const SERVICE_id_t* service_id, C_addr_t server, C_addr_t redirect_address) {
	SLP_PURGE_packet_t purge;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Broadcasting PURGE for service \"%s\"...\n", service_id);
	memset(&purge, 0, sizeof(SLP_PURGE_packet_t));
	SERVICE_copy_service_id(purge.service_id, service_id);
	purge.redirect_address = redirect_address;
	C_sendto(C_BROADCAST_ADDRESS, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SLP, SLP_PACKET_TYPE_PURGE, &purge, sizeof(SLP_PURGE_packet_t));
}
#endif	// SLP_USE_ANNOUNCEMENTS

/**
 * @brief Handle an incoming SLP QUERY.
 *
 * Send a reply to this query if the service is hosted on this node, forward
 * the query otherwise.
 *
 * @param header Pointer to header of incoming packet.
 * @param query Pointer to SLP query.
 */
static void handle_SLP_QUERY(const net_header_t* header, const SLP_QUERY_packet_t* query) {
	if(header->source_address == C_LOCAL_ADDRESS) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_ERROR, "SLP: Dropping own QUERY...\n");
		return;
	}

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_VERBOSE, "SLP: Received QUERY from %s.\n", addr_to_str(header->source_address));
	if(SERVICE_get_running_service(query->service_id) != NULL) {
		if(query->current_try <= (SLP_QUERY_TRIES + 1) / 2)
			C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_NOTE, "SLP: Service \"%s\" found for query from node %s after %d hop(s), sending reply...\n", query->service_id, addr_to_str(header->source_address), header->hop_count);
		else
			C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_WARN, "SLP: Service \"%s\" found for query from node %s after %d hop(s) on retry #%d, sending reply...\n", query->service_id, addr_to_str(header->source_address), header->hop_count, query->current_try);
		send_SLP_REPLY(header->source_address, query->service_id);
	} else if(header->ttl > 0) {
		// NOTE: One could also check the local service cache here.
		// However, this would get in the way of directory agents.
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Forwarding QUERY...\n");
		forward(header, query);
	} else {
#ifdef SLP_USE_EXPANDING_RING_SEARCH
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Dropping QUERY with expired TTL...\n");
#else
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_WARN, "SLP: Dropping QUERY with expired TTL...\n");
#endif	// SLP_USE_EXPANDING_RING_SEARCH
	}
}

/**
 * @brief Check whether the service ID of two queries match.
 *
 * @param item Query to check.
 * @param cmp_item Reference query.
 * @return <code>true</code> if the service ID of the two queries matches,
 * <code>false</code> otherwise.
 */
static bool matches_query_service_id(const void* item, const void* cmp_item) {
	SLP_query_t* query = (SLP_query_t*) item;
	SLP_query_t* cmp_query = (SLP_query_t*) cmp_item;
	return SERVICE_same_service_id(query->service_id, cmp_query->service_id);
}

/**
 * @brief Handle an incoming SLP REPLY.
 *
 * @param header Pointer to the header of the incoming reply.
 * @param reply Pointer to the SLP reply.
 */
static void handle_SLP_REPLY(const net_header_t* header, const SLP_REPLY_packet_t* reply) {
	SLP_query_t cmp_query;
	SLP_query_t* query = NULL;
#ifdef SLP_USE_CACHE
	bool cache_updated = false;
#endif	// SLP_USE_CACHE
	bool matched_active_query = false;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_NOTE, "SLP: Received reply to query for service \"%s\" from %s.\n", reply->service_id, addr_to_str(header->source_address));

	SERVICE_copy_service_id(cmp_query.service_id, reply->service_id);
	while((query = (SLP_query_t*) list_find(C_SLP_active_queries, &cmp_query, matches_query_service_id)) != NULL) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_VERBOSE, "SLP: Calling callback function...\n");

#ifdef SLP_USE_CACHE
		// Only update the with the data of the first item in the queue
		// as this is the item that triggered the query, and hence the
		// start time is correct.
		if(!cache_updated) {
			cache_update(query->service_id, header->source_address, header->hop_count, (C_time() - query->start_time), true);
			cache_updated = true;
		}
#endif	// SLP_USE_CACHE

		query->result->error_code = C_ERROR_OK;
		query->result->server = header->source_address;
		query->function(query->result);
		query->result = NULL;
		list_remove(C_SLP_active_queries, query);
		if(query->timeout_callback_id)
			callback_cancel(query->timeout_callback_id);
		FREE(query);

		matched_active_query = true;
	}

	// Only report and count stats for one query. Queued queries don't
	// count towards the statistics.
	if(matched_active_query) {
		SERVICE_service_t* service = SERVICE_get_service(reply->service_id);
		if(service != NULL)
			PLACEMENT_report_known_service_instance(service, header->source_address);
		C_SLP_successful_queries++;
	}
}

#ifdef SLP_USE_ANNOUNCEMENTS
/**
 * @brief Handle an incoming SLP ANNOUNCE.
 *
 * @param header Pointer to the packet header.
 * @param announcement Pointer to the SLP announcement.
 */
static void handle_SLP_ANNOUNCE(const net_header_t* header, const SLP_ANNOUNCE_packet_t* announcement) {
	SERVICE_service_t* service = NULL;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Received announcement of service \"%s\" hosted on %s at %d hop(s).\n", announcement->service_id, addr_to_str(header->source_address), header->hop_count);

	if(
		(header->ttl > 0)
		&& !cache_contains(announcement->service_id, header->source_address)
	) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Forwarding announcement of service \"%s\" hosted on %s...\n", announcement->service_id, addr_to_str(header->source_address));
		forward(header, announcement);
	}

	cache_update(announcement->service_id, header->source_address, header->hop_count, (float) C_NETWORK_TRAVERSAL_TIME, true);
	if(announcement->replace_address != C_BROADCAST_ADDRESS)
		cache_remove(announcement->service_id, announcement->replace_address);

	service = SERVICE_get_service(announcement->service_id);
	if(service != NULL) {
		PLACEMENT_report_known_service_instance(service, header->source_address);
		if(announcement->replace_address != C_BROADCAST_ADDRESS)
			PLACEMENT_report_invalid_service_instance(service, announcement->replace_address);
	}
}

/**
 * @brief Handle an incoming SLP PURGE.
 *
 * @param header Pointer to the packet header.
 * @param purge Pointer to the SLP purge request.
 */
static void handle_SLP_PURGE(const net_header_t* header, const SLP_PURGE_packet_t* purge) {
	SERVICE_service_t* service = NULL;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Received purge request of service \"%s\" hosted on %s.\n", purge->service_id, addr_to_str(header->source_address));

	if(
		(header->ttl > 0)
		&& cache_contains(purge->service_id, header->source_address)
	) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Removing cache entry of service \"%s\" hosted on %s and forwarding purge request...\n", purge->service_id, addr_to_str(header->source_address));
		cache_remove(purge->service_id, header->source_address);
		if(purge->redirect_address != C_BROADCAST_ADDRESS)
			cache_update(purge->service_id, purge->redirect_address, C_MAXIMAL_NETWORK_DIAMETER, (float) C_NETWORK_TRAVERSAL_TIME, false);
		forward(header, purge);
	}

	service = SERVICE_get_service(purge->service_id);
	if(service != NULL) {
		PLACEMENT_report_invalid_service_instance(service, header->source_address);
		if(purge->redirect_address != C_BROADCAST_ADDRESS)
			PLACEMENT_report_known_service_instance(service, purge->redirect_address);
	}
}
#endif	// SLP_USE_ANNOUNCEMENTS

/**
 * @brief SLP packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void SLP_recv(const net_header_t* header, const void* payload) {
	C_ABORT_ON(header->app != APP_TYPE_SLP, "SLP: Received non-SLP packet.\n");

	switch(header->type) {
	case SLP_PACKET_TYPE_QUERY:
		C_ABORT_ON(header->payload_length != sizeof(SLP_QUERY_packet_t), "SLP: Received QUERY packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SLP_QUERY_packet_t));
		handle_SLP_QUERY(header, (SLP_QUERY_packet_t*) payload);
		break;
	case SLP_PACKET_TYPE_REPLY:
		C_ABORT_ON(header->payload_length != sizeof(SLP_REPLY_packet_t), "SLP: Received REPLY packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SLP_REPLY_packet_t));
		handle_SLP_REPLY(header, (SLP_REPLY_packet_t*) payload);
		break;
#ifdef SLP_USE_ANNOUNCEMENTS
	case SLP_PACKET_TYPE_ANNOUNCE:
		C_ABORT_ON(header->payload_length != sizeof(SLP_ANNOUNCE_packet_t), "SLP: Received ANNOUNCE packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SLP_ANNOUNCE_packet_t));
		handle_SLP_ANNOUNCE(header, (SLP_ANNOUNCE_packet_t*) payload);
		break;
	case SLP_PACKET_TYPE_PURGE:
		C_ABORT_ON(header->payload_length != sizeof(SLP_PURGE_packet_t), "SLP: Received PURGE packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SLP_PURGE_packet_t));
		handle_SLP_PURGE(header, (SLP_PURGE_packet_t*) payload);
		break;
#endif	// SLP_USE_ANNOUNCEMENTS
	default:
		C_ABORT_ON(true, "SLP: Received packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
	}
}



/*
 * Interface provided by this module.
 */

#ifdef SLP_USE_CACHE
/**
 * @brief Timeout handler for removing service instances from the blacklist.
 *
 * @param parameter Service instance that is to be removed from the blacklist.
 */
static void remove_from_unavailable_services(callback_parameter_t parameter) {
	SLP_service_instance_t* unavailable_service_instance = (SLP_service_instance_t*) parameter;
	list_remove(C_SLP_unavailable_services, unavailable_service_instance);
	FREE(unavailable_service_instance);
}
#endif	// SLP_USE_CACHE

/**
 * @brief Timeout handler for locating a service.
 *
 * @param parameter Query for which the timeout has expired.
 */
static void SLP_locate_service_timeout(callback_parameter_t parameter) {
	SLP_query_t* query = (SLP_query_t*) parameter;
	SLP_query_t* query_with_same_service_id = NULL;
#ifdef SLP_USE_CACHE
	SLP_service_instance_t* unavailable_service_instance = NULL;
#endif	// SLP_USE_CACHE

	// Retry if no service instances have been found.
	if(query->current_try < SLP_QUERY_TRIES) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Timeout for QUERY (0x%x) for service \"%s\" has expired (try #%d).\n", query, query->service_id, query->current_try);
		query->current_try++;
		query->start_time = C_time();	// Update start time for the next retry.
		query->timeout_callback_id = callback_add(SLP_locate_service_timeout, query, SLP_QUERY_TIMEOUT);
		send_SLP_QUERY(query->service_id, query->current_try);

	// Abort and call callback if no retries are left.
	} else {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_WARN, "SLP: Timeout for QUERY (0x%x) for service \"%s\" has expired (try #%d). No retries left, service was not found.\n", query, query->service_id, query->current_try);
		query->result->error_code = C_ERROR_TIMEOUT;
		query->function(query->result);
		list_remove(C_SLP_active_queries, query);

		while((query_with_same_service_id = (SLP_query_t*) list_find(C_SLP_active_queries, query, matches_query_service_id)) != NULL) {
			query_with_same_service_id->result->error_code = C_ERROR_TIMEOUT;
			query_with_same_service_id->function(query_with_same_service_id->result);
			list_remove(C_SLP_active_queries, query_with_same_service_id);
			FREE(query_with_same_service_id);
			// Queued queries don't count towards the statistics.
		}

#ifdef SLP_USE_CACHE
		// Add service to blacklist and start timeout.
		alloc_service_instance(&unavailable_service_instance, query->service_id, C_BROADCAST_ADDRESS);
		list_add(C_SLP_unavailable_services, unavailable_service_instance);
		unavailable_service_instance->delayed_removal_callback_id = callback_add(remove_from_unavailable_services, unavailable_service_instance, SLP_UNAVAILABLE_SERVICE_TIMEOUT);
#endif	// SLP_USE_CACHE

		FREE(query);
		C_SLP_failed_queries++;
	}
}

/**
 * @brief Locate a service.
 *
 * @param service_id Service ID of the service to be located.
 * @param function Callback function.
 * @param result Pointer to query results.
 */
void SLP_locate_service(const SERVICE_id_t* service_id, callback_function_t function, SLP_query_result_t* result) {
#ifdef SLP_USE_CACHE
	SLP_service_instance_t cmp_service_instance;
	C_addr_t server = C_BROADCAST_ADDRESS;
#endif	// SLP_USE_CACHE
	SLP_query_t* query = NULL;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	if((function == NULL) || (result == NULL)) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_ERROR, "SLP: Callback function or parameter not given (%p, %p).\n", function, result);
		result->error_code = C_ERROR_PARAMETER;
		function(result);
		return;
	}

#ifdef SLP_USE_CACHE
	// Check whether service is cached.
	switch(cache_lookup(service_id, &server)) {
	case C_ERROR_OK:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Reusing cached server %s for service \"%s\"...\n", addr_to_str(server), service_id);
		result->server = server;
		result->error_code = C_ERROR_OK;
		function(result);
		return;
	case C_ERROR_NOT_FOUND:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: No server in cache for service \"%s\", starting SLP query...\n", service_id);
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Cache lookup returned invalid error code.\n");
	}

	// Check whether service is on blacklist.
	SERVICE_copy_service_id(cmp_service_instance.service_id, service_id);
	if(list_find(C_SLP_unavailable_services, &cmp_service_instance, matches_service_instance_service_id) != NULL) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Service \"%s\" is tagged as not available.\n", service_id);
		result->server = C_BROADCAST_ADDRESS;
		result->error_code = C_ERROR_TIMEOUT;
		function(result);
		return;
	}
#endif	// SLP_USE_CACHE

	// Do stats and abort if this component is busy or the request can be
	// trivially satisfied.
	if(list_size(C_SLP_active_queries) >= SLP_MAX_ACTIVE_QUERIES) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_ERROR, "SLP: No empty slots for query processing.\n");
		result->error_code = C_ERROR_BUSY;
		function(result);
		C_SLP_total_queries++;
		C_SLP_failed_queries++;
		return;
	}
	if(SERVICE_get_running_service(service_id) != NULL) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Service \"%s\" is hosted on the node itself.\n", service_id);
#ifdef SLP_USE_CACHE
		// Apparently the cache entry has been removed due to lack of
		// client demand on the local node. Add it once more.
		cache_update(service_id, C_LOCAL_ADDRESS, 0, 0.0, true);
#endif	// SLP_USE_CACHE
		result->server = C_LOCAL_ADDRESS;
		result->error_code = C_ERROR_OK;
		function(result);
		C_SLP_total_queries++;
		C_SLP_successful_queries++;
		return;
	}

	// Prepare query for service.
	MALLOC(query, SLP_query_t);
	SERVICE_copy_service_id(query->service_id, service_id);
	query->current_try = 1;
	query->start_time = C_time();
	query->function = function;
	query->result = result;
	query->result->error_code = C_ERROR_NOT_SET;
	if(list_find(C_SLP_active_queries, query, matches_query_service_id) == NULL) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_NOTE, "SLP: Locating service \"%s\"...\n", service_id);
		query->timeout_callback_id = callback_add(SLP_locate_service_timeout, query, SLP_QUERY_TIMEOUT);
		send_SLP_QUERY(service_id, 0);
		C_SLP_total_queries++;
	} else {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Look-up of service \"%s\" already in progress, queueing query...\n", service_id);
		// Queued queries don't count towards the statistics.
	}
	list_add(C_SLP_active_queries, query);
}

/**
 * @brief Announce a service hosted on the local node.
 *
 * @param service_id Service ID of the service to be announced.
 * @param replace_address Address of node that this host replaces,
 * C_BROADCAST_ADDRESS if none.
 */
void SLP_announce_service(const SERVICE_id_t* service_id, C_addr_t replace_address) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#ifdef SLP_USE_ANNOUNCEMENTS
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_NOTE, "SLP: Announcing service \"%s\"...\n", service_id);

	// Add to own cache to avoid retransmissions.
	cache_update(service_id, C_LOCAL_ADDRESS, 0, 0.0, true);
	if(replace_address != C_BROADCAST_ADDRESS)
		cache_remove(service_id, replace_address);
	send_SLP_ANNOUNCE(service_id, C_LOCAL_ADDRESS, replace_address, 1);
#else
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Announcing services is disabled.\n");
#endif	// SLP_USE_ANNOUNCEMENTS
}

/**
 * @brief Purge a service hosted on the local node from all caches.
 *
 * @param service_id Service ID of the service to be purged.
 * @param redirect_address Address of node to redirect requests to.
 */
void SLP_purge_service(const SERVICE_id_t* service_id, C_addr_t redirect_address) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#ifdef SLP_USE_ANNOUNCEMENTS
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_NOTE, "SLP: Purging service \"%s\"...\n", service_id);

	// Remove from own cache to avoid retransmissions.
	cache_remove(service_id, C_LOCAL_ADDRESS);
	if(redirect_address != C_BROADCAST_ADDRESS)
		cache_update(service_id, redirect_address, C_MAXIMAL_NETWORK_DIAMETER, (float) C_NETWORK_TRAVERSAL_TIME, false);
	send_SLP_PURGE(service_id, C_LOCAL_ADDRESS, redirect_address);
#else
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Purging services is disabled.\n");
#endif	// SLP_USE_ANNOUNCEMENTS
}

/**
 * @brief Set the entry in the service cache to a known server address.
 *
 * @param service_id Service ID of the cached service.
 * @param server Address of the server for this service.
 */
void SLP_cache_update(const SERVICE_id_t* service_id, C_addr_t server) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#ifdef SLP_USE_CACHE
	cache_update(service_id, server, C_MAXIMAL_NETWORK_DIAMETER, (float) C_NETWORK_TRAVERSAL_TIME, false);
#else
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Caching is disabled.\n");
#endif	// SLP_USE_CACHE
}

/**
 * @brief Set the entry in the service cache to a known server address
 * including mertrics.
 *
 * @param service_id Service ID of the cached service.
 * @param server Address of the server for this service.
 * @param hop_count Distance to the server in hops.
 * @param delay Time it took the reply to arrive.
 */
void SLP_cache_update_with_metrics(const SERVICE_id_t* service_id, C_addr_t server, uint8_t hop_count, C_time_t delay) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#ifdef SLP_USE_CACHE
	cache_update(service_id, server, hop_count, delay, true);
#else
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Caching is disabled.\n");
#endif	// SLP_USE_CACHE
}

/**
 * @brief Remove a cache entry.
 *
 * @param service_id Service ID of the cached service to be removed.
 * @param server Address of the server for this service.
 */
void SLP_cache_remove(const SERVICE_id_t* service_id, C_addr_t server) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

#ifdef SLP_USE_CACHE
	cache_remove(service_id, server);
#else
	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Caching is disabled.\n");
#endif	// SLP_USE_CACHE
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Callback of command function to locate a service.
 *
 * @param parameter Result of the query.
 */
static void SLP_command_locate_service_callback(callback_parameter_t parameter) {
	SLP_query_result_t* result = (SLP_query_result_t*) parameter;
	char* service_id = (char*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_NOTE, "SLP: Service \"%s\" located on node %s.\n", service_id, addr_to_str(result->server));
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_ERROR, "SLP: Wrong parameter value.\n");
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_WARN, "SLP: Service \"%s\" not found.\n", service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Query for service \"%s\" has been aborted.\n", service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_WARN, "SLP: Too many service queries in progress.\n");
		break;
	default:
		C_ABORT_ON(true, "SLP: Callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(service_id);
	FREE(result);
}

/**
 * @brief SLP command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool SLP_command(int argc, const char*const* argv) {
	SLP_query_result_t* result = NULL;

	if(argc == 3) {
		if(strcmp(argv[1], "SLP_locate_service") == 0) {
			MALLOC(result, SLP_query_result_t);
			SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, argv[2]);
			SLP_locate_service(argv[2], SLP_command_locate_service_callback, result);
			return true;
		} else if(strcmp(argv[1], "SLP_announce_service") == 0) {
			SLP_announce_service(argv[2], C_BROADCAST_ADDRESS);
			return true;
		}
	}
	return false;
}
