/**
 * @file callback.c
 *
 * @brief Callback functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: callback.c 3298 2010-02-15 18:31:42Z gwitten $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "debug.h"
#include "list.h"
#include "callback.h"



/**
 * @brief Maximal supported number of simultaneous callbacks.
 */
#define MAX_CALLBACKS	1024

/**
 * @brief Debug level for this component.
 */
#define CALLBACK_DEBUG_LEVEL	C_MSG_WARN



/**
 * @brief Check whether a callback expires later than another one.
 *
 * @param item Pointer to callback to check.
 * @param cmp_item Pointer to reference callback.
 * @return <code>true</code> if the callback to check expires later than the
 * reference callback, <code>false</code> otherwise.
 */
static bool callback_is_later(const void* item, const void* cmp_item) {
	callback_t* callback = (callback_t*) item;
	callback_t* cmp_callback = (callback_t*) cmp_item;
	return (callback->time > cmp_callback->time);
}

/**
 * @brief Get the next valid callback ID.
 *
 * @return Next valid callback ID.
 */
static callback_id_t get_next_callback_id() {
	C_next_callback_id++;
	if(C_next_callback_id == 0) {
		C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_WARN, "Detected callback ID wrap-around.\n");
		C_next_callback_id = 1;
	}
	return C_next_callback_id;
}

/**
 * @brief Add a new callback to be executed later on.
 *
 * @param function Pointer to the callback function.
 * @param parameter Pointer to the parameter passed to the callback function.
 * @param delay Delay before the callback expires in seconds.
 * @return Unique ID of the callback (guanranteed not to be 0).
 */
callback_id_t callback_add(callback_function_t function, callback_parameter_t parameter, C_time_t delay) {
	callback_t* callback = NULL;

	C_ABORT_ON(function == NULL, "Trying to register callback with NULL function...\n");
	C_ABORT_ON(delay < 0.0, "Trying to register callback to function at %p with negative delay (%f)...\n", delay);
	C_ABORT_ON(list_size(C_callbacks) >= MAX_CALLBACKS, "Trying to register callback to function at %p, but %d callbacks are already running...\n", function, MAX_CALLBACKS);

	MALLOC(callback, callback_t);
	callback->id = get_next_callback_id();
	callback->function = function;
	callback->parameter = parameter;
	callback->time = C_time() + delay;
	list_insert_where(C_callbacks, callback, callback_is_later);

	if(delay == 0.0) {
		C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_INFO, "Registering callback #%d to function at %p and parameter at %p without delay...\n", callback->id, function, parameter);
	} else {
		C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_INFO, "Registering callback #%d to function at %p and parameter at %p with a delay of %.2f seconds...\n", callback->id, function, parameter, delay);
	}

	C_activate_callback(((callback_t*) list_head(C_callbacks))->time);
	C_registered_callbacks++;
	return callback->id;
}

/**
 * @brief Check whether two callback have the same unique ID.
 *
 * @param item Callback to check.
 * @param cmp_item Reference callback.
 * @return <code>true</code> if the callbacks have the same unique ID,
 * <code>false</code> otherwise.
 */
static bool cmp_callback_id(const void* item, const void* cmp_item) {
	callback_t* callback = (callback_t*) item;
	callback_t* cmp_callback = (callback_t*) cmp_item;
	return (callback->id == cmp_callback->id);
}

/**
 * @brief Cancel callback, thus preventing it from ever being executed.
 *
 * @param id Unique ID of the callback that is to be cancelled.
 */
void callback_cancel(callback_id_t id) {
	callback_t* callback = NULL;
	callback_t cmp_callback;
	bool was_active_callback = false;

	C_ABORT_ON(id == 0, "Trying to cancel unused callback (i.e., with ID #0).\n");

	memset(&cmp_callback, 0, sizeof(callback_t));
	cmp_callback.id = id;
	callback = (callback_t*) list_find(C_callbacks, &cmp_callback, cmp_callback_id);
	C_ABORT_ON(callback == NULL, "Trying to cancel callback #%d, but no callback with that ID is scheduled for execution.\n", id);

	C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_INFO, "Cancelling callback #%d to function at %p and parameter at %p...\n", callback->id, callback->function, callback->parameter);
	was_active_callback = (callback == (callback_t*) list_head(C_callbacks));
	list_remove(C_callbacks, callback);
	FREE(callback);
	C_executed_callbacks++;

	if(was_active_callback) {
		if(!list_empty(C_callbacks))
			C_activate_callback(((callback_t*) list_head(C_callbacks))->time);
		else
			C_activate_callback(0.0);
	}
}

/**
 * @brief This function is called when a callback has expired.
 */
void callback_expire() {
	C_time_t time = C_time();
	callback_t* callback = NULL;

	if(list_empty(C_callbacks)) {
		C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_DEBUG, "Ignoring call to cancelled callback...\n");
		return;
	}
	callback = (callback_t*) list_head(C_callbacks);

	// Allow callbacks to occur a bit too early to work around rounding errors.
	C_ABORT_ON((callback->time - time) > 0.1, "Trying to call expired callback, but next callback is only due in %f seconds.\n", callback->time - time);

	if((time - callback->time) > 1.0) {
		C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_WARN, "Calling expired callback #%d with function at %p and with parameter at %p (%.2fs overdue)...\n", callback->id, callback->function, callback->parameter, (time - callback->time));
	} else {
		C_DEBUG(CALLBACK_DEBUG_LEVEL, C_MSG_INFO, "Calling expired callback #%d with function at %p and with parameter at %p (%.2fs overdue)...\n", callback->id, callback->function, callback->parameter, (time - callback->time));
	}
	C_executed_callbacks++;
	list_remove(C_callbacks, callback);
	callback->function(callback->parameter);
	FREE(callback);

	// Schedule next callback, if there is one and if this module hasn't
	// been shut down in the mean time.
	if((C_callbacks != NULL) && !list_empty(C_callbacks))
		C_activate_callback(((callback_t*) list_head(C_callbacks))->time);
}
