/**
 * @file debug.c
 *
 * @brief Debugging functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: debug.c 3238 2010-01-24 19:25:06Z wittenbu $
 */

#ifndef ARCH_WIN32
#include <execinfo.h>
#endif	// ARCH_WIN32
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "debug.h"
#include "net.h"



#ifdef C_ENABLE_DEBUG

#ifndef ARCH_WIN32

/**
 * @brief Use color in output.
 */
#define DEBUG_USE_COLOR

#endif

/**
 * @brief Raise a signal on errors.
 *
 * Flag to make the system raise a signal on log messages with severity greater
 * or equal to C_MSG_ERROR.
 */
#define DEBUG_SIGNAL_ON_ERRORS

/**
 * @brief Size of backtrace on FATAL errors.
 */
#define DEBUG_BACKTRACE_SIZE	64



/**
 * @brief Convert numeric debug level to a string.
 *
 * @param msg_type Numeric type of the message.
 * @return A string representing the debug level.
 */
static const char* C_MSG_TO_STRING(uint8_t msg_type) {
	switch(msg_type & C_MSG_PRIORITY) {
		case C_MSG_FATAL:	return "FATAL";
		case C_MSG_ERROR:	return "ERROR";
		case C_MSG_WARN:	return "WARN";
		case C_MSG_NOTE:	return "NOTE";
		case C_MSG_INFO:	return "INFO";
		case C_MSG_DEBUG:	return "DEBUG";
		case C_MSG_VERBOSE:	return "VERB";
		default: C_printf("FATAL: Unknown message type 0x%X.\n", msg_type); abort(); return "NOT REACHED";
	}
}

#ifdef DEBUG_USE_COLOR
/**
 * @brief Return ASCII control sequence to highlight debugging output.
 *
 * @param msg_type Numeric type of the message.
 * @return A string with a ASCII control sequence for appropriate highlighting.
 */
static const char* C_MSG_TO_COLOR(uint8_t msg_type) {
	switch(msg_type & C_MSG_PRIORITY) {
		case C_MSG_FATAL:	return "\033[41m\033[1;31m";	// bright red on red
		case C_MSG_ERROR:	return "\033[40m\033[1;31m";	// bright red on black
		case C_MSG_WARN:	return "\033[40m\033[1;33m";	// yellow on black
		case C_MSG_NOTE:	return "\033[40m\033[1;32m";	// green on black
		case C_MSG_INFO:	return "\033[40m\033[1;37m";	// white on black
		case C_MSG_DEBUG:	return "\033[40m\033[2;37m";	// dim white on black
		case C_MSG_VERBOSE:	return "\033[40m\033[1;30m";	// dark gray on black
		default: C_printf("FATAL: Unknown message type 0x%X.\n", msg_type); abort(); return "NOT REACHED";
	}
}
#endif	// DEBUG_USE_COLOR

#endif	// C_ENABLE_DEBUG

/**
 * @brief Print formatted debuggging output.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param format A format string.
 * @param ... Parameters for the format string.
 */
void debug_printf(uint8_t debug_level, uint8_t msg_type, const char *format, ...) {
#ifdef C_ENABLE_DEBUG

#ifdef ARCH_WIN32
	SYSTEMTIME time;
#else
	// Variable to print a backtrace.
	void *array[DEBUG_BACKTRACE_SIZE];
	size_t size;
	char **strings;
	size_t i;
#endif	// ARCH_WIN32

	if(debug_level > C_MSG_ERROR) {
		C_printf("Abort: High debug level (0x%x) might cause information loss.\n", debug_level);
		abort();
	}

	if((msg_type & C_MSG_PRIORITY) >= debug_level) {
		va_list argp;

		if((msg_type & C_MSG_MODIFIER) != C_MSG_CONT) {
#ifdef DEBUG_USE_COLOR
			C_printf("%s", C_MSG_TO_COLOR(msg_type));
#endif	// DEBUG_USE_COLOR

#ifdef ARCH_LINUX
			time_t time = (time_t) C_time();
			char time_string[9];
			strftime(time_string, 9, "%T", localtime(&time));
			C_printf("[%s @ %s: %s]", time_string, addr_to_str(C_LOCAL_ADDRESS), C_MSG_TO_STRING(msg_type));
#elif defined(ARCH_WIN32)
			GetLocalTime(&time);
			C_printf("[%02d:%02d:%02d @ %s: %s]", time.wHour, time.wMinute, time.wSecond, addr_to_str(C_LOCAL_ADDRESS), C_MSG_TO_STRING(msg_type));
#else
			C_printf("[%.4f @ %s: %s]", C_time(), addr_to_str(C_LOCAL_ADDRESS), C_MSG_TO_STRING(msg_type));
#endif	// ARCH_*

#ifdef DEBUG_USE_COLOR
			C_printf("\033[0m");	// Reset colors to default.
#endif	// DEBUG_USE_COLOR

			C_printf("\t");
		}

		va_start(argp, format);
		C_vprintf(format, argp);
		va_end(argp);
	}

	// Special handling for certain message types.
	switch(msg_type & C_MSG_PRIORITY) {
		case C_MSG_FATAL:
			C_printf("\n");
			C_printf("\t\t\t\t\\|/ ____ \\|/\n");
			C_printf("\t\t\t\t\"@'/ .. \\`@\"\n");
			C_printf("\t\t\t\t/_| \\__/ |_\\\n");
   			C_printf("\t\t\t\t   \\__U_/\n");
			C_printf("\n");

#ifndef ARCH_WIN32
			// Print a backtrace.
			size = backtrace(array, DEBUG_BACKTRACE_SIZE);
			strings = backtrace_symbols(array, size);
			C_printf("Backtrace of %zd stack frame(s):\n", size);
			for(i = 0; i < size; i++)
				C_printf("%s\n", strings[i]);
			free (strings);
			C_printf("\n");
#endif	// ARCH_WIN32

			break;
		case C_MSG_ERROR:
			C_DEBUG_ERRORs++;
			break;
		case C_MSG_WARN:
			C_DEBUG_WARNs++;
			break;
		default: ;	// Do nothing.
	}

#ifdef DEBUG_SIGNAL_ON_ERRORS
	if((msg_type & C_MSG_PRIORITY) == C_MSG_ERROR) {
		fflush(NULL);
		raise(SIGTRAP);
	}
#endif	// DEBUG_SIGNAL_ON_ERRORS

#endif	// C_ENABLE_DEBUG
}
