/**
 * @file mapping.c
 *
 * @brief Network mapping functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: mapping.c 3365 2010-02-27 20:55:49Z gwitten $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "debug.h"
#include "list.h"
#include "net.h"



/**
 * @brief Debug level for this component.
 */
#define MAPPING_DEBUG_LEVEL	C_MSG_WARN

/**
 * @brief Distance value for disconneted nodes.
 *
 * NOTE: This should correspond to the variable type of the edge cost and
 * distance matrices.
 */
#define MAPPING_DISCONNECTED_NODE_DISTANCE	0xFF

#ifdef C_USE_NETWORK_MAPPING
/**
 * @brief Time after which inactive, previously known links are purged from the
 * network map.
 *
 * A good value depends highly on the volatily of the network topology, e.g.,
 * caused by mobility. In these scenarios, it would make sense to make it
 * dependant on the beacon interval, e.g., by using a value such as
 * (10 * C_BEACON_INTERVAL + C_NETWORK_TRAVERSAL_TIME). However, since we focus
 * on static network for know, a very conservative value makes more sense.
 */
#define MAPPING_KNOWN_LINK_TIMEOUT	180.0
#endif	// C_USE_NETWORK_MAPPING



/**
 * @brief Returns the number of currently known nodes.
 *
 * The variable type of this function should correspond to the value of
 * <code>C_UNKNOWN_NODE_ID</code>.
 *
 * @return Number of known nodes.
 */
uint16_t get_number_of_known_nodes() {
	return (uint16_t) list_size(C_known_nodes);
}

#ifdef C_USE_NETWORK_MAPPING

/**
 * @brief Calculate the edge cost matrix for the known network.
 *
 * Time complexity:	O(nodes ^ 2
 *			+ nodes * avg_routing_hops)
 *			= O(nodes ^ 2)
 * Space complexity:	O(nodes ^ 2)
 *
 * @param number_of_nodes Number of nodes.
 * @param edge_cost_matrix Pointer to edge cost matrix to be calculated.
 */
static void calculate_edge_cost(uint16_t number_of_nodes, uint8_t* edge_cost_matrix) {
	uint16_t i, j;

	C_ABORT_ON(number_of_nodes != list_size(C_known_nodes), "Trying to calculate edge cost, but size of matrix doesn't match the number of known nodes (%d != %d)\n", number_of_nodes, list_size(C_known_nodes));
	C_ABORT_ON(edge_cost_matrix == NULL, "Trying to calculate edge cost with NULL edge cost matrix.\n");

	C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_VERBOSE, "Calculating edge cost matrix for %d nodes with %d known nodes in the network...\n", number_of_nodes, list_size(C_known_nodes));

	// Inititialize edge cost matrix.
	memset(edge_cost_matrix, MAPPING_DISCONNECTED_NODE_DISTANCE, (number_of_nodes * number_of_nodes));
	for(i = 0; i < number_of_nodes; i++)
		edge_cost_matrix[(i * number_of_nodes) + i] = 0;

	for(i = 0; i < number_of_nodes; i++) {
		for(j = 0; j < number_of_nodes; j++) {
			if(C_network_map[(i * C_MAXIMAL_NUMBER_OF_NODES) + j] != 0)
				edge_cost_matrix[(i * number_of_nodes) + j] = 1;
		}
	}

	// Print the edge cost matrix for debugging.
	C_IF_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG) {
		C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "Edge cost:\t");
		for(i = 0; i < number_of_nodes; i++)
			C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "%s\t", addr_to_str(node_id_to_address(i)));
		C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "\n");
		for(i = 0; i < number_of_nodes; i++) {
			C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "%s:\t", addr_to_str(node_id_to_address(i)));
			for(j = 0; j < number_of_nodes; j++) {
				C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "%d\t", edge_cost_matrix[(i * number_of_nodes) + j]);
			}
			C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "\n");
		}
	}
}

/**
 * @brief Floyd-Warshall algorithm.
 *
 * Solves the all-pairs shortest path problem.
 *
 * Adapted from Joshua Robinson's implementation at
 * http://www.ece.rice.edu/~jpr/research/fwarsh.html
 *
 * Time complexity:	O(nodes ^ 3)
 * Space complexity:	O(nodes ^ 2)
 *
 * @param number_of_nodes Number of nodes.
 * @param edge_cost_matrix Pointer to edge cost matrix.
 * @param distance_matrix Pointer to distance matrix to be calculated.
 * @param predicate_matrix Pointer to predicate matrix to be calculated. May be NULL, if predicate matrix is not the be calculated.
 */
static void floyd_warshall(uint16_t number_of_nodes, const uint8_t* edge_cost_matrix, uint8_t* distance_matrix, uint16_t* predicate_matrix)
{
	uint16_t i, j, k;
	bool graph_is_disjunct = false;

	C_ABORT_ON(edge_cost_matrix == NULL, "Trying to run Floyd-Warshall algorithm with NULL edge cost matrix.\n");
	C_ABORT_ON(distance_matrix == NULL, "Trying to run Floyd-Warshall algorithm with NULL distance matrix.\n");

	// Floyd-Warshall initialization.
	memcpy(distance_matrix, edge_cost_matrix, (sizeof(uint8_t) * number_of_nodes * number_of_nodes));
	if(predicate_matrix != NULL) {
		memset(predicate_matrix, C_UNKNOWN_NODE_ID, (sizeof(uint8_t) * number_of_nodes * number_of_nodes));
		for(i = 0; i < number_of_nodes; i++) {
			for(j = 0; j < number_of_nodes; j++) {
				if(edge_cost_matrix[(i * number_of_nodes) + j] != MAPPING_DISCONNECTED_NODE_DISTANCE)
					predicate_matrix[(i * number_of_nodes) + j] = i;
			}
		}
	}

	// Floyd-Warshall main loop.
	for(k = 0; k < number_of_nodes; k++) {
		for(i = 0; i < number_of_nodes; i++) {
			for(j = 0; j < number_of_nodes; j++) {
				if(distance_matrix[(i * number_of_nodes) + j] > (distance_matrix[(i * number_of_nodes) + k] + distance_matrix[(k * number_of_nodes) + j])) {
					distance_matrix[(i * number_of_nodes) + j] = distance_matrix[(i * number_of_nodes) + k] + distance_matrix[(k * number_of_nodes) + j];
					if(predicate_matrix != NULL)
						predicate_matrix[(i * number_of_nodes) + j] = predicate_matrix[(k * number_of_nodes) + j];
				}
			}
		}
	}

	C_IF_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG) {
		// Print the distance matrix for debugging.
		C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "Distance:\t");
		for(i = 0; i < number_of_nodes; i++)
			C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "%s\t", addr_to_str(node_id_to_address(i)));
		C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "\n");
		for(i = 0; i < number_of_nodes; i++) {
			C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "%s:\t", addr_to_str(node_id_to_address(i)));
			for(j = 0; j < number_of_nodes; j++) {
				C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "%d\t", distance_matrix[(i * number_of_nodes) + j]);
				if(distance_matrix[(i * number_of_nodes) + j] == MAPPING_DISCONNECTED_NODE_DISTANCE)
					graph_is_disjunct = true;
			}
			C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "\n");
		}

		// Print the predicate matrix for debugging.
		if(predicate_matrix != NULL) {
			C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "Predicate:\t");
			for(i = 0; i < number_of_nodes; i++)
				C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "%s\t", addr_to_str(node_id_to_address(i)));
			C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "\n");
			for(i = 0; i < number_of_nodes; i++) {
				C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "%s:\t", addr_to_str(node_id_to_address(i)));
				for(j = 0; j < number_of_nodes; j++) {
					C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "%s\t", addr_to_str(node_id_to_address(predicate_matrix[(i * number_of_nodes) + j])));
				}
				C_DEBUG(MAPPING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), "\n");
			}
		}
	}

	if(graph_is_disjunct)
		C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_WARN, "Detected disjunct graph while running Floyd-Warshall algorithm.\n");
}

/**
 * @brief Purge outdated entries from the network map.
 */
static void purge_network_map() {
	uint16_t i, j;
	uint16_t local_id = node_address_to_id(C_LOCAL_ADDRESS);
	C_time_t last_valid_time = C_time() - MAPPING_KNOWN_LINK_TIMEOUT;
	C_time_t link_time = 0.0;
	bool has_changed = false;

	// Make sure that the network map is only purged when new information
	// has become available.
	if(!C_network_map_may_need_purging)
		return;
	C_network_map_may_need_purging = false;

	for(i = 0; i < C_MAXIMAL_NUMBER_OF_NODES; i++) {
		for(j = 0; j < C_MAXIMAL_NUMBER_OF_NODES; j++) {
			link_time = C_network_map[(i * C_MAXIMAL_NUMBER_OF_NODES) + j];
			if(
				(link_time != 0)
				&& (link_time < last_valid_time)
				&& (link_time > (-1.0 * last_valid_time))
			) {
				if((i == local_id) || (j == local_id))
					C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_WARN, "Purging link between %s and %s (one of which is the local node) from the network map...\n", addr_to_str(node_id_to_address(i)), addr_to_str(node_id_to_address(j)));
				else
					C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_DEBUG, "Purging link between %s and %s from the network map...\n", addr_to_str(node_id_to_address(i)), addr_to_str(node_id_to_address(j)));
				C_network_map[(i * C_MAXIMAL_NUMBER_OF_NODES) + j] = 0;
				has_changed = true;
			}
		}
	}

	if(has_changed) {
		// Invalidate previous edge cost and distance calculations.
		if(C_edge_cost_matrix != NULL)
			FREE(C_edge_cost_matrix);
		if(C_distance_matrix != NULL)
			FREE(C_distance_matrix);
	}
}

/**
 * @brief Returns whether a link is known or not.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 * @return <code>true</code> if the link is known, <code>false</code>
 * otherwise.
 */
bool exists_known_link(C_addr_t src_address, C_addr_t dst_address) {
	uint16_t src_id = node_address_to_id(src_address);
	uint16_t dst_id = node_address_to_id(dst_address);

	purge_network_map();

	return (C_network_map[(src_id * C_MAXIMAL_NUMBER_OF_NODES) + dst_id] != 0);
}

/**
 * @brief Returns whether a link is directly known (as opposed to reported by
 * other nodes) or not.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 * @return <code>true</code> if the link is directly known, <code>false</code>
 * otherwise.
 */
bool exists_directly_known_link(C_addr_t src_address, C_addr_t dst_address) {
	uint16_t src_id = node_address_to_id(src_address);
	uint16_t dst_id = node_address_to_id(dst_address);

	return (C_network_map[(src_id * C_MAXIMAL_NUMBER_OF_NODES) + dst_id] > 0);
}

/**
 * @brief Returns whether a path between two known nodes is known.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 * @return <code>true</code> if a path is known, <code>false</code>
 * otherwise.
 */
bool exists_known_path(C_addr_t src_address, C_addr_t dst_address) {
	uint16_t number_of_nodes = get_number_of_known_nodes();
	uint8_t distance = MAPPING_DISCONNECTED_NODE_DISTANCE;

	C_ABORT_ON(((C_edge_cost_matrix == NULL) && (C_distance_matrix != NULL)) || ((C_edge_cost_matrix != NULL) && (C_distance_matrix == NULL)), "Edge cost matrix and distance matrics are initialized inconsistently (%p, %p).\n", C_edge_cost_matrix, C_distance_matrix);

	purge_network_map();

	if((C_edge_cost_matrix == NULL) || (C_distance_matrix == NULL)) {
		// Calculate the edge cost matrix.
		MALLOC_ARRAY(C_edge_cost_matrix, uint8_t, (number_of_nodes * number_of_nodes));
		calculate_edge_cost(number_of_nodes, C_edge_cost_matrix);

		// Calculate the distance matrix.
		MALLOC_ARRAY(C_distance_matrix, uint8_t, (number_of_nodes * number_of_nodes));
		floyd_warshall(number_of_nodes, C_edge_cost_matrix, C_distance_matrix, NULL);
	}
	distance = C_distance_matrix[(node_address_to_id(src_address) * number_of_nodes) + node_address_to_id(dst_address)];

	return (distance != MAPPING_DISCONNECTED_NODE_DISTANCE);
}

/**
 * @brief Get the distance between two known nodes.
 *
 * The distance metric corresponds to the hop count in the current
 * implementation, but it generally doesn't have to. If no path exists between
 * the two nodes, the network diameter is returned.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 * @return Distance between the two nodes, or network diameter if disconnected.
 */
float get_distance_between_nodes(C_addr_t src_address, C_addr_t dst_address) {
	uint16_t number_of_nodes = get_number_of_known_nodes();
	uint8_t distance = MAPPING_DISCONNECTED_NODE_DISTANCE;

	C_ABORT_ON(((C_edge_cost_matrix == NULL) && (C_distance_matrix != NULL)) || ((C_edge_cost_matrix != NULL) && (C_distance_matrix == NULL)), "Edge cost matrix and distance matrics are initialized inconsistently (%p, %p).\n", C_edge_cost_matrix, C_distance_matrix);

	purge_network_map();

	if((C_edge_cost_matrix == NULL) || (C_distance_matrix == NULL)) {
		// Calculate the edge cost matrix.
		MALLOC_ARRAY(C_edge_cost_matrix, uint8_t, (number_of_nodes * number_of_nodes));
		calculate_edge_cost(number_of_nodes, C_edge_cost_matrix);

		// Calculate the distance matrix.
		MALLOC_ARRAY(C_distance_matrix, uint8_t, (number_of_nodes * number_of_nodes));
		floyd_warshall(number_of_nodes, C_edge_cost_matrix, C_distance_matrix, NULL);
	}
	distance = C_distance_matrix[(node_address_to_id(src_address) * number_of_nodes) + node_address_to_id(dst_address)];

	if(distance == MAPPING_DISCONNECTED_NODE_DISTANCE) {
		C_DEBUG(MAPPING_DEBUG_LEVEL, C_MSG_VERBOSE, "Trying to retrieve distance between disconnected nodes %s and %s, returning network diameter (%d hops)...\n", addr_to_str(src_address), addr_to_str(dst_address), C_MAXIMAL_NETWORK_DIAMETER);
		return (float) C_MAXIMAL_NETWORK_DIAMETER;
	}

	return (float) distance;
}

#endif	// C_USE_NETWORK_MAPPING
