/**
 * @file stats.c
 *
 * @brief Functions for processing and output of statistics.
 *
 * Copyright (c) 2010 Georg Wittenburg, Freie Universitaet Berlin
 * (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: stats.c 3465 2010-04-21 19:16:14Z gwitten $
 */

#include <float.h>
#include <math.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "callback.h"
#include "debug.h"
#include "list.h"
#include "mapping.h"
#include "net.h"
#include "stats.h"
#include "../dymo/dymo.h"
#include "../icmp-light/icmp-light.h"
#include "../rudp-light/rudp-light.h"
#include "../slp-light/slp-light.h"
#include "../service/service.h"
#include "../placement/placement.h"



/**
 * @brief Debug level for this component.
 */
#define STATS_DEBUG_LEVEL	C_MSG_INFO

/**
 * @brief Flag whether to print naked stats or not.
 */
#define STATS_PRINT_NAKED_STATS



#ifdef C_USE_NETWORK_MAPPING
/**
 * @brief Update the aggregate network statistics counters.
 */
static void calculate_network_statistics() {
	list_iterator_t C_known_nodes_iterator;
	node_t* node = NULL;
	double sum_of_distances = 0.0;
	unsigned int connected_nodes = 0;

	list_iterator_init(C_known_nodes, &C_known_nodes_iterator);
	while((node = (node_t*) list_iterator_next(&C_known_nodes_iterator)) != NULL) {

// TODO: Skip local node for neighborhood calculation.
// We started the current evaluation without this snipped, so connected_nodes
// is off by one. This affects the measurements for C_average_node_distance and
// any evaluation needs to take this into account.
/*
		if(node->address == C_LOCAL_ADDRESS)
			continue;
*/

		if(node->minimal_hop_count > C_MAXIMAL_NETWORK_DIAMETER)
			continue;
 
		sum_of_distances += node->minimal_hop_count;
		connected_nodes++;
		if(node->minimal_hop_count <= 1) {
			C_DEBUG(STATS_DEBUG_LEVEL, C_MSG_VERBOSE, "%s and %s are direct neighbors.\n", addr_to_str(C_LOCAL_ADDRESS), addr_to_str(node->address));
			C_node_degree_running_sum++;
		}
	}
	if(connected_nodes > 0)
		C_node_distance_running_sum += sum_of_distances / ((double) connected_nodes);

	if((C_state == C_STATE_INITIALIZED) || (C_state == C_STATE_RUNNING)) {
		C_ABORT_ON(C_periodic_statistics_measurement_samples == 0, "Node is initialized or running (%d), but no measurement samples have been collected.\n", C_state);
		C_average_node_degree = ((double) C_node_degree_running_sum) / ((double) C_periodic_statistics_measurement_samples);
		C_average_node_distance = C_node_distance_running_sum / ((double) C_periodic_statistics_measurement_samples);
	}
}
#endif	// C_USE_NETWORK_MAPPING

/**
 * @brief Function to perform periodic statistics measurements.
 */
void periodic_statistics_measurement() {
	C_periodic_statistics_measurement_samples++;

	C_transmitted_datarate = calculate_datarate(C_transmitted_datarate_stat);
	C_received_datarate = calculate_datarate(C_received_datarate_stat);
#ifdef C_USE_NETWORK_MAPPING
	calculate_network_statistics();
#endif	// C_USE_NETWORK_MAPPING
	SERVICE_calculate_statistics();

#ifndef ARCH_NS2
	print_statistics(false, false, true);
#endif	// ARCH_NS2
}

/**
 * @brief Callback function for periodic statistics measurements.
 *
 * @param parameter Callback parameter (unused).
 */
void periodic_statistics_measurement_callback(callback_parameter_t parameter) {
	periodic_statistics_measurement();
	C_periodic_statistics_measurement_callback_id = callback_add(periodic_statistics_measurement_callback, NULL, C_PERIODIC_MEASUREMENT_INTERVAL);
}

#ifndef ARCH_NS2

#include <locale.h>

/**
 * @brief Convert an integer to a string with thousands separators.
 *
 * Adapted from http://www.eskimo.com/~scs/c-faq.com/stdio/commaprint.html.
 * Copyright 1995-2005 by Steve Summit, apparently free to use.
 *
 * NOTE: The string is stored in a internal buffer. Only call once per output!
 *
 * @param n The integer to be converted.
 */
static char* thousands(long n) {
	static char thousands_sep = '\0';
	static char retbuf[32];
	char *p = &retbuf[sizeof(retbuf) - 1];
	int i = 0;
	bool is_negative = false;

	// Initialize thousands_sep on first run.
	if(thousands_sep == '\0') {
		struct lconv *lcp = localeconv();
		if(
			(lcp != NULL)
			&& (lcp->thousands_sep != NULL)
			&& (*lcp->thousands_sep != '\0')
		) {
			thousands_sep = *lcp->thousands_sep;
		} else {
			thousands_sep = ',';
		}
	}

	if(n < 0) {
		n = n * -1;
		is_negative = true;
	}

	*p = '\0';
	do {
		if(((i % 3) == 0) && (i != 0))
			*--p = thousands_sep;
		*--p = '0' + n % 10;
		n /= 10;
		i++;
	} while(n != 0);

	if(is_negative)
		*--p = '-';

	return p;
}

/**
 * @brief Convert a float to a string with thousands separators.
 *
 * Adapted from http://www.eskimo.com/~scs/c-faq.com/stdio/commaprint.html.
 * Copyright 1995-2005 by Steve Summit, apparently free to use.
 *
 * NOTE: The string is stored in a internal buffer. Only call once per output!
 *
 * @param d The float to be converted.
 * @param precision The precision to be printed in the string.
 */
static char* fthousands(double d, unsigned int precision) {
	static char decimal_point = '\0';
	static char thousands_sep = '\0';
	static char retbuf[32];
	char *p = &retbuf[sizeof(retbuf) - 1];
	long n = 0;
	int i = 0;
	bool is_negative = false;

	// Initialize decimal_point and thousands_sep on first run.
	if((decimal_point == '\0') || (thousands_sep == '\0')) {
		struct lconv *lcp = localeconv();
		decimal_point = '.';
		thousands_sep = ',';
		if(lcp != NULL) {
			if((lcp->decimal_point != NULL) && (*lcp->decimal_point != '\0'))
				decimal_point = *lcp->decimal_point;
			if((lcp->thousands_sep != NULL) && (*lcp->thousands_sep != '\0'))
				thousands_sep = *lcp->thousands_sep;
		}
	}

	if(d < 0.0) {
		d = d * -1;
		is_negative = true;
	}

	*p = '\0';
	if(precision > 0) {
		n = (long) ((d - floor(d)) * pow (10.0, precision));
		do {
			*--p = '0' + n % 10;
			precision--;
		} while(precision > 0);
		*--p = decimal_point;
	}
	n = (long) floor(d);
	do {
		if(((i % 3) == 0) && (i != 0))
			*--p = thousands_sep;
		*--p = '0' + n % 10;
		n /= 10;
		i++;
	} while(n != 0);

	if(is_negative)
		*--p = '-';

	return p;
}

/**
 * @brief Print statistics as they are being collected.
 *
 * @param print_readable_ouput Flag whether to print readable stats or not.
 * @param print_parseable_header Flag whether to print parseable stats header
 * or not.
 * @param print_parseable_results Flag whether to print parseable stats or not.
 */
void print_statistics(bool print_readable_ouput, bool print_parseable_header, bool print_parseable_results) {

	float average_unicast_hop_count = 0.0;
	unsigned int maximal_unicast_hop_count = 0;
	float unicast_pdr = 0.0;
	float broadcast_pdr = 0.0;
	float average_transmitted_datarate = 0.0;
	float average_received_datarate = 0.0;
	long overhead_wrt_service_traffic_absolute = 0;
	float overhead_wrt_service_traffic_relative = 0.0;
	float DYMO_unicast_pdr = 0.0;
	float DYMO_broadcast_pdr = 0.0;
	float RUDP_pdr = 0.0;
	float ICMP_nop_pdr = 0.0;
	float SLP_recall = 0.0;
	float SERVICE_data_request_recall = 0.0;
	float SERVICE_average_successful_data_request_delay = 0.0;
	float SERVICE_data_update_recall = 0.0;
	float SERVICE_stream_request_recall = 0.0;
	float SERVICE_stream_packet_recall = 0.0;
	float SERVICE_request_recall = 0.0;
	float SERVICE_average_transmitted_datarate = 0.0;
	float SERVICE_average_received_datarate = 0.0;
	float SERVICE_average_request_hop_count = 0.0;
	unsigned int SERVICE_maximal_request_hop_count = 0;
	float SERVICE_sync_overhead_relative = 0.0;
	float SERVICE_replication_success_rate = 0.0;
	float SERVICE_migration_success_rate = 0.0;
	float PLACEMENT_adaptation_success_rate = 0.0;

	unsigned int i = 0;
	C_time_t current_time = (C_time() - C_init_time);
	C_time_t total_duration = current_time;
	unsigned int total_unicast_packets_with_hop_count = 0;
	unsigned int SERVICE_total_request_packets_with_hop_count = 0;

#ifndef STATS_PRINT_NAKED_STATS
	if(!print_readable_ouput && !print_parseable_header)
		print_parseable_results = false;
#endif	// STATS_PRINT_NAKED_STATS

	if(C_stop_time > 0.0) {		// Only count the active phase of the run towards the total duration.
		total_duration = (C_stop_time - C_init_time);
		total_duration -= 60.0;	// Reduce duration to match the (incorrect) behavior of the simulation scripts. TODO: Fix this.
	}
	for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++) {
		average_unicast_hop_count += (C_unicast_packets_per_hop_count[i] * i);
		total_unicast_packets_with_hop_count += C_unicast_packets_per_hop_count[i];
		if(C_unicast_packets_per_hop_count[i] > maximal_unicast_hop_count)
			maximal_unicast_hop_count = i;
	}
	if(total_unicast_packets_with_hop_count > 0)
		average_unicast_hop_count = (average_unicast_hop_count / ((float) total_unicast_packets_with_hop_count));
	if(C_transmitted_unicast_packets > 0)
		unicast_pdr = (((float) C_received_unicast_packets) / ((float) C_transmitted_unicast_packets));
	if(C_transmitted_broadcast_packets > 0)
		broadcast_pdr = (((float) C_received_broadcast_packets) / ((float) C_transmitted_broadcast_packets));
	average_transmitted_datarate = (((float) C_transmitted_bytes) / total_duration);
	average_received_datarate = (((float) C_received_bytes) / total_duration);
	overhead_wrt_service_traffic_absolute = (C_transmitted_bytes - C_SERVICE_transmitted_bytes);
	if(C_transmitted_bytes > 0)
		overhead_wrt_service_traffic_relative = (((float) overhead_wrt_service_traffic_absolute) / ((float) C_transmitted_bytes));
	if(C_DYMO_transmitted_unicast_packets > 0)
		DYMO_unicast_pdr = (((float) C_DYMO_received_unicast_packets) / ((float) C_DYMO_transmitted_unicast_packets));;
	if(C_DYMO_transmitted_broadcast_packets > 0)
		DYMO_broadcast_pdr = (((float) C_DYMO_received_broadcast_packets) / ((float) C_DYMO_transmitted_broadcast_packets));;
	if(C_RUDP_total_transmissions > 0)
		RUDP_pdr = (((float) C_RUDP_successful_transmissions) / ((float) C_RUDP_total_transmissions));;
	if(C_ICMP_transmitted_nops > 0)
		ICMP_nop_pdr = (((float) C_ICMP_received_nops) / ((float) C_ICMP_transmitted_nops));;
	if(C_SLP_total_queries > 0)
		SLP_recall = (((float) C_SLP_successful_queries) / ((float) C_SLP_total_queries));;
	if(C_SERVICE_total_data_requests > 0) {
		SERVICE_data_request_recall = (((float) C_SERVICE_successful_data_requests) / ((float) C_SERVICE_total_data_requests));;
		SERVICE_average_successful_data_request_delay = (C_SERVICE_total_successful_data_request_delay / ((float) C_SERVICE_total_data_requests));
	}
	if(C_SERVICE_total_data_updates > 0)
		SERVICE_data_update_recall = (((float) C_SERVICE_successful_data_updates) / ((float) C_SERVICE_total_data_updates));;
	if(C_SERVICE_total_stream_requests > 0)
		SERVICE_stream_request_recall = (((float) C_SERVICE_successful_stream_requests) / ((float) C_SERVICE_total_stream_requests));;
	if(C_SERVICE_expected_stream_packets > 0)
		SERVICE_stream_packet_recall = (((float) C_SERVICE_received_stream_packets) / ((float) C_SERVICE_expected_stream_packets));;
	if(C_SERVICE_total_requests > 0)
		SERVICE_request_recall = (((float) C_SERVICE_successful_requests) / ((float) C_SERVICE_total_requests));;

	SERVICE_average_transmitted_datarate = (((float) C_SERVICE_transmitted_bytes) / total_duration);
	SERVICE_average_received_datarate = (((float) C_SERVICE_received_bytes) / total_duration);
	for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++) {
		SERVICE_average_request_hop_count += (C_SERVICE_request_packets_per_hop_count[i] * i);
		SERVICE_total_request_packets_with_hop_count += C_SERVICE_request_packets_per_hop_count[i];
		if(C_SERVICE_request_packets_per_hop_count[i] > SERVICE_maximal_request_hop_count)
			SERVICE_maximal_request_hop_count = i;
	}
	if(SERVICE_total_request_packets_with_hop_count > 0)
		SERVICE_average_request_hop_count = (SERVICE_average_request_hop_count / ((float) SERVICE_total_request_packets_with_hop_count));
	if(C_SERVICE_transmitted_bytes > 0)
		SERVICE_sync_overhead_relative = (((float) C_SERVICE_sync_overhead_bytes) / ((float) C_SERVICE_transmitted_bytes));;
	if(C_SERVICE_total_replications > 0)
		SERVICE_replication_success_rate = (((float) C_SERVICE_successful_replications) / ((float) C_SERVICE_total_replications));;
	if(C_SERVICE_total_migrations > 0)
		SERVICE_migration_success_rate = (((float) C_SERVICE_successful_migrations) / ((float) C_SERVICE_total_migrations));;
	if(C_PLACEMENT_total_adaptations > 0)
		PLACEMENT_adaptation_success_rate = (((float) C_PLACEMENT_successful_adaptations) / ((float) C_PLACEMENT_total_adaptations));;

	if(print_readable_ouput) {
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "ERRORs:\t\t\t\t\t%d\n", C_DEBUG_ERRORs);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "WARNings:\t\t\t\t%d\n", C_DEBUG_WARNs);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Run duration:\t\t\t\t%.1fs\n", total_duration);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Average (mean) node degree:\t\t%.2f\n", C_average_node_degree);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Average (mean) node distance:\t\t%.2f hop(s)\n", C_average_node_distance);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Average (mean) unicast hop count:\t%.2f hop(s)\n", average_unicast_hop_count);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Maximal unicast hop count:\t\t%d hop(s)\n", maximal_unicast_hop_count);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Unicast packets by hop counts:\t\t%d", C_unicast_packets_per_hop_count[0]);
		for(i = 1; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_unicast_packets_per_hop_count[i]);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
//		for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
//			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Unicast packets with hop count %d:\t%s\n", i, thousands(C_unicast_packets_per_hop_count[i]));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Transmitted unicast packets:\t\t%s\n", thousands(C_transmitted_unicast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Unicast transmission failures:\t\t%s\n", thousands(C_unicast_transmission_failures));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Received unicast packets:\t\t%s\n", thousands(C_received_unicast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Unicast packet delivery ratio:\t\t%.2f%%\n", (unicast_pdr * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Transmitted broadcast packets:\t\t%s\n", thousands(C_transmitted_broadcast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Received broadcast packets:\t\t%s\n", thousands(C_received_broadcast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Broadcast packet delivery ratio:\t%.2f%%\n", (broadcast_pdr * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Total traffic (sent):\t\t\t%s bytes\n", thousands(C_transmitted_bytes));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Total traffic (received):\t\t%s bytes\n", thousands(C_received_bytes));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Current datarate (sent):\t\t%s bytes/s\n", thousands(C_transmitted_datarate));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Current datarate (received):\t\t%s bytes/s\n", thousands(C_received_datarate));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Average datarate (sent):\t\t%s bytes/s\n", fthousands(average_transmitted_datarate, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Average datarate (received):\t\t%s bytes/s\n", fthousands(average_received_datarate, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Overhead (wrt. SERVICE, absolute):\t%s bytes\n", thousands(overhead_wrt_service_traffic_absolute));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Overhead (wrt. SERVICE, relative):\t%.2f%%\n", (overhead_wrt_service_traffic_relative * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Network mapping overhead (absolute):\t%s bytes\n", thousands(C_network_mapping_overhead));
		// NOTE: network_mapping_overhead_relative is not included because it requires data about layer 2 packets.
		// C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Network mapping overhead (relative):\t%.2f%%\n", (network_mapping_overhead_relative * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO transmitted route requests (RREQ):\t%s\n", thousands(C_DYMO_transmitted_RREQs));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO transmitted route replies (RREP):\t%s\n", thousands(C_DYMO_transmitted_RREPs));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO transmitted route errors (RERR):\t%s\n", thousands(C_DYMO_transmitted_RERRs));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO transmitted unicast packets:\t%s\n", thousands(C_DYMO_transmitted_unicast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO received unicast packets:\t\t%s\n", thousands(C_DYMO_received_unicast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO unicast packet delivery ratio:\t%.2f%%\n", (DYMO_unicast_pdr * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO transmitted broadcast packets:\t%s\n", thousands(C_DYMO_transmitted_broadcast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO received broadcast packets:\t%s\n", thousands(C_DYMO_received_broadcast_packets));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "DYMO broadcast packet delivery ratio:\t%.2f%%\n", (DYMO_broadcast_pdr * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "RUDP total transmissions:\t\t%s\n", thousands(C_RUDP_total_transmissions));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "RUDP successful transmissions:\t\t%s\n", thousands(C_RUDP_successful_transmissions));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "RUDP packet delivery ratio:\t\t%.2f%%\n", (RUDP_pdr * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "ICMP transmitted NOPs:\t\t\t%s\n", thousands(C_ICMP_transmitted_nops));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "ICMP received NOPs:\t\t\t%s\n", thousands(C_ICMP_received_nops));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "ICMP NOP packet delivery ratio:\t\t%.2f%%\n", (ICMP_nop_pdr * 100.0));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SLP total queries:\t\t\t%s\n", thousands(C_SLP_total_queries));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SLP successful queries:\t\t\t%s\n", thousands(C_SLP_successful_queries));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SLP recall:\t\t\t\t%.2f%%\n", (SLP_recall * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total data requests:\t\t%s\n", thousands(C_SERVICE_total_data_requests));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE successful data requests:\t%s\n", thousands(C_SERVICE_successful_data_requests));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE data request recall:\t\t%.2f%%\n", (SERVICE_data_request_recall * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE avg. succ. data request delay:\t%.4fs\n", SERVICE_average_successful_data_request_delay);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total data updates:\t\t%s\n", thousands(C_SERVICE_total_data_updates));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE successful data updates:\t%s\n", thousands(C_SERVICE_successful_data_updates));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE data update recall:\t\t%.2f%%\n", (SERVICE_data_update_recall * 100.0));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total stream requests:\t\t%s\n", thousands(C_SERVICE_total_stream_requests));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE successful stream requests:\t%s\n", thousands(C_SERVICE_successful_stream_requests));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE stream request recall:\t\t%.2f%%\n", (SERVICE_stream_request_recall * 100.0));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE expected stream packets:\t%s\n", thousands(C_SERVICE_expected_stream_packets));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE received stream packets:\t%s\n", thousands(C_SERVICE_received_stream_packets));
//		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE stream packet recall:\t\t%.2f%%\n", (SERVICE_stream_packet_recall * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total requests:\t\t\t%s\n", thousands(C_SERVICE_total_requests));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE successful requests:\t\t%s\n", thousands(C_SERVICE_successful_requests));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE request recall:\t\t\t%.2f%%\n", (SERVICE_request_recall * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total traffic (sent):\t\t%s bytes\n", thousands(C_SERVICE_transmitted_bytes));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total traffic (received):\t%s bytes\n", thousands(C_SERVICE_received_bytes));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE current datarate (sent):\t%s bytes/s\n", thousands(C_SERVICE_transmitted_datarate));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE current datarate (received):\t%s bytes/s\n", thousands(C_SERVICE_received_datarate));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE average datarate (sent):\t%s bytes/s\n", fthousands(SERVICE_average_transmitted_datarate, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE average datarate (received):\t%s bytes/s\n", fthousands(SERVICE_average_received_datarate, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE average request hop count:\t%.2f hop(s)\n", SERVICE_average_request_hop_count);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE maximal request hop count:\t%d hop(s)\n", SERVICE_maximal_request_hop_count);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE requests by hop counts:\t\t%d", C_SERVICE_request_packets_per_hop_count[0]);
		for(i = 1; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_request_packets_per_hop_count[i]);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
//		for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
//			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE requests with hop count %d:\t%s\n", i, thousands(C_SERVICE_request_packets_per_hop_count[i]));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE currently hosted services:\t%d\n", C_SERVICE_currently_hosted_services);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE average hosted services:\t%.2f\n", C_SERVICE_average_hosted_services);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE available until:\t\t%.2fs\n", C_SERVICE_available_until_time);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE interhost sync overhead (abs.):\t%s bytes\n", thousands(C_SERVICE_sync_overhead_bytes));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE interhost sync overhead (rel.):\t%.2f%%\n", (SERVICE_sync_overhead_relative * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total replications:\t\t%s\n", thousands(C_SERVICE_total_replications));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE successful replications:\t%s\n", thousands(C_SERVICE_successful_replications));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE replications success rate:\t%.2f%%\n", (SERVICE_replication_success_rate * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE total migrations:\t\t%s\n", thousands(C_SERVICE_total_migrations));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE successful migrations:\t\t%s\n", thousands(C_SERVICE_successful_migrations));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "SERVICE migrations success rate:\t%.2f%%\n", (SERVICE_migration_success_rate * 100.0));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT current service cost:\t\t%s\n", fthousands(C_PLACEMENT_current_service_cost, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT curr. optimal service cost:\t%s\n", fthousands(C_PLACEMENT_currently_optimal_service_cost, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT curr. opt. number of hosts:\t%s\n", thousands(C_PLACEMENT_currently_optimal_number_of_hosts));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT curr. est. adaptation cost:\t%s\n", fthousands(C_PLACEMENT_currently_estimated_adaptation_cost, 2));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT total adaptations:\t\t%s\n", thousands(C_PLACEMENT_total_adaptations));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT successful adaptations:\t%s\n", thousands(C_PLACEMENT_successful_adaptations));
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "PLACEMENT adaptations success rate:\t%.2f%%\n", (PLACEMENT_adaptation_success_rate * 100.0));
	}

	if(print_parseable_header) {
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n###\n");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "current_time,nodes,area_x,area_y,setup_duration,run_duration,total_duration,averageNodeDegree,averageNodeDistance,averageUnicastHopCount,maximalUnicastHopCount");
		for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",unicastPacketsPerHopCount_%d", i);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",transmittedUnicastPackets,unicastTransmissionFailures,receivedUnicastPackets,unicastPDR,transmittedBroadcastPackets,receivedBroadcastPackets,broadcastPDR,transmittedBytes,receivedBytes,currentTransmittedDatarate,currentReceivedDatarate,averageTransmittedDatarate,averageReceivedDatarate,overhead_wrt_serviceTraffic_absolute,overhead_wrt_serviceTraffic_relative,networkMappingOverhead_absolute,MAC_transmittedPackets,MAC_transmissionFailures,MAC_receivedPackets,MAC_droppedPackets_capture,MAC_droppedPackets_collision,MAC_droppedPackets_transmit,MAC_droppedPackets_bit_error,MAC_transmittedBytes,MAC_receivedBytes,MAC_averageTransmittedDatarate,MAC_averageReceivedDatarate,MAC_overhead_wrt_serviceTraffic_absolute,MAC_overhead_wrt_serviceTraffic_relative,DYMO_transmittedRREQs,DYMO_transmittedRREPs,DYMO_transmittedRERRs,DYMO_transmittedUnicastPackets,DYMO_receivedUnicastPackets,DYMO_unicastPdr,DYMO_transmittedBroadcastPackets,DYMO_receivedBroadcastPackets,DYMO_broadcastPdr,RUDP_totalTransmissions,RUDP_successfulTransmissions,RUDP_pdr,ICMP_transmittedNops,ICMP_receivedNops,ICMP_nopPdr,SLP_totalQueries,SLP_successfulQueries,SLP_recall,SERVICE_totalDataRequests,SERVICE_successfulDataRequests,SERVICE_dataRequestRecall,SERVICE_averageSuccessfulDataRequestDelay,SERVICE_totalDataUpdates,SERVICE_successfulDataUpdates,SERVICE_dataUpdateRecall,SERVICE_totalStreamRequests,SERVICE_successfulStreamRequests,SERVICE_streamRequestRecall,SERVICE_expectedStreamPackets,SERVICE_receivedStreamPackets,SERVICE_streamPacketRecall,SERVICE_totalRequests,SERVICE_successfulRequests,SERVICE_requestRecall,SERVICE_transmittedBytes,SERVICE_receivedBytes,SERVICE_currentTransmittedDatarate,SERVICE_currentReceivedDatarate,SERVICE_averageTransmittedDatarate,SERVICE_averageReceivedDatarate,SERVICE_averageRequestHopCount,SERVICE_maximalRequestHopCount");
		for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",SERVICE_requestPacketsPerHopCount_%d", i);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",SERVICE_currentlyHostedServices,SERVICE_averageHostedServices,SERVICE_availableUntilTime,SERVICE_syncOverhead_absolute,SERVICE_syncOverhead_relative,SERVICE_totalReplications,SERVICE_successfulReplications,SERVICE_replicationSuccessRate,SERVICE_totalMigrations,SERVICE_successfulMigrations,SERVICE_migrationSuccessRate,PLACEMENT_currentServiceCost,PLACEMENT_currentlyOptimalServiceCost,PLACEMENT_currentlyOptimalNumberOfHosts,PLACEMENT_currentlyEstimatedAdaptationCost,PLACEMENT_totalAdaptations,PLACEMENT_successfulAdaptations,PLACEMENT_adaptationSuccessRate\n");
	}

	if(print_parseable_results) {
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "%f", current_time);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",n/a");	// nodes
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",n/a");	// area_x
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",n/a");	// area_y
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",n/a");	// setup_duration
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",n/a");	// run_duration
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", total_duration);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_average_node_degree);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_average_node_distance);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", average_unicast_hop_count);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", maximal_unicast_hop_count);
		for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_unicast_packets_per_hop_count[i]);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_transmitted_unicast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_unicast_transmission_failures);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_received_unicast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", unicast_pdr);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_transmitted_broadcast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_received_broadcast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", broadcast_pdr);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_transmitted_bytes);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_received_bytes);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_transmitted_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_received_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", average_transmitted_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", average_received_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", overhead_wrt_service_traffic_absolute);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", overhead_wrt_service_traffic_relative);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_network_mapping_overhead);
		// NOTE: See above for network_mapping_overhead_relative.
		// Print "n/a" for all MAC-related values: MAC_transmittedPackets,MAC_transmissionFailures,MAC_receivedPackets,MAC_droppedPackets_capture,MAC_droppedPackets_collision,MAC_droppedPackets_transmit,MAC_droppedPackets_bit_error,MAC_transmittedBytes,MAC_receivedBytes,MAC_averageTransmittedDatarate,MAC_averageReceivedDatarate,MAC_overhead_wrt_serviceTraffic_absolute,MAC_overhead_wrt_serviceTraffic_relative
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",n/a,n/a,n/a,n/a,n/a,n/a,n/a,n/a,n/a,n/a,n/a,n/a,n/a");
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_transmitted_RREQs);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_transmitted_RREPs);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_transmitted_RERRs);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_transmitted_unicast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_received_unicast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", DYMO_unicast_pdr);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_transmitted_broadcast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_DYMO_received_broadcast_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", DYMO_broadcast_pdr);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_RUDP_total_transmissions);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_RUDP_successful_transmissions);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", RUDP_pdr);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_ICMP_transmitted_nops);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_ICMP_received_nops);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", ICMP_nop_pdr);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SLP_total_queries);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SLP_successful_queries);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SLP_recall);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_total_data_requests);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_successful_data_requests);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_data_request_recall);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_average_successful_data_request_delay);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_total_data_updates);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_successful_data_updates);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_data_update_recall);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_total_stream_requests);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_successful_stream_requests);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_stream_request_recall);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_expected_stream_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_received_stream_packets);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_stream_packet_recall);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_total_requests);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_successful_requests);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_request_recall);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_transmitted_bytes);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_received_bytes);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_transmitted_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_received_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_average_transmitted_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_average_received_datarate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_average_request_hop_count);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", SERVICE_maximal_request_hop_count);
		for(i = 0; i <= C_MAXIMAL_NETWORK_DIAMETER; i++)
			C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_request_packets_per_hop_count[i]);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_currently_hosted_services);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_SERVICE_average_hosted_services);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_SERVICE_available_until_time);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_sync_overhead_bytes);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_sync_overhead_relative);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_total_replications);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_successful_replications);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_replication_success_rate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_total_migrations);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_SERVICE_successful_migrations);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", SERVICE_migration_success_rate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_PLACEMENT_current_service_cost);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_PLACEMENT_currently_optimal_service_cost);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_PLACEMENT_currently_optimal_number_of_hosts);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", C_PLACEMENT_currently_estimated_adaptation_cost);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_PLACEMENT_total_adaptations);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%d", C_PLACEMENT_successful_adaptations);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), ",%f", PLACEMENT_adaptation_success_rate);
		C_DEBUG(STATS_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
	}
}
#endif	// ARCH_NS2
