/**
 * @file vars.h
 *
 * @brief Global variables and constants.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: vars.h 3505 2010-05-12 16:54:05Z gwitten $
 */

#ifndef VARS_H
#define VARS_H



#include "config.h"
#include "types.h"
#include "rudp-light/rudp-light_vars.h"



#define C_STATE_INITIALIZING		0x00	///< Identifier for initializing state.
#define C_STATE_INITIALIZED		0x01	///< Identifier for initialized state.
#define C_STATE_RUNNING			0x02	///< Identifier for running state.
#define C_STATE_STOPPED			0x03	///< Identifier for stopped state.
#define C_STATE_FINALIZED		0x04	///< Identifier for finalized state.



#define C_ERROR_NOT_SET			0x00	///< Error code not set (must not happen).
#define C_ERROR_OK			0x01	///< No error, everything OK.
#define C_ERROR_PARAMETER		0x02	///< Error due to invalid parameter.
#define C_ERROR_NOT_FOUND		0x03	///< Error due to requested resource not being found.
#define C_ERROR_TIMEOUT			0x04	///< Error due to timeout.
#define C_ERROR_ABORTED			0x05	///< Error due to component being shutdown.
#define C_ERROR_OUT_OF_MEMORY		0x06	///< Error due to lack of memory.
#define C_ERROR_BUSY			0x07	///< Error due to a component being busy.
#define C_ERROR_IN_PROGRESS		0x08	///< Error due to a similar request already being in progress.
#define C_ERROR_COMM_ERROR		0x09	///< Error due to a communication error.
#define C_ERROR_REMOTE_BUSY		0x10	///< Error due to the remote component / node being busy.



#define C_MSG_PRIORITY	0x0F	///< Bitmask for message priority.
#define C_MSG_MODIFIER	0xF0	///< Bitmask for message modifier.

#define C_MSG_VERBOSE	0x00	///< Identifier for verbose debug output.
#define C_MSG_DEBUG	0x01	///< Identifier for debug output.
#define C_MSG_INFO	0x02	///< Identifier for regular information.
#define C_MSG_NOTE	0x03	///< Identifier for important notes.
#define C_MSG_WARN	0x04	///< Identifier for warnings.
#define C_MSG_ERROR	0x05	///< Identifier for regular errors.
#define C_MSG_FATAL	0x06	///< Identifier for fatal errors.

#define C_MSG_CONT	0x10	///< Modifier for continued output.



/**
 * @brief Number of ERRORs encountered during execution.
 */
uint32_t C_DEBUG_ERRORs;

/**
 * @brief Number of WARNings encountered during execution.
 */
uint32_t C_DEBUG_WARNs;



/**
 * @brief Current state of this component.
 */
uint8_t C_state;

/**
 * @brief Time at which this component was initialized.
 */
C_time_t C_init_time;

/**
 * @brief Time at which this component was stopped.
 */
C_time_t C_stop_time;

/**
 * @brief Flag whether consistency checks should be performed when shutting
 * down.
 */
bool C_check_consistency_on_shutdown;

/**
 * @brief List of pending callbacks.
 */
list_t* C_callbacks;

/**
 * @brief Unique identifier for the next callback.
 *
 * The value 0 is not used, but rather reserved to tag unused callbacks.
 */
callback_id_t C_next_callback_id;

/**
 * @brief Number of registered callbacks.
 */
uint32_t C_registered_callbacks;

/**
 * @brief Number of executed callbacks.
 */
uint32_t C_executed_callbacks;



/**
 * @brief Convenience macro for address of the local node.
 */
#define C_LOCAL_ADDRESS			C_local_address

/**
 * @brief Address of the local node.
 */
C_addr_t C_local_address;

/**
 * @brief Convenience macro for broadcast address.
 */
#define C_BROADCAST_ADDRESS		C_broadcast_address

/**
 * @brief Address of the local node.
 */
C_addr_t C_broadcast_address;



/**
 * @brief Size of buffer for converting numeric addresses to a string
 * representation.
 *
 * With a buffer size of 8 we are definately on the safe side. Less is
 * probably OK, too, especially when running on memory constrained devices.
 */
#define C_ADDR_TO_STR_BUFFER_SIZE	8

/**
 * @brief Length of strings in buffer for converting numeric addresses to a
 * string representation.
 *
 * This has to be big enough to support an IPv4 address in the dotted decimal
 * notation, but may be smaller on platforms with other addressing schemes.
 */
#define C_ADDR_TO_STR_BUFFER_STR_LEN	16

/**
 * @brief Buffer for converting numeric addresses to a string representation.
 *
 * Multiple address strings may need to be valid at the same point in time,
 * e.g. when printing debugging output. Hence, we need a buffer with multiple
 * slots to avoid overwriting strings that still need to be processed.
 */
char C_addr_to_str_buffer[C_ADDR_TO_STR_BUFFER_SIZE][C_ADDR_TO_STR_BUFFER_STR_LEN];

/**
 * @brief Currently used index of buffer for converting numeric addresses to a
 * string representation.
 */
uint8_t C_addr_to_str_buffer_index;



#define PACKET_PRIORITY_DATA			0x01	///< Priority of application data packet.
#define PACKET_PRIORITY_SIGNALLING_SERVICE	0x02	///< Priority of service-level signalling packet.
#define PACKET_PRIORITY_SIGNALLING_PLACEMENT	0x03	///< Priority of placement-level signalling packet.
#define PACKET_PRIORITY_SIGNALLING_TRANSPORT	0x04	///< Priority of transport-level signalling packet.
#define PACKET_PRIORITY_SIGNALLING_ROUTING	0x05	///< Priority of network-level signalling packet.
#define PACKET_PRIORITY_SIGNALLING_LINK		0x06	///< Priority of link-level signalling packet.
#define PACKET_PRIORITY_MIN			PACKET_PRIORITY_DATA			///< Minimal packet priority.
#define PACKET_PRIORITY_MAX			PACKET_PRIORITY_SIGNALLING_LINK		///< Maximal packet priority.

#define APP_TYPE_NONE		0x0	///< Packets not belonging to any component.
#define APP_TYPE_INTERNAL	0x1	///< Internal packets.
#define APP_TYPE_DYMO		0x2	///< Packets for the DYMO component.
#define APP_TYPE_ICMP		0x3	///< Packets for the ICMP component.
#define APP_TYPE_RUDP		0x4	///< Packets for the RUDP component.
#define APP_TYPE_SLP		0x5	///< Packets for the SLP component.
#define APP_TYPE_SERVICE	0x6	///< Packets for the SERVICE component.
#define APP_TYPE_PLACEMENT	0x7	///< Packets for the PLACEMENT component.
#define APP_TYPE_INTERFERENCE	0x8	///< Packets for the INTERFERENCE component.

#define PACKET_TYPE_BEACON	0x0	///< BEACON packet type.



/**
 * @brief Current sequence number for transmitted packets.
 */
uint32_t C_sequence_number;

/**
 * @brief Callback ID of the send callback.
 */
callback_id_t C_send_callback_id;

/**
 * @brief Size of cache for recently received packets.
 */
#define C_RECEIVED_PACKET_CACHE_SIZE	1024

/**
 * @brief Cache for recently received packets.
 */
packet_fingerprint_t C_received_packet_cache[C_RECEIVED_PACKET_CACHE_SIZE];



/**
 * @brief Value for an unknown node ID.
 *
 * This value should correspond to the variable type of
 * <code>node_address_to_id()</code>.
 *
 * NOTE: A node ID is used for internal node identification and does NOT
 * correspond to the node's address.
 */
#define C_UNKNOWN_NODE_ID		0xFFFF

/**
 * @brief List of known nodes.
 */
list_t* C_known_nodes;

/**
 * @brief Network map with support for aging information about links.
 *
 * Two-dimensional array indexed by node ID. The value is the time of the last
 * known communication between two neighboring nodes.
 */
C_time_t* C_network_map;

/**
 * @brief Flag that the network map may need purging.
 *
 * For efficiency, the network map is only be recalculated once per incoming
 * event. This optimization is valid, since no new information becomes
 * available during event (e.g., packet) processing.
 */
bool C_network_map_may_need_purging;

/**
 * @brief Callback ID of the beaconing callback.
 */
callback_id_t C_beacon_callback_id;



/**
 * @brief Matrix of edge costs for the currently known network graph.
 */
uint8_t* C_edge_cost_matrix;

/**
 * @brief Matrix of distances for the currently known network graph.
 */
uint8_t* C_distance_matrix;



#ifndef ARCH_NS2

/**
 * @brief Default UDP port number.
 *
 * NOTE: According to http://www.iana.org/assignments/port-numbers and
 * http://tools.ietf.org/html/draft-ietf-manet-iana-07 we should use port 269.
 * However, this would require root privileges to execute the program, which
 * is unpractical for testing.
 */
#define C_DEFAULT_UDP_PORT	10269

/**
 * @brief UDP socket handle.
 */
int C_udp_socket;

#endif	// ARCH_NS2



/**
 * @brief Callback ID of the the periodic statistics measurement callback.
 */
callback_id_t C_periodic_statistics_measurement_callback_id;

/**
 * @brief Number of samples taken for the periodic statistics measurements.
 */
uint32_t C_periodic_statistics_measurement_samples;

/**
 * @brief Statistics counter for running sum of degree of this node.
 */
uint32_t C_node_degree_running_sum;

/**
 * @brief Statistics counter for average (over time) degree of this node.
 */
double C_average_node_degree;

/**
 * @brief Statistics counter for running sum of the lengths of all routes
 * leading to this node.
 */
double C_node_distance_running_sum;

/**
 * @brief Statistics counter for average (over time) length of all routes
 * leading to this node.
 */
double C_average_node_distance;

/**
 * @brief Statistics counter array to count received layer-3 unicast packets
 * depending on their hop count.
 *
 * The size of the array is C_MAXIMAL_NETWORK_DIAMETER + 1 to make the
 * processing more natural for packets with hop_count = 0.
 */
uint32_t C_unicast_packets_per_hop_count[C_MAXIMAL_NETWORK_DIAMETER + 1];

/**
 * @brief Statistics counter for transmitted unicast packets.
 */
uint32_t C_transmitted_unicast_packets;

/**
 * @brief Statistics counter for failed transmissions.
 */
uint32_t C_unicast_transmission_failures;

/**
 * @brief Statistics counter for received unicast packets.
 */
uint32_t C_received_unicast_packets;

/**
 * @brief Statistics counter for transmitted broadcast packets.
 */
uint32_t C_transmitted_broadcast_packets;

/**
 * @brief Statistics counter for received broadcast packets.
 */
uint32_t C_received_broadcast_packets;

/**
 * @brief Statistics counter for transmitted bytes.
 */
uint32_t C_transmitted_bytes;

/**
 * @brief Statistics counter for received bytes.
 */
uint32_t C_received_bytes;

/**
 * @brief Internal statistics struct for transmitted datarate.
 */
datarate_stat_t* C_transmitted_datarate_stat;

/**
 * @brief Statistics counter for transmitted datarate.
 */
uint32_t C_transmitted_datarate;

/**
 * @brief Internal statistics struct for transmitted datarate.
 */
datarate_stat_t* C_received_datarate_stat;

/**
 * @brief Statistics counter for transmitted datarate.
 */
uint32_t C_received_datarate;

/**
 * @brief Statistics counter for total overhead caused by network mapping in bytes.
 */
uint32_t C_network_mapping_overhead;

#endif	// VARS_H
