#!/usr/bin/perl

##
#
# This script compresses all the subdirectories contained in a directory
# into an intelligently named archive file and computes the MD5 checksum.
#
# Example:	archive_directories.pl parent_directory
#
# Copyright (c) 2010 Georg Wittenburg, Freie Universitaet Berlin
# (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: archive_directories.pl 3429 2010-03-23 12:58:54Z gwitten $
#
##

use strict;

use File::stat;

my $ARCHIVE_CMD = "tar -czf";
my $ARCHIVE_SUFFIX = "tgz";

my $CHECKSUM_CMD = "md5sum -b";
my $CHECKSUM_SUFFIX = "md5";

if(scalar(@ARGV) != 1) {
	print "Usage:\tarchive_directories.pl parent_directory\n\n";
	exit(1);
}

my $parent_directory = $ARGV[0];

if(!-e $parent_directory) {
	print "The directory \"$parent_directory\" does not exist.\n";
	exit(1);
}

print "Archiving subdirectories of $parent_directory...\n";
chdir($parent_directory);

opendir(my $DIR, ".") or die $!;
while(my $directory = readdir($DIR)) {

	# Only consider "normal" directories.
	if(
		(!-d $directory)
		|| ($directory =~ /\A\..*/o)
	) {
		next;
	}

	# Skip directories for which an up-to-date archive already exists.
	my $archive_exists = 0;
	opendir(my $DIR2, ".") or die $!;
	while(my $file = readdir($DIR2)) {
		if(
			(-f $file)
			&& ($file =~ /\A$directory\_.+/)
			&& (stat($file)->mtime > stat($directory)->mtime)
		) {
			print "Skipping directory $directory (up-to-date archive in $file)...\n";
			$archive_exists = 1;
			last;
		}
	}
	closedir($DIR2);
	if($archive_exists) {next;}

	my($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time);
	$sec = ($sec < 10) ? "0$sec" : $sec;
	$min = ($min < 10) ? "0$min" : $min;
	$hour = ($hour < 10) ? "0$hour" : $hour;
	$mday = ($mday < 10) ? "0$mday" : $mday;
	$mon++;
	$mon = ($mon < 10) ? "0$mon" : $mon;
	$year += 1900;
	my $archive_file = "$directory\_$year$mon$mday";
	if(-e "$archive_file.$ARCHIVE_SUFFIX") {
		my $i = 2;
		while(-e "$archive_file\_$i.$ARCHIVE_SUFFIX") {
			$i++;
		}
		$archive_file = "$archive_file\_$i";
	}

	print "Archiving directory $directory (archive file is $archive_file.$ARCHIVE_SUFFIX)...\n";

	print `$ARCHIVE_CMD $archive_file.$ARCHIVE_SUFFIX $directory`;
	if($? != 0) {
		print "Error while archiving directory $directory ($?), aborting...";
		exit(2);
	}

	print "Calculating checksum for archive $archive_file.$ARCHIVE_SUFFIX...\n";

	print `$CHECKSUM_CMD $archive_file.$ARCHIVE_SUFFIX > $archive_file.$ARCHIVE_SUFFIX.$CHECKSUM_SUFFIX`;
	if($? != 0) {
		print "Error while calculating checksum of archive $archive_file.$ARCHIVE_SUFFIX ($?), aborting...";
		exit(2);
	}

}
closedir($DIR);
print "\n";
