#!/usr/bin/perl

##
#
# Generate a topology for evaluations on a (wired) testbed.
#
# Copyright (c) 2010 Georg Wittenburg, Freie Universitaet Berlin
# (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: eval_generate_topology.pl 3489 2010-05-04 10:05:30Z gwitten $
#
##

use strict;

# List of machines is at https://www.mi.fu-berlin.de/w/IT/ServicesStudentPools.
my @HOSTS = (
	# Takustr. 9, K38 "China" (10 hosts)
	"shanghai.imp.fu-berlin.de",
	"peking.imp.fu-berlin.de",
	"guangzhou.imp.fu-berlin.de",
	"harbin.imp.fu-berlin.de",
	"chongqing.imp.fu-berlin.de",
	"shenyang.imp.fu-berlin.de",
	"wuhan.imp.fu-berlin.de",
	"chengdu.imp.fu-berlin.de",
	"tianjin.imp.fu-berlin.de",
	"xian.imp.fu-berlin.de",
	# Compute servers.
# 	"knecht.imp.fu-berlin.de",
	# Takustr. 9, K48 "Wales" (10 hosts)
	"llanberis.imp.fu-berlin.de",
	"corris.imp.fu-berlin.de",
	"bala.imp.fu-berlin.de",
	"tywyn.imp.fu-berlin.de",
	"porthmadog.imp.fu-berlin.de",
	"harlech.imp.fu-berlin.de",
	"ffestiniog.imp.fu-berlin.de",
	"welshpool.imp.fu-berlin.de",
	"aberystwyth.imp.fu-berlin.de",
	"llangollen.imp.fu-berlin.de",
	# Arnimallee 6, Raum 005 (7 hosts)
	"poncelet.imp.fu-berlin.de",
	"aiken.imp.fu-berlin.de",
	"mises.imp.fu-berlin.de",
	"kronecker.imp.fu-berlin.de",
	# "russel.imp.fu-berlin.de",	21.4.10: Unable to resolve host name.
	"banach.imp.fu-berlin.de",
	"bachet.imp.fu-berlin.de",
	# Arnimallee 6, Raum 030 (18 hosts)
	"rellich.imp.fu-berlin.de",
	"lalla.imp.fu-berlin.de",
	"cauchy.imp.fu-berlin.de",
	"breteuil.imp.fu-berlin.de",
	"hayes.imp.fu-berlin.de",
	"euler.imp.fu-berlin.de",
	"rudin.imp.fu-berlin.de",
	"germain.imp.fu-berlin.de",
	"fenchel.imp.fu-berlin.de",
	"hamel.imp.fu-berlin.de",
	"kendall.imp.fu-berlin.de",
	"neumann.imp.fu-berlin.de",
	"runge.imp.fu-berlin.de",
	"kutta.imp.fu-berlin.de",
	"peter.imp.fu-berlin.de",
	"landau.imp.fu-berlin.de",
	"hypathia.imp.fu-berlin.de",
	"jordan.imp.fu-berlin.de",
	# Takustr. 9, K60 "Niedersachsen" (9 hosts)
# 	"wolfsburg.imp.fu-berlin.de",
# 	"hannover.imp.fu-berlin.de",
# 	"salzgitter.imp.fu-berlin.de",
# 	"oldenburg.imp.fu-berlin.de",
# 	"braunschweig.imp.fu-berlin.de",
# 	"emden.imp.fu-berlin.de",
# 	"cuxhaven.imp.fu-berlin.de",
# 	"goslar.imp.fu-berlin.de",
# 	"wilhelmshaven.imp.fu-berlin.de",
);

my $USER = "gwitten";
# Only use public key auth, set connection timeout, and disable X11 forwarding.
my $SSH_OPTIONS = "-o PreferredAuthentications=publickey -o ConnectTimeout=10 -x";

# Different node degrees depending on how many nodes are available.
my %AVERAGE_NODE_DEGREES = (
	16	=> 4.88384664,
	25	=> 5.60304496,
	36	=> 6.04250566,
	49	=> 6.34986811,
	64	=> 6.6596065,
	81	=> 6.8052939,
	100	=> 6.98249473,
);

sub check_host {
	my($host, $check_if_busy, $print_status) = @_;

	`ping -c1 -q $host`;
	if($? != 0) {
		if($print_status) {print "NOT REACHABLE";}
		return 0;
	}

	my $ps_wc_str = `ssh $SSH_OPTIONS $USER\@$host "/bin/ps aux | grep $USER | grep ns | grep -v '>>' | grep -v grep | wc"`;
	if(($? != 256) && ($? != 0)) {
		if($print_status) {print "LOGIN FAILURE";}
		return 0;
	}

	if($check_if_busy) {
		chomp $ps_wc_str;
		$ps_wc_str =~ /(\d+).+\d+.+\d+/o;
		my $ns_count = $1;
		if($ns_count > 0) {
			if($print_status) {print "RUNNING ($ns_count)";}
			return 0;
		}
	}

	my $uptime_str = `ssh $SSH_OPTIONS $USER\@$host uptime`;
	if($? != 0) {
		if($print_status) {print "LOGIN FAILURE";}
		return 0;
	}

	if($check_if_busy) {
		chomp $uptime_str;
		$uptime_str =~ /([\.0-9]*), ([\.0-9]*), ([\.0-9]*)$/o;
		my $load_average_1m = $1;
		my $load_average_5m = $2;
		my $load_average_15m = $3;
		if(
			($load_average_1m > 2.5)
			&& ($load_average_5m > 2.5)
			&& ($load_average_15m > 2.5)
		) {
			if($print_status) {print "BUSY ($load_average_1m, $load_average_5m, $load_average_15m)";}
			return 0;
		}
	}

	if($print_status) {print "OK";}
	return 1;
}

sub average_node_degree {
	my($topology_ref) =  @_;
	my %topology = %{$topology_ref};

	if(scalar(keys %topology) == 0) {
		return 0.0;
	}

	my $average_node_degree = 0.0;
	for my $host1 (keys %topology) {
		for my $host2 (keys %{$topology{$host1}}) {
# TODO: Skip local node for neighborhood calculation.
# We started the current evaluation without this snipped, so connected_nodes
# is off by one. This affects the measurements for C_average_node_distance and
# any evaluation needs to take this into account.
#			if($host1 ne $host2) {
				$average_node_degree++;
#			}
		}
	}
	return ($average_node_degree / scalar(keys %topology));
}

sub all_nodes_connected {
	my($topology_ref) =  @_;
	my %topology = %{$topology_ref};

	if(scalar(keys %topology) == 0) {
		return 0;
	}

	for my $host1 (keys %topology) {
		if(scalar(keys %{$topology{$host1}}) < 1) {
			return 0;
		}
	}
	return 1;
}

if(scalar(@ARGV) != 3) {
	print "Usage:\teval_generate_topology.pl number_of_nodes number_of_topologies\n";
	print "\toutput_directory\n\n";
	exit(1);
}

my $number_of_nodes = $ARGV[0];
my $number_of_topologies = $ARGV[1];
my $output_directory = $ARGV[2];

if(!exists $AVERAGE_NODE_DEGREES{$number_of_nodes}) {
	print "No average node degree has been specified for $number_of_nodes node(s)\n\n";
	exit(2);
}
if(-e $output_directory) {
	print "Output directory \"$output_directory\" already exists.\n\n";
	exit(2);
}

print "\nGenerating $number_of_topologies topology/ies in $output_directory/...\n\n";

print "Checking which hosts are up:\n";
my @hosts;
for my $host (@HOSTS) {
	print "Checking $host... ";
	if(check_host($host, 0, 1)) {
		print ", adding to list...\n";
		push(@hosts, $host);
	} else {
		print ", ignoring...\n";
	}
	if(scalar(@hosts) >= $number_of_nodes) {
		last;
	}
}
if(scalar(@hosts) == $number_of_nodes) {
	print "".scalar(@hosts)." host(s) are available ($number_of_nodes requested).\n\n";
} else {
	print "\n".scalar(@hosts)." host(s) are available, but $number_of_nodes were requested.\n\n";
	exit(2);
}

mkdir($output_directory);
open(HOSTS, ">$output_directory/hosts.txt") or die "Can't open file \"$output_directory/hosts.txt\" for writing";
my @adresses;
for my $host (@hosts) {
	my $host_output = `host $host`;
	$host_output =~ /$host has address (\d+\.\d+\.\d+\.\d+)/;
	print "$host has address $1.\n";
	push(@adresses, $1);
	print HOSTS "$host\n";
}
print "\n";
close(HOSTS);

for my $topology (1..$number_of_topologies) {
	if($topology < 10) {
		$topology = '0'.$topology;
	}
	my $topology_file = "topology_$topology.txt";
	print "Generating topology #$topology...\n";
	my %topology;
	while(
		(average_node_degree(\%topology) < $AVERAGE_NODE_DEGREES{$number_of_nodes})
		|| (!all_nodes_connected(\%topology))
	) {
		my $adress1 = $adresses[rand(@adresses)];
		my $adress2 = $adresses[rand(@adresses)];
		# print "Adding bi-directional link between $host1 and $host2...\n";
		$topology{$adress1}{$adress2} = 1;
		$topology{$adress2}{$adress1} = 1;
	}

	open(TOPOLOGY, ">$output_directory/$topology_file") or die "Can't open file \"$output_directory/$topology_file\" for writing";
	for my $adress1 (@adresses) {
		my @neighbors;
		for my $adress2 (@adresses) {
			if($topology{$adress1}{$adress2} == 1) {
				push(@neighbors, $adress2);
			}
		}
		print TOPOLOGY "$adress1:".join(',', @neighbors)."\n";
	}
	close(TOPOLOGY);
}
print "\n";
