#!/usr/bin/perl

##
#
# This script runs simulations corresponding to all TCL files in a directory on
# a set of hosts. Output of the simulations is written to *.out files with
# the same base name as the TCL script.
#
# Example:	sim_run.pl ~/usr/bin/ns /home/fenn/gwitten/simulation_scripts/
#
# Known bugs: There seems to be a problem with file locking when several
# simulations start at exactly the same time.
#
# Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
# Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: sim_run.pl 3645 2010-07-08 16:21:53Z gwitten $
#
##

use strict;
use Cwd qw(abs_path chdir);
use Fcntl ':flock';

# List of machines is at https://www.mi.fu-berlin.de/w/IT/ServicesStudentPools.
my @HOSTS = (
	# Takustr. 9, K38 "China" (10 hosts)
	"shanghai.imp.fu-berlin.de",
	"peking.imp.fu-berlin.de",
	"guangzhou.imp.fu-berlin.de",
	"harbin.imp.fu-berlin.de",
	"chongqing.imp.fu-berlin.de",
	"shenyang.imp.fu-berlin.de",
	"wuhan.imp.fu-berlin.de",
	"chengdu.imp.fu-berlin.de",
	"tianjin.imp.fu-berlin.de",
	"xian.imp.fu-berlin.de",
	# Compute servers.
	"knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", "knecht.imp.fu-berlin.de", # 8/16 cores
	# Takustr. 9, K48 "Wales" (10 hosts)
# 	"llanberis.imp.fu-berlin.de",
# 	"corris.imp.fu-berlin.de",
# 	"bala.imp.fu-berlin.de",
# 	"tywyn.imp.fu-berlin.de",
# 	"porthmadog.imp.fu-berlin.de",
# 	"harlech.imp.fu-berlin.de",
# 	"ffestiniog.imp.fu-berlin.de",
# 	"welshpool.imp.fu-berlin.de",
# 	"aberystwyth.imp.fu-berlin.de",
# 	"llangollen.imp.fu-berlin.de",
	# Arnimallee 6, Raum 005 (7 hosts)
	"poncelet.imp.fu-berlin.de",
	"aiken.imp.fu-berlin.de",
	"mises.imp.fu-berlin.de",
	"kronecker.imp.fu-berlin.de",
	# "russel.imp.fu-berlin.de",	21.4.10: Unable to resolve host name.
	"banach.imp.fu-berlin.de",
	"bachet.imp.fu-berlin.de",
	# Arnimallee 6, Raum 030 (18 hosts)
	"rellich.imp.fu-berlin.de",
	"lalla.imp.fu-berlin.de",
	"cauchy.imp.fu-berlin.de",
	"breteuil.imp.fu-berlin.de",
	"hayes.imp.fu-berlin.de",
	"euler.imp.fu-berlin.de",
	"rudin.imp.fu-berlin.de",
	"germain.imp.fu-berlin.de",
	"fenchel.imp.fu-berlin.de",
	"hamel.imp.fu-berlin.de",
	"kendall.imp.fu-berlin.de",
	"neumann.imp.fu-berlin.de",
	"runge.imp.fu-berlin.de",
	"kutta.imp.fu-berlin.de",
	"peter.imp.fu-berlin.de",
	"landau.imp.fu-berlin.de",
	"hypathia.imp.fu-berlin.de",
	"jordan.imp.fu-berlin.de",
	# Takustr. 9, K60 "Niedersachsen" (9 hosts)
# 	"wolfsburg.imp.fu-berlin.de",
# 	"hannover.imp.fu-berlin.de",
# 	"salzgitter.imp.fu-berlin.de",
# 	"oldenburg.imp.fu-berlin.de",
# 	"braunschweig.imp.fu-berlin.de",
# 	"emden.imp.fu-berlin.de",
# 	"cuxhaven.imp.fu-berlin.de",
# 	"goslar.imp.fu-berlin.de",
# 	"wilhelmshaven.imp.fu-berlin.de",
);

my $USER = "gwitten";
# Only use public key auth, set connection timeout, and disable X11 forwarding.
my $SSH_OPTIONS = "-o PreferredAuthentications=publickey -o ConnectTimeout=10 -x";
my $SYNC_FILE_NAME = "sim_run.tmp";
my $OUTPUT_RUNNING_SUFFIX = "_out.txt_RUNNING";
my $OUTPUT_DONE_SUFFIX = "_out.txt";
my $VARIABLE_DELIMITER = "%%";

sub print_time {
	my($string) = @_;
	my($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time);
	$sec = ($sec < 10) ? "0$sec" : $sec;
	$min = ($min < 10) ? "0$min" : $min;
	$hour = ($hour < 10) ? "0$hour" : $hour;
	$mday = ($mday < 10) ? "0$mday" : $mday;
	$mon++;
	$mon = ($mon < 10) ? "0$mon" : $mon;
	$year += 1900;
	print "$mday/$mon $hour:$min:$sec\t$string";
}

sub by_run_index {
	my $run_index_a;
	my $run_index_b;
	if($a =~ /_run_(\d+)/) {
		$run_index_a = $1;
	} else {
		return $a cmp $b;
	}
	if($b =~ /_run_(\d+)/) {
		$run_index_b = $1;
	} else {
		return $a cmp $b;
	}
	if($run_index_a == $run_index_b) {
		return $a cmp $b;
	} else {
		return $run_index_a <=> $run_index_b;
	}
}

sub check_host {
	my($host, $check_if_used, $check_if_busy, $print_status) = @_;

	`ping -c1 -q $host`;
	if($? != 0) {
		if($print_status) {print "NOT REACHABLE";}
		return 0;
	}

	my $ps_wc_str = `ssh $SSH_OPTIONS $USER\@$host "/bin/ps aux | grep $USER | grep ns | grep -v '>>' | grep -v grep | wc"`;
	if(($? != 256) && ($? != 0)) {
		if($print_status) {print "LOGIN FAILURE";}
		return 0;
	}

	if($check_if_used) {
		chomp $ps_wc_str;
		$ps_wc_str =~ /(\d+).+\d+.+\d+/o;
		my $ns_count = $1;
		if($ns_count > 0) {
			if($print_status) {print "RUNNING ($ns_count)";}
			return 0;
		}
	}

	my $uptime_str = `ssh $SSH_OPTIONS $USER\@$host uptime`;
	if($? != 0) {
		if($print_status) {print "LOGIN FAILURE";}
		return 0;
	}

	if($check_if_busy) {
		chomp $uptime_str;
		$uptime_str =~ /([\.0-9]*), ([\.0-9]*), ([\.0-9]*)$/o;
		my $load_average_1m = $1;
		my $load_average_5m = $2;
		my $load_average_15m = $3;
		if(
			($load_average_1m > 1.2)
			&& ($load_average_5m > 1.2)
			# && ($load_average_15m > 1.2)
		) {
			if($print_status) {print "BUSY ($load_average_1m, $load_average_5m, $load_average_15m)";}
			return 0;
		}
	}

	if($print_status) {print "OK";}
	return 1;
}

if(scalar(@ARGV) != 2) {
	print "Usage:\tsim_run.pl ns2_executable_with_path absolute_path_to_tcl_directory\n\n";
	exit(1);
}

my $ns2_executable = $ARGV[0];
my $tcl_directory = $ARGV[1];

if($tcl_directory !~ /\/\Z/o) {
	$tcl_directory .= '/';
}
if($tcl_directory =~ / /o) {
	print "ERROR: tcl_directory may not contain whitespaces.\n";
	exit(2);
}

if(abs_path($ENV{PWD}) ne abs_path($tcl_directory)) {
	chdir $tcl_directory;
	if($ENV{PWD} ne $tcl_directory) {
		die "Unable to change working directory to $tcl_directory";
	}
	print "\n";
	print_time "Changing working directory from $ENV{PWD} to $tcl_directory...\n";
}

print "\n";
print_time "Starting simulation using $ns2_executable.\n";
print_time "Simulation files are in $tcl_directory.\n";
print "\n";

if(-e $SYNC_FILE_NAME) {
	print_time "Removing stale sync file ($SYNC_FILE_NAME)...\n";
	unlink $SYNC_FILE_NAME;
	print "\n";
}

opendir(my $DIR, $tcl_directory) or die $!;
while(my $file = readdir($DIR)) {
	if($file =~ /$OUTPUT_RUNNING_SUFFIX\Z/) {
		print_time "Removing stale output file ($file)...\n";
		unlink $file;
	}
}
closedir($DIR);

print_time "Scanning for simulation scripts...\n";
my @tcl_scripts;
opendir(my $DIR, $tcl_directory) or die $!;
while(my $file = readdir($DIR)) {
	if(($file =~ /(.+)\.tcl\Z/) && (!-e $1.$OUTPUT_DONE_SUFFIX)) {
		open(my $FILE, $file);
		while(<$FILE>) {
			if($_ =~ /$VARIABLE_DELIMITER/) {
				print "ERROR: Variable delimiter \"$VARIABLE_DELIMITER\" found in script $file.\n";
				exit(2);
			}
		}
		close($FILE);
		push(@tcl_scripts, $file);
	}
}
closedir($DIR);
@tcl_scripts = sort by_run_index @tcl_scripts;
if(scalar(@tcl_scripts) > 0) {
	print_time "Found ".scalar(@tcl_scripts)." work-needing simulation script(s).\n";
	print "\n";
} else {
	print_time "No work-needing simulation scripts found, aborting...\n";
	print "\n";
	exit();
}

print_time "Checking which hosts are up:\n";
my @hosts;
for my $host (@HOSTS) {
	print_time "Checking $host... ";
	if(check_host($host, 1, 0, 1)) {
		print ", adding to list...\n";
		push(@hosts, $host);
	} else {
		print ", ignoring...\n";
	}
	if(scalar(@hosts) >= scalar(@tcl_scripts)) {
		last;
	}
}
if(scalar(@hosts) > 0) {
	print_time "".scalar(@hosts)." host(s) are available.\n";
	print "\n";
} else {
	print_time "No hosts are available, aborting...\n";
	print "\n";
	exit();
}

open(my $SYNC_FILE, "> $SYNC_FILE_NAME") or die $!;
print $SYNC_FILE "0";
close($SYNC_FILE);

my $process_count = 0;
my $pid;
while($process_count < scalar(@hosts)) {
	$process_count++;
	$pid = fork();
	if($pid == 0) {
		last;
	}
}

if($pid != 0) {
	#
	# Parent process.
	#
	while($process_count) {
		wait();
		$process_count--;
		print_time "$process_count host(s) still running...\n";
	}
	unlink $SYNC_FILE_NAME;
	print_time "All simulations complete.\n";
	exit();
}

#
# Child process.
#

# Retrieve an unique id.
sleep 1;	# Avoid race conditions with parent.
my $id;
open(my $SYNC_FILE, "+< $SYNC_FILE_NAME") or die $!;
flock($SYNC_FILE, LOCK_EX) or die $!;
sleep 1;	# Avoid race conditions later on.
$id = <$SYNC_FILE>;
if(!defined $id) {die "Didn't find an ID in sync file";}
seek($SYNC_FILE, 0, 0);
print $SYNC_FILE ($id + 1);
flock($SYNC_FILE, LOCK_UN);
close($SYNC_FILE);

# Set local host.
my $host = $hosts[$id];
my $script_id = $id;

while($script_id < scalar(@tcl_scripts)) {

	my $script = $tcl_scripts[$script_id];
	$script =~ /(.+)\.tcl\Z/o;
	my $script_name = $1;

	my $output_running = $script_name.$OUTPUT_RUNNING_SUFFIX;
	my $output_done = $script_name.$OUTPUT_DONE_SUFFIX;

	my $progess_info = "".($script_id + 1)."/".scalar(@tcl_scripts);
	if((!-e $output_running) && (!-e $output_done)) {
		print_time "$host: $script_name -> START ($progess_info)\n";
		while(1) {

			if(!check_host($host, 0, 0, 0)) {
				print_time "$host: Host is not available, shutting down...\n";
				exit();
			}

			# FIXME: Quick hack to avoid re-nicing on compute servers.
			my $nice = "nice";
			if($host =~ /knecht/) {
				$nice = "";
			}

			print `echo $host: $ns2_executable $script > $tcl_directory/$output_running`;
			# NOTE: This should be "&>>" instead of ">>" to
			# redirect and append. But bash doesn't support this at
			# the moment.
			print `ssh $SSH_OPTIONS $USER\@$host "cd $tcl_directory; $nice $ns2_executable $script >> $output_running"`;
			print_time "$host: $script_name -> STOP ($progess_info)\n";

			if($? == 0) {
				last;
			}

			open(my $OUTPUT_RUNNING, "<$output_running") or die $!;
			while(<$OUTPUT_RUNNING>) {
				if($_ =~ /Simulation complete/) {
					last;
				}
			}

			my($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time);
			$sec = ($sec < 10) ? "0$sec" : $sec;
			$min = ($min < 10) ? "0$min" : $min;
			$hour = ($hour < 10) ? "0$hour" : $hour;
			$mday = ($mday < 10) ? "0$mday" : $mday;
			$mon++;
			$mon = ($mon < 10) ? "0$mon" : $mon;
			$year += 1900;
			rename($output_running, $output_done."_ERROR_$?_".$host."_$year$mon$mday$hour$min$sec");
			print_time "===> $host: $script_name -> RESTART ($progess_info terminated with $?) <===\n";
		}
		rename($output_running, $output_done);
	}
	$script_id++;
}
exit();
